/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#include <glib.h>
#include <glib/gi18n-lib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkscrolledwindow.h>

#include <pan/base/argset.h>
#include <pan/base/article.h>
#include <pan/base/vserver.h>
#include <pan/base/debug.h>
#include <pan/base/group.h>
#include <pan/base/log.h>
#include <pan/base/status-item.h>
#include <pan/base/pan-base-signals.h>

#include <pan/action.h>
#include <pan/article-actions.h>
#include <pan/articlelist.h>
#include <pan/article-toolbar.h>
#include <pan/flagset.h>
#include <pan/globals.h>
#include <pan/header-pane-renderer.h>
#include <pan/prefs.h>
#include <pan/queue.h>
#include <pan/task-bodies.h>
#include <pan/task-save.h>
#include <pan/task-headers.h>
#include <pan/text.h>
#include <pan/util.h>
#include <pan/signals.h>
#include <pan/sorted-filter.h>

#include <pan/filters/filter-top.h>
#include <pan/filters/filter-score.h>
#include <pan/filters/score.h>

/***
****
****  Abandon hope, all ye who enter here
****
***/

/**********************************************************
 * Private Variables
 **********************************************************/

/* Since freeze/thaw has some overhead, don't do it if there
 * are only a few nodes to refresh.  10u is an arbitrary number. */
#define FREEZE_THRESHOLD 10u

/* state */
static Group * my_group                          = NULL;
static int sort_type                             = COLUMN_SUBJECT;

static GtkWidget * threads_popup_menu = NULL;
static GtkItemFactory * threads_popup_factory = NULL;
static GtkItemFactoryEntry threads_popup_entries[];
static int threads_popup_entries_qty;

typedef struct
{
	GNode *node;
	Article *a;
	GtkTreeIter iter;
} iter_tree;

static GMemChunk *iter_tree_chunk=NULL;
static GTree *article_to_it=NULL;
static GNode *article_tree=NULL;
static GPtrArray *my_articles=NULL; // not refd

static Article *article_cur=NULL,*article_prev=NULL; // not refd

static guint articles_changed_id=0;
static guint group_headers_retreived_id=0;
static guint selection_changed_id=0;

static void articlelist_repopulate_nolock(Group*, GPtrArray*,GPtrArray*);
static gboolean popup_menu_cb(GtkWidget *wig);
static void articlelist_menu_popup_nolock(GdkEventButton* bevent);
static void header_pane_refresh(int actions);

enum {
	REFRESH_TREE          = (1<<0),
	REFRESH_SCORE         = (1<<2),
	REFRESH_FILTER        = (1<<3),
	REFRESH_ALL           = (REFRESH_TREE | REFRESH_SCORE | REFRESH_FILTER)
};

/**********************************************************
 * Private Functions
 **********************************************************/

static int gtree_compare(gconstpointer a, gconstpointer b)
{
	if(a<b)
		return -1;
	if(a>b)
		return 1;
	return 0;
}

static iter_tree* new_iter_tree(void)
{
	return g_mem_chunk_alloc0(iter_tree_chunk);
}

static void free_iter_tree(gpointer v)
{
	g_mem_chunk_free(iter_tree_chunk,v);
}

static GTree* make_tree(void)
{
	return g_tree_new_full((GCompareDataFunc)gtree_compare,NULL,NULL,free_iter_tree);
}

static void gpa_unref(gpointer a, gpointer user)
{
	g_object_unref(ARTICLE(a));
}
static void gpa_ref(gpointer a, gpointer user)
{
	g_object_ref(ARTICLE(a));
}
static void gpa_mark_read(gpointer a, gpointer user)
{
	article_set_read_simple(ARTICLE(a),TRUE);
}

static gboolean filter_is_vis(GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	Article *a;
	gtk_tree_model_get(model,iter,ARTICLE_COL_OBJECT,&a,-1);
	return a->passes_filter;
}

static void row_expanded_cb(GtkTreeView *view, GtkTreeIter *iter, GtkTreePath *p, gpointer data)
{
	GtkTreeModel *m=gtk_tree_model_filter_get_model(GTK_TREE_MODEL_FILTER(gtk_tree_view_get_model(view)));
	GtkTreeStore *s=GTK_TREE_STORE(m);
	gtk_tree_store_set(s,iter,ARTICLE_COL_EXPANDED,TRUE,-1);
}
static void row_collapsed_cb(GtkTreeView *view, GtkTreeIter *iter, GtkTreePath *p, gpointer data)
{
	GtkTreeModel *m=gtk_tree_model_filter_get_model(GTK_TREE_MODEL_FILTER(gtk_tree_view_get_model(view)));
	GtkTreeStore *s=GTK_TREE_STORE(m);
	gtk_tree_store_set(s,iter,ARTICLE_COL_EXPANDED,FALSE,-1);
}

static void child_to_filter_iter(GtkTreeIter *client, GtkTreeIter *filter)
{
	GtkTreeModelFilter *f=GTK_TREE_MODEL_FILTER(gtk_tree_view_get_model(Pan.article_tree));
	gtk_tree_model_filter_convert_child_iter_to_iter(f,filter,client);
}

/***
****
****
****  SELECTIONS
****
****
***/

void articlelist_select_all_nolock(void)
{
	gtk_tree_selection_select_all(gtk_tree_view_get_selection(Pan.article_tree));
}

void articlelist_deselect_all_nolock(void)
{
	gtk_tree_selection_unselect_all(gtk_tree_view_get_selection(Pan.article_tree));
}

// iter is from filter model
static Article* articlelist_get_article_from_iter(GtkTreeIter* iter)
{
	Article* article = NULL;
	
	if(iter!=NULL)
		gtk_tree_model_get(gtk_tree_view_get_model(Pan.article_tree),ARTICLE_COL_OBJECT,&article,-1);
	
	return article;
}

// Free the iters with gtk_tree_iter_free
static void sel_to_iter(GtkTreeModel *m, GtkTreePath *p, GtkTreeIter *i, gpointer data)
{
	GPtrArray *list=data;
	g_ptr_array_add(list,gtk_tree_iter_copy(i));
}

// Free the iters with gtk_tree_iter_free
static GPtrArray* articlelist_get_selected_iters_nolock(void)
{
	GPtrArray * a;
	GtkTreeSelection *sel;
	guint32 c;
	debug_enter("articlelist_get_selected_iters_nolock");
	
	sel=gtk_tree_view_get_selection(Pan.article_tree);
	c=sel!=NULL?gtk_tree_selection_count_selected_rows(sel):0;
	a=g_ptr_array_sized_new(c);
	if(c>0)
		gtk_tree_selection_selected_foreach(sel,sel_to_iter,a);
	debug_exit("articlelist_get_selected_iters_nolock");
	return a;
}

static void tree_path_free_gfunc(gpointer a, gpointer d)
{
	gtk_tree_path_free((GtkTreePath*)a);
}

static gboolean articlelist_get_selected_iter(GtkTreeIter *iter)
{
	GtkTreeSelection *sel=gtk_tree_view_get_selection(Pan.article_tree);
	gboolean ret=FALSE;
	
	if(iter!=NULL && (gtk_tree_selection_count_selected_rows(sel)>0))
	{
		GtkTreeModel *m=NULL;
		GList *list=gtk_tree_selection_get_selected_rows(sel,&m);
		
		gtk_tree_model_get_iter(m,iter,(GtkTreePath*)(list->data));
		g_list_foreach(list, tree_path_free_gfunc, NULL);
		g_list_free(list);
		ret=TRUE;
	}
	return ret;
}

static void sel_to_article(GtkTreeModel *m, GtkTreePath *p, GtkTreeIter *i, gpointer data)
{
	GPtrArray *list=data;
	Article *a;
	gtk_tree_model_get(m,i,ARTICLE_COL_OBJECT,&a,-1);
	g_ptr_array_add(list,a);
}

static GPtrArray* _articlelist_get_selected_articles_nolock(void)
{
	GPtrArray * a;
	GtkTreeSelection *sel;
	guint32 c;
	debug_enter("articlelist_get_selected_articles_nolock");
	
	sel=gtk_tree_view_get_selection(Pan.article_tree);
	c=sel!=NULL?gtk_tree_selection_count_selected_rows(sel):0;
	a=g_ptr_array_sized_new(c);
	if(c>0)
		gtk_tree_selection_selected_foreach(sel,sel_to_article,a);
	debug_exit("articlelist_get_selected_articles_nolock");
	return a;
}

static void add_descendants_to_gptr_array(GtkTreeModel *m, GtkTreeIter *i, GPtrArray *array)
{
	Article *a;
	GtkTreeIter n;
	// first child
	gtk_tree_model_iter_children(m,&n,i);
	gtk_tree_model_get(m,&n,ARTICLE_COL_OBJECT,&a,-1);
	g_ptr_array_add(array,a);
	if(gtk_tree_model_iter_has_child(m,&n))
		add_descendants_to_gptr_array(m,&n,array);
	// other children
	while(gtk_tree_model_iter_next(m,&n))
	{
		gtk_tree_model_get(m,&n,ARTICLE_COL_OBJECT,&a,-1);
		g_ptr_array_add(array,a);
		if(gtk_tree_model_iter_has_child(m,&n))
			add_descendants_to_gptr_array(m,&n,array);
	}
}

static GPtrArray* _articlelist_get_active_articles_nolock(void)
{
	guint i;
	GPtrArray * retval;
	GPtrArray *iters;
	Article *a;
	GtkTreeModel *m;
	GtkTreeIter *iter;
	gboolean expnd;
	debug_enter("articlelist_get_active_articles_nolock");
	
	m=gtk_tree_view_get_model(Pan.article_tree);
	iters=articlelist_get_selected_iters_nolock();
	retval=g_ptr_array_sized_new(iters->len);
	
	for(i=0;i<iters->len;++i)
	{
		iter=g_ptr_array_index(iters,i);
		gtk_tree_model_get(m,iter,ARTICLE_COL_OBJECT,&a,ARTICLE_COL_EXPANDED,&expnd,-1);
		g_ptr_array_add(retval,a);
		if(!expnd && gtk_tree_model_iter_has_child(m,iter))
			add_descendants_to_gptr_array(m,iter,retval);
		gtk_tree_iter_free(iter);
	}
	g_ptr_array_free(iters,TRUE);
	
	debug_exit("articlelist_get_active_articles_nolock");
	return retval;
}

void header_pane_foreach_selected(ArticleForeach foreach, gpointer user_data)
{
	guint i;
	GPtrArray * a = _articlelist_get_selected_articles_nolock();
	for (i=0; i!=a->len; ++i)
		(foreach)(ARTICLE(g_ptr_array_index(a,i)), user_data);
	g_ptr_array_free(a, TRUE);
}

static void header_pane_first_selected_or_null(ArticleForeach foreach, gpointer user_data)
{
	GtkTreeIter i;
	Article * article=NULL;
	if(articlelist_get_selected_iter(&i))
		article = articlelist_get_article_from_iter(&i);
	(foreach)(article, user_data);
}

void header_pane_first_selected(ArticleForeach foreach, gpointer user_data)
{
	GtkTreeIter i={0,NULL,NULL,NULL};
	Article * article;
	if (articlelist_get_selected_iter(&i))
	{ 
		article= articlelist_get_article_from_iter(&i);
		(foreach)(article, user_data);
	}
}

void
header_pane_forall_selected(ArticleForall forall, gpointer user_data, gboolean call_forall_even_if_none_selected)
{
	GPtrArray * a = _articlelist_get_selected_articles_nolock();
	if (call_forall_even_if_none_selected || a->len)
		(forall)(a, user_data);
	g_ptr_array_free(a, TRUE);
}

void header_pane_foreach_active(ArticleForeach foreach, gpointer user_data)
{
	guint i;
	GPtrArray * a = _articlelist_get_active_articles_nolock();
	for (i=0; i!=a->len; ++i)
		(foreach)(ARTICLE(g_ptr_array_index(a,i)), user_data);
	g_ptr_array_free(a, TRUE);
}

void header_pane_forall_active(ArticleForall forall, gpointer user_data)
{
	GPtrArray * a = _articlelist_get_active_articles_nolock();
	if (a->len > 0)
		(forall)(a, user_data);
	g_ptr_array_free(a, TRUE);
}

guint articlelist_get_selected_count_nolock(void)
{
	if(Pan.article_tree==NULL)
		return 0;
	guint retval=0;
	GtkTreeSelection *s;
	debug_enter("articlelist_get_selected_count_nolock");
	
	if(s=gtk_tree_view_get_selection(Pan.article_tree))
		retval = gtk_tree_selection_count_selected_rows( s);

	debug_exit("articlelist_get_selected_count_nolock");
	return retval;

}

gboolean header_pane_has_selection(void)
{
	return (gboolean)articlelist_get_selected_count_nolock();
}

// free paths with gtk_tree_path_free
static GPtrArray* iter_to_path(GtkTreeModel *m, GPtrArray *iters)
{
	GPtrArray *ret=g_ptr_array_sized_new(iters->len);
	GtkTreeIter *i;
	GtkTreePath *p;
	int x;
	for(x=0;x<iters->len;++x)
	{
		i=g_ptr_array_index(iters,x);
		p=gtk_tree_model_get_path(m,i);
		g_ptr_array_add(ret,p);
	}
	return ret;
}

static void articlelist_set_selected_iters_nolock(GPtrArray *iters)
{
	g_return_if_fail(iters!=NULL);
	g_return_if_fail(iters->len>0);
	
	guint i;
	GtkTreeSelection *sel=NULL;
	GPtrArray *paths=NULL;
	GtkTreeModelFilter *f;
	GtkTreePath *p;
	gboolean is_tree=FALSE;
	debug_enter("articlelist_set_selected_iters_nolock");
	
	f=GTK_TREE_MODEL_FILTER(gtk_tree_view_get_model(Pan.article_tree));
	sel=gtk_tree_view_get_selection(Pan.article_tree);
	paths=iter_to_path(GTK_TREE_MODEL(f),iters);
	is_tree=GTK_IS_TREE_STORE(gtk_tree_model_filter_get_model(f));
	
	gdk_window_freeze_updates(GTK_WIDGET(Pan.article_tree)->window);
	
	gtk_tree_selection_unselect_all(sel);
	
	// NOTE: not syncing focus on selection
	for(i=0;i<paths->len;++i)
	{
		p=g_ptr_array_index(paths,i);
		if(is_tree)
			gtk_tree_view_expand_to_path(Pan.article_tree,p);
		gtk_tree_selection_select_path(sel,p);
		gtk_tree_path_free(p);
	}
	g_ptr_array_free(paths,TRUE);
	paths=NULL;
	
	gdk_window_thaw_updates(GTK_WIDGET(Pan.article_tree)->window);

	debug_enter("articlelist_set_selected_iters_nolock");
}

static void select_these_articles(GPtrArray *articles, gpointer unused)
{
	g_return_if_fail(articles!=NULL);
	GPtrArray *iters=g_ptr_array_sized_new(articles->len);
	GtkTreeIter *c,f;
	GtkTreeModelFilter *fltr=GTK_TREE_MODEL_FILTER(gtk_tree_view_get_model(Pan.article_tree));
	guint32 i;
	
	for(i=0;i<articles->len;++i)
	{
		c=g_tree_lookup(article_to_it,g_ptr_array_index(articles,i));
		if(c)
		{
			child_to_filter_iter(c,&f);
			g_ptr_array_add(iters,gtk_tree_iter_copy(&f));
		}
	}
	if(iters->len>0)
	{
		articlelist_set_selected_iters_nolock(iters);
		for(i=0;i<iters->len;++i)
			gtk_tree_iter_free((GtkTreeIter*)(g_ptr_array_index(iters,i)));
	}
	g_ptr_array_free(iters,TRUE);
}

static void get_top_iter(GtkTreeModel *m, GtkTreeIter *iter, GtkTreeIter *base)
{
	GtkTreeIter p;
	*iter=*base;
	while(gtk_tree_model_iter_parent(m,&p,iter))
		*iter=p;
}

static void articlelist_forall_in_subthread_i(GtkTreeIter *i, ArticleForall forall, gpointer user_data)
{
	GPtrArray *array=g_ptr_array_new();
	GtkTreeModel *m=gtk_tree_view_get_model(Pan.article_tree);
	
	g_ptr_array_add(array, articlelist_get_article_from_iter(i));
	
	if(gtk_tree_model_iter_has_child(m,i))
		add_descendants_to_gptr_array(m,i,array);
	forall(array,user_data);
	g_ptr_array_free(array,TRUE);
}
static void articlelist_forall_in_subthread(Article *a, ArticleForall forall, gpointer user_data)
{
	GPtrArray *array=g_ptr_array_new();;
	GtkTreeModel *m=gtk_tree_view_get_model(Pan.article_tree);
	GtkTreeIter *i=&((iter_tree*)g_tree_lookup(article_to_it,a))->iter;
	GtkTreeIter fi;
	
	child_to_filter_iter(i,&fi);
	g_ptr_array_add(array,a);
	if(gtk_tree_model_iter_has_child(m,&fi))
		add_descendants_to_gptr_array(m,&fi,array);
	forall(array,user_data);
	g_ptr_array_free(array,TRUE);
}
static void articlelist_forall_in_thread_i(GtkTreeIter *i, ArticleForall forall, gpointer user_data)
{
	GtkTreeModel *m=gtk_tree_view_get_model(Pan.article_tree);
	GtkTreeIter c;
	
	get_top_iter(m,&c,i);
	articlelist_forall_in_subthread_i(&c,forall,user_data);
}
static void articlelist_forall_in_thread(Article *a, ArticleForall forall, gpointer user_data)
{
	Article *h=NULL;
	GtkTreeIter *i=&((iter_tree*)g_tree_lookup(article_to_it,a))->iter;
	GtkTreeModel *m=gtk_tree_view_get_model(Pan.article_tree);
	GtkTreeIter c,p;
	
	child_to_filter_iter(i,&c);
	get_top_iter(m,&p,&c);
	h=articlelist_get_article_from_iter(&c);
	articlelist_forall_in_subthread(h,forall,user_data);
	
}
static void add_to_tree(GPtrArray *articles, gpointer data)
{
	GTree *tree=data;
	Article *a;
	guint i;
	
	for(i=0;i<articles->len;++i)
	{
		a=g_ptr_array_index(articles,i);
		g_tree_insert(tree,a,a);
	}
}
static gboolean tree_to_gpa(gpointer key, gpointer v, gpointer data)
{
	GPtrArray *a=data;
	g_ptr_array_add(a,key);
	return FALSE;
}
static void articlelist_forall_in_threads(GPtrArray *articles , 
		ThreadGet threadget, ArticleForall forall_func, 
		gpointer user_data)
{
	g_return_if_fail(articles!=NULL);
	g_return_if_fail(forall_func!=NULL);
	
	// use a hash to avoid duplicates
	GTree *tree=g_tree_new(gtree_compare);
	GPtrArray *gpa;
	Article *a;
	guint i;
	
	if(threadget==GET_SUBTHREAD)
	{
		for(i=0;i<articles->len;++i)
		{
			a=g_ptr_array_index(articles,i);
			articlelist_forall_in_subthread(a,add_to_tree,tree);
		}
	}
	else
	{
		for(i=0;i<articles->len;++i)
		{
			a=g_ptr_array_index(articles,i);
			articlelist_forall_in_thread(a,add_to_tree,tree);
		}
	}
	
	gpa=g_ptr_array_sized_new(g_tree_nnodes(tree));
	g_tree_foreach(tree,tree_to_gpa,gpa);
	g_tree_destroy(tree);
	tree=NULL;
	
	(forall_func)(gpa,user_data);
	g_ptr_array_free(gpa,TRUE);
}

static void select_these_articles_and_their_threads(GPtrArray *articles, gpointer thread_get_gpointer)
{
	const ThreadGet thread_get = GPOINTER_TO_INT (thread_get_gpointer);
	articlelist_forall_in_threads(articles, thread_get, select_these_articles, NULL);
}

void articlelist_add_replies_to_selection_nolock(void)
{
	header_pane_forall_selected(select_these_articles_and_their_threads, GINT_TO_POINTER(GET_SUBTHREAD), FALSE);
}
void articlelist_add_thread_to_selection_nolock(void)
{
	header_pane_forall_selected(select_these_articles_and_their_threads, GINT_TO_POINTER(GET_WHOLE_THREAD), FALSE);
}

static void select_iter_nolock(GtkTreeIter *iter)
{
	if(iter!=NULL)
		gtk_tree_selection_select_iter(gtk_tree_view_get_selection(Pan.article_tree),iter);
}

static void activate_iter_nolock(GtkTreeIter *iter)
{
	GtkTreePath *p;
	GtkTreeViewColumn *c;
	if(iter==NULL)
		return;
	gtk_tree_selection_select_iter(gtk_tree_view_get_selection(Pan.article_tree),iter);
	c=gtk_tree_view_get_column(Pan.article_tree,0);
	p=gtk_tree_model_get_path(gtk_tree_view_get_model(Pan.article_tree),iter);
	gtk_tree_view_row_activated(Pan.article_tree,p,c);
	gtk_tree_path_free(p);
}

/***
****
****
****  NAVIGATION
****
****
***/

typedef void (*IterFunc)(GtkTreeIter * iter, gpointer user_data);

static void tree_iter_next(GtkTreeIter *iter, GtkTreeIter *next)
{
	GtkTreeModel *m=gtk_tree_view_get_model(Pan.article_tree);
	GtkTreeIter p;
	
	if(iter==NULL)
	{
		gtk_tree_model_get_iter_first(m,next);
		return;
	}
	
	if(!gtk_tree_model_iter_children(m,next,iter))
	{
		*next=*iter;
		p=*iter;
		while(!gtk_tree_model_iter_next(m,next))
		{
			if(!gtk_tree_model_iter_parent(m,next,&p))
			{
				gtk_tree_model_get_iter_first(m,next);
				break;
			}
			p=*next;
		}
	}
}

static void tree_iter_prev(GtkTreeIter *iter, GtkTreeIter *prev)
{
	GtkTreeModel *m=gtk_tree_view_get_model(Pan.article_tree);
	GtkTreePath *path;
	int pos;
	
	// get last row
	if(iter==NULL)
	{
		pos=gtk_tree_model_iter_n_children(NULL,prev);
		gtk_tree_model_iter_nth_child(m,prev,NULL,pos-1);
		return;
	}
	// get previous row
	path=gtk_tree_model_get_path(m,iter);
	if(gtk_tree_path_prev(path))
	{
		gtk_tree_model_get_iter(m,prev,path);
		// get last child of previous row
		while(gtk_tree_model_iter_has_child(m,prev))
		{
			pos=gtk_tree_model_iter_n_children(NULL,prev);
			gtk_tree_model_iter_nth_child(m,prev,prev,pos-1);
		}
		return;
	}
	// get parent
	if(gtk_tree_model_iter_parent(m,prev,iter))
		return;
	// get last row
	pos=gtk_tree_model_iter_n_children(NULL,prev);
	gtk_tree_model_iter_nth_child(m,prev,NULL,pos-1);
}

/**
***
**/

static void articlelist_read_article(Article * article)
{
	g_return_if_fail(article!=NULL);
	
	article_prev=article_cur;
	article_cur=article;
	
	if(article->parts>1)
	{
		GPtrArray *parts=article_get_parts_weed_duplicates(article);
		text_set_from_articles(group_get_acache_key(my_group), group_get_default_charset(my_group),
			(Article**)parts->pdata, parts->len);
		g_ptr_array_foreach(parts,pan_g_ptr_array_unref,NULL);
		g_ptr_array_free(parts,TRUE);
	}
	else
		text_set_from_articles(group_get_acache_key(my_group), group_get_default_charset(my_group),&article,1);
}

static void articlelist_activate_article(Article * a)
{
	g_return_if_fail(a!=NULL);
	
	if (a->parts>3 && a->multipart_state==MULTIPART_STATE_ALL)
	{
		/* clicked on a big multipart; pop up the save dialog */
		article_action_selected_save_as();
	}
	else if (a->multipart_state!=MULTIPART_STATE_SOME && (queue_is_online() || article_is_cached(a)))
	{
		/* clicked on an article that we've got, or that we can get */
		articlelist_read_article(a);
	}
}

static void activate_article_cb(GtkTreeView *v, GtkTreePath *p, GtkTreeViewColumn *c, gpointer data)
{
	Article *a;
	GtkTreeModel *m=gtk_tree_view_get_model(Pan.article_tree);
	GtkTreeIter i;
	
	gtk_tree_model_get_iter(m,&i,p);
	gtk_tree_model_get(m,&i,ARTICLE_COL_OBJECT,&a,-1);
	articlelist_activate_article(a);
}

static void header_pane_download_all_single_part_except(GPtrArray *articles, gpointer except_me)
{
	guint i;
	Article *a;
    
	for (i=0; i!=articles->len; ++i)
	{
		a=g_ptr_array_index(articles,i);
		if (a!=except_me && a->parts<2)
			queue_add(TASK(task_bodies_new(&a, 1)));
	}
}

static void articlelist_read_article_and_download_other_active(Article * article)
{
	g_return_if_fail(article!=NULL);

	articlelist_read_article(article);

	header_pane_forall_active(header_pane_download_all_single_part_except, article);
}

void articlelist_read_selected(void)
{
	header_pane_first_selected((ArticleForeach)articlelist_read_article_and_download_other_active, NULL);
}

static void articlelist_activate_article_and_download_other_active(Article * article)
{
	g_return_if_fail(article!=NULL);
	
	articlelist_activate_article(article);

	header_pane_forall_active(header_pane_download_all_single_part_except, article);
}

void articlelist_activate_selected(void)
{
	header_pane_first_selected((ArticleForeach)articlelist_activate_article_and_download_other_active, NULL);
}

/***
****
***/

typedef void   (*NodeIteratorFunc)  (GtkTreeIter *start, GtkTreeIter *end);

static void
header_pane_find_next_iterated_node_from(NodeIteratorFunc     node_iterator_func,
                                          ArticleTest          test_func,
                                          gpointer             test_func_user_data,
                                          IterFunc             node_func,
                                          gpointer             node_func_user_data,
                                          GtkTreeIter       * start_node,
                                          gboolean             loop_around_when_end_reached)
{
	gboolean match = FALSE;
	GtkTreeIter march,next;

	/* sanity clause */
	g_return_if_fail(start_node!=NULL);

	/* walk through the nodes */
	march = *start_node;
	for (;;)
	{
		const Article * article;

		/* have we looped around to where we started? */
		(node_iterator_func)(&march,&next);
		march=next;
		if (memcmp(&march,start_node,sizeof(GtkTreeIter))==0)
			break;

		/* is there an article? */
		article = articlelist_get_article_from_iter(&march);

		/* do we have a match? */
		match = (test_func)(article, test_func_user_data);
		if (match)
			break;
	}

	if (match)
		(node_func)(&march, node_func_user_data);
}

static void
header_pane_find_next_iterated_node(NodeIteratorFunc     node_iterator_func,
                                     ArticleTest          test_func,
                                     gpointer             test_func_user_data,
                                     IterFunc             node_func,
                                     gpointer             node_func_user_data,
                                     gboolean             loop_around_when_end_reached)
{
	GtkTreeIter start_node;
	gboolean run;

	/* get the starting node */
	run=articlelist_get_selected_iter(&start_node);
	if (!run)
	{
		run=gtk_tree_model_get_iter_first(gtk_tree_view_get_model(Pan.article_tree),&start_node);
	}

	/* if we have a starting node, start iterating */
	if (run)
		header_pane_find_next_iterated_node_from(node_iterator_func,
		                                          test_func, test_func_user_data,
		                                          node_func, node_func_user_data,
		                                          &start_node,
		                                          loop_around_when_end_reached);
}

/***
****
***/

static void activate_node(GtkTreeIter * node, gpointer user_data)
{
	activate_iter_nolock(node);
}

static void select_node(GtkTreeIter * node, gpointer user_data)
{
	select_iter_nolock(node);
}

void header_pane_select_next_if(ArticleTest test_func, gpointer test_func_user_data)
{
	header_pane_find_next_iterated_node(tree_iter_next, test_func, test_func_user_data, select_node, NULL, TRUE);
}

void header_pane_select_prev_if(ArticleTest test_func, gpointer test_func_user_data)
{
	header_pane_find_next_iterated_node(tree_iter_prev, test_func, test_func_user_data, select_node, NULL, TRUE);
}

static void header_pane_read_next_if(ArticleTest test_func, gpointer test_func_user_data)
{
	header_pane_find_next_iterated_node(tree_iter_next, test_func, test_func_user_data, activate_node, NULL, TRUE);
}

static void header_pane_read_prev_if(ArticleTest test_func, gpointer test_func_user_data)
{
	header_pane_find_next_iterated_node(tree_iter_prev, test_func, test_func_user_data, activate_node, NULL, TRUE);
}

/**
***
**/

static gboolean test_article_for_ok(const Article * article, gpointer user_data)
{
	return article!=NULL && IS_ARTICLE(article) && !article_get_error_flag(article);
}
void header_pane_read_next(void)
{
	header_pane_read_next_if(test_article_for_ok, NULL);
}

static gboolean test_article_for_unread(const Article * article, gpointer user_data)
{
	return test_article_for_ok(article,NULL) && !article_is_read(article);
}
void header_pane_read_next_unread(void)
{
	header_pane_read_next_if(test_article_for_unread, NULL);
}

static gboolean test_article_for_new(const Article * article, gpointer user_data)
{
	return test_article_for_ok(article,NULL) && article_is_new(article);
}
void header_pane_read_next_new(void)
{
	header_pane_read_next_if(test_article_for_new, NULL);
}

static gboolean test_article_for_unread_and_positive_score(const Article * article, gpointer user_data)
{
	return test_article_for_unread(article,NULL) && article_get_score(article,my_group->id)>0;
}
void header_pane_read_next_score(void)
{
	header_pane_read_next_if(test_article_for_unread_and_positive_score, NULL);
}

static gboolean test_article_for_different_thread(const Article * article, gpointer other_article)
{
	return test_article_for_ok(article,NULL) && (other_article==NULL || !articles_are_in_same_thread(article,ARTICLE(other_article)));
}
static void read_thread_after_this_article(Article * article, gpointer unused)
{
	header_pane_read_next_if(test_article_for_different_thread, article);
}
void header_pane_read_next_thread(void)
{
	header_pane_first_selected_or_null(read_thread_after_this_article, NULL);
}

static gboolean test_article_for_unread_in_different_thread(const Article * article, gpointer other_article)
{
	return test_article_for_different_thread(article,other_article) && !article_is_read(article);
}
static void read_unread_after_this_thread(Article * article, gpointer unused)
{
	header_pane_read_next_if(test_article_for_unread_in_different_thread, article);
}
void header_pane_read_next_unread_thread(void)
{
	header_pane_first_selected_or_null(read_unread_after_this_thread, NULL);
}

static gboolean test_article_for_new_in_different_thread(const Article * article, gpointer other_article)
{
	return test_article_for_different_thread(article,other_article) && article_is_new(article);
}
static void read_new_after_this_thread(Article * article, gpointer unused)
{
	header_pane_read_next_if(test_article_for_new_in_different_thread, article);
}
void header_pane_read_next_new_thread(void)
{
	header_pane_first_selected_or_null(read_new_after_this_thread, NULL);
}

void header_pane_read_prev(void)
{
	header_pane_read_prev_if(test_article_for_ok, NULL);
}

static void read_thread_before_this_article(Article * article, gpointer unused)
{
	header_pane_read_prev_if(test_article_for_different_thread, article);
}
void header_pane_read_prev_thread(void)
{
	header_pane_first_selected_or_null(read_thread_before_this_article, NULL);
}

static gboolean test_article_for_identity(const Article * article, gpointer compare)
{
	return article == compare;
}

static void read_parent_of_this_article(Article * article, gpointer unused)
{
	Article *p;
	GtkTreeModel *m=gtk_tree_view_get_model(Pan.article_tree);
	GtkTreeIter *i,fi,pi;
	i=&((iter_tree*)g_tree_lookup(article_to_it,p))->iter;
	child_to_filter_iter(i,&fi);
	if (gtk_tree_model_iter_parent(m,&pi,&fi))
	{
		gtk_tree_model_get(m,&pi,ARTICLE_COL_OBJECT,&p,-1);
		header_pane_read_prev_if(test_article_for_identity, p);
	}
}
void header_pane_read_parent(void)
{
	header_pane_first_selected(read_parent_of_this_article, NULL);
}

/**
***
**/

gboolean articlelist_has_prev_read(void)
{
	return article_prev!=NULL;
}

void header_pane_read_prev_read(void)
{
	header_pane_read_prev_if(test_article_for_ok, NULL);
	if (articlelist_has_prev_read())
	{
		GtkTreeIter node,*c=&((iter_tree*)g_tree_lookup(article_to_it,article_prev))->iter;
		
		child_to_filter_iter(c,&node);
		activate_iter_nolock(&node);
	}
}


/**
***
**/

static const int SECONDS_IN_DAY = 60 * 60 * 24;

static void set_to_selection(GtkTreeModel *model,GtkTreePath *p, GtkTreeIter *i, gpointer data)
{
	GPtrArray *list=data;
	Article *a,*b;
	guint32 x;
	
	gtk_tree_model_get(model,i,ARTICLE_COL_OBJECT,&a,-1);
	
	for(x=0;x<my_articles->len;++x)
	{
		b=g_ptr_array_index(my_articles,x);
		if(abs(difftime(a->date, b->date)) < SECONDS_IN_DAY
			&& article_same_author(a,b)
			&& article_subjects_are_similar(a,b))
			g_ptr_array_add(list,&((iter_tree*)g_tree_lookup(article_to_it,b))->iter);
	}
}
void articlelist_add_set_to_selection_nolock(void)
{
	debug_enter("articlelist_add_set_to_selection_nolock");
	GtkTreeSelection *sel=gtk_tree_view_get_selection(Pan.article_tree);
	if(gtk_tree_selection_count_selected_rows(sel)>0)
	{
		int x;
		GtkTreeIter *c,f;
		GtkTreeModelFilter *filter=GTK_TREE_MODEL_FILTER(gtk_tree_view_get_model(Pan.article_tree));
		GPtrArray *list=g_ptr_array_sized_new(128);
		
		gtk_tree_selection_selected_foreach(sel,set_to_selection,list);
		for(x=0;x<list->len;++x)
		{
			c=g_ptr_array_index(list,x);
			gtk_tree_model_filter_convert_child_iter_to_iter(filter,&f,c);
			gtk_tree_selection_select_iter(sel,&f);
		}
		g_ptr_array_free(list,TRUE);
	}
	debug_exit("articlelist_add_set_to_selection_nolock");
}

/***
****
****
****   SORTING
****
****
***/
static int iter_comp(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer data)
{
	Article *aa, *ab;
	gtk_tree_model_get(model,a,ARTICLE_COL_OBJECT,&aa,-1);
	gtk_tree_model_get(model,b,ARTICLE_COL_OBJECT,&ab,-1);
	
	switch(GPOINTER_TO_INT(data))
	{
		case COLUMN_SCORE:
		{
			int ua=article_get_score(aa,my_group->id);
			int ub=article_get_score(ab,my_group->id);
			if(ua==ub) return 0;
			if(ua<ub) return -1;
			return 1;
		}
		case COLUMN_SUBJECT:
		{
			char ba[1024],bb[1024];
			article_get_subject(aa,ba,1024);
			article_get_subject(ab,bb,1024);
			return g_utf8_collate(ba,bb);
		}
		case COLUMN_LINES:
			if(aa->line_qty==ab->line_qty) return 0;
			if(aa->line_qty<ab->line_qty) return -1;
			return 1;
		case COLUMN_AUTHOR:
		{
			char ba[1024],bb[1024];
			article_get_author_str(aa,ba,1024);
			article_get_author_str(ab,bb,1024);
			//return g_utf8_collate(ba,bb);
			// FIXME: not right for non-ascii
			return g_ascii_strcasecmp(ba,bb);
		}
		case COLUMN_DATE:
			if(aa->date==ab->date) return 0;
			if(aa->date<ab->date) return -1;
			return 1;
	}
	return 0;
}

/****
*****
*****  MARKING ARTICLES FLAGED
*****
****/

static void header_pane_unflag_articles(GPtrArray *articles, gpointer user_data)
{
	flagset_remove_articles(articles);
}

void articlelist_selected_unflag_for_dl_nolock(void)
{
	header_pane_forall_active(header_pane_unflag_articles, NULL);
}

static void header_pane_flag_articles(GPtrArray *articles, gpointer user_data)
{
	flagset_add_articles(articles);
}

void articlelist_selected_flag_for_dl_nolock(void)
{
	header_pane_forall_active(header_pane_flag_articles, NULL);
}

/***
****  MAKE THE STORES
***/
static void build_trees(Group *g, GPtrArray *arts,GTree *tree, GNode *root)
{
	g_return_if_fail(g!=NULL || arts!=NULL || tree!=NULL || root!=NULL);
	
	iter_tree *it;
	Article *a,*p=NULL;
	GNode *node=NULL;
	guint i;
	
	for(i=0;i<arts->len;++i)
	{
		it=new_iter_tree();
		a=ARTICLE(g_ptr_array_index(arts,i));
		a->passes_filter=TRUE;
		it->a=a;
		p=article_get_parent(a,g);
		if(p)
			node=((iter_tree*)g_tree_lookup(tree,p))->node;
		if(!node)
			node=root;
		it->node=g_node_prepend_data(node,it);
		g_tree_insert(tree,a,it);
	}
}

static GtkTreeModel* make_list_store(GPtrArray *arts, GTree *tree)
{
	GtkListStore *store;
	GtkTreeModelFilter *m;
	debug_enter("make_list_store");
	
	store=gtk_list_store_new(NUM_ARTICLE_COLS,G_TYPE_OBJECT,G_TYPE_BOOLEAN);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store),COLUMN_SUBJECT,
		iter_comp,GINT_TO_POINTER(COLUMN_SUBJECT),NULL);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store),COLUMN_SCORE,
		iter_comp,GINT_TO_POINTER(COLUMN_SCORE),NULL);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store),COLUMN_LINES,
		iter_comp,GINT_TO_POINTER(COLUMN_LINES),NULL);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store),COLUMN_AUTHOR,
		iter_comp,GINT_TO_POINTER(COLUMN_AUTHOR),NULL);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store),COLUMN_DATE,
		iter_comp,GINT_TO_POINTER(COLUMN_DATE),NULL);
	if(arts!=NULL && arts->len>0)
	{
		g_assert(tree!=NULL);
		GtkTreeIter *iter;
		Article *a;
		iter_tree *it;
		int i;
		for(i=0;i<arts->len;++i)
		{
			a=ARTICLE(arts->pdata[i]);
			it=g_tree_lookup(tree,a);
			iter=&it->iter;
			gtk_list_store_append(store,iter);
			gtk_list_store_set(store,iter,
				ARTICLE_COL_OBJECT,a,
				ARTICLE_COL_EXPANDED,FALSE,
				-1);
		}
	}
	
	m=GTK_TREE_MODEL_FILTER(sorted_filter_new(store));
	gtk_tree_model_filter_set_visible_func(m,filter_is_vis,NULL,NULL);
	g_object_unref(store);
	
	debug_exit("make_list_store");
	return GTK_TREE_MODEL(m);
}

static gboolean node_into_store(GNode *node, gpointer data)
{
	if(node->data==NULL)
		return FALSE;
	
	GtkTreeStore *store=GTK_TREE_STORE(data);
	iter_tree *it=node->data;
	GtkTreeIter *iter=&it->iter;
	GtkTreeIter *p=NULL;
	
	if(node->parent && node->parent->data)
		p=&((iter_tree*)(node->parent->data))->iter;
	gtk_tree_store_append(store,iter,p);
	gtk_tree_store_set(store,iter,
				ARTICLE_COL_OBJECT,it->a,
				ARTICLE_COL_EXPANDED, FALSE,
				-1);
	return FALSE;
}

static GtkTreeModel* make_tree_store(GNode *root, GTree *tree)
{
	debug_enter("make_tree_store");
	GtkTreeStore *store;
	GtkTreeModelFilter *m;
	store=gtk_tree_store_new(NUM_ARTICLE_COLS,G_TYPE_OBJECT,G_TYPE_BOOLEAN);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store),COLUMN_SUBJECT,
		iter_comp,GINT_TO_POINTER(COLUMN_SUBJECT),NULL);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store),COLUMN_SCORE,
		iter_comp,GINT_TO_POINTER(COLUMN_SCORE),NULL);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store),COLUMN_DATE,
		iter_comp,GINT_TO_POINTER(COLUMN_DATE),NULL);
	if(root!=NULL && root->children!=NULL && tree!=NULL)
		g_node_traverse(root, G_LEVEL_ORDER, G_TRAVERSE_ALL, -1, node_into_store, store);
	
	m=GTK_TREE_MODEL_FILTER(sorted_filter_new(store));
	gtk_tree_model_filter_set_visible_func(m,filter_is_vis,NULL,NULL);
	g_object_unref(store);
	
	debug_exit("make_tree_store");
	return GTK_TREE_MODEL(m);
}

/***
****
****
****
***/

Group* articlelist_get_group(void)
{
	return my_group;
}

/***
****
****
****  POPULATING THE TREE:  UTILITY FUNCTIONS
****
****
***/


/* siblings get matched up to, as per the user's 'show' settings,
 * with one caveat: "don't show ignored articles" outweighs 
 * "show relatives". This way ignored articles must be explicitly
 * turned on in `match ignored articles' before they're shown. */
static void pass_filter_if_not_ignored(GPtrArray *articles, gpointer unused)
{
	Article *a;
	guint i;
	for (i=0; i!=articles->len; ++i)
	{
		a=ARTICLE(articles->pdata[i]);
		a->passes_filter = filter_score_get_score_mode(article_get_score(a,my_group->id)) != SCORE_IGNORED;
	}
}

static gboolean node_into_gpa(GNode *node, gpointer data)
{
	GPtrArray *array=data;
	g_ptr_array_add(array,((iter_tree*)(node->data))->a);
	return FALSE;
}

static void filter_forall_in_subthread(Article *a, ArticleForall func, gpointer data)
{
	iter_tree *it=g_tree_lookup(article_to_it,a);
	GPtrArray *array=g_ptr_array_sized_new(g_node_n_children(it->node)+1);
	g_node_traverse(it->node, G_LEVEL_ORDER, G_TRAVERSE_ALL, -1, node_into_gpa, array);
	(func)(array,data);
	g_ptr_array_free(array,TRUE);
}
static void filter_forall_in_thread(Article *a, ArticleForall func, gpointer data)
{
	iter_tree *it=g_tree_lookup(article_to_it,a);
	GNode *n=it->node;
	while(n->parent->data!=NULL)
		n=n->parent;
	it=(iter_tree*)n->data;
	filter_forall_in_subthread(it->a, func, data);
}

static void apply_filter_tests(Filter * filter, FilterShow show, GPtrArray *articles)
{
	register guint i;
	GPtrArray * tmp;

	g_return_if_fail (articles!=NULL);
	g_return_if_fail (articles->len > 0);

	/* clear the filter state */
	for (i=0; i!=articles->len; ++i)
		ARTICLE(g_ptr_array_index(articles,i))->passes_filter = FALSE;

	/* make a working copy of the headers */
	tmp = g_ptr_array_sized_new(articles->len);
	pan_g_ptr_array_append(tmp, articles->pdata, articles->len);

	/* remove the articles that don't pass */
	if (filter != NULL)
		filter_remove_failures(filter, tmp, my_group);

	/* process matching article's families */
	for (i=0; i<tmp->len; ++i)
	{
		register Article * a = (Article*) g_ptr_array_index(tmp, i);

		if (show == FILTER_SHOW_SUBTHREADS)
			filter_forall_in_subthread(a, pass_filter_if_not_ignored, NULL);
		else if (show == FILTER_SHOW_THREAD)
			filter_forall_in_thread(a, pass_filter_if_not_ignored, NULL);
	}

	/* process matching articles.
	   we do this after doing the families because pass_filter_if_not_ignored
	   will fail ignored articles even if we've got `show ignored' turned on. */
	for (i=0; i<tmp->len; ++i)
		ARTICLE(g_ptr_array_index(tmp, i))->passes_filter = TRUE;

	/* cleanup */
	g_ptr_array_free(tmp, TRUE);
}

/***
****
****
****  POPULATING THE TREE:  THE DIRTY WORK
****
****
***/

static void add_expanded(GtkTreeView *view, GtkTreePath *path, gpointer data)
{
	GPtrArray *exp=(GPtrArray*)data;
	Article *a;
	GtkTreeModel *model=gtk_tree_view_get_model(Pan.article_tree);
	GtkTreeIter iter;
	
	gtk_tree_model_get_iter(model,&iter,path);
	gtk_tree_model_get(model,&iter,ARTICLE_COL_OBJECT,&a,-1);
	g_ptr_array_add(exp,a);
}

static void refilter_threads(GPtrArray *articles);
// takes ownership of article
static void articlelist_repopulate_nolock(Group *group, GPtrArray *articles, GPtrArray *sel)
{
	debug_enter("articlelist_repopulate_nolock");
	
	GTree *tree=make_tree();
	GtkTreeModelFilter *filter;
	GtkTreeModel *model;
	GtkTreeViewColumn *fc;
	Article *cursor=NULL;
	GPtrArray *exp=NULL;
	GNode *root=g_node_new(NULL);
	gboolean is_tree=my_group!=NULL && GTK_IS_TREE_MODEL(gtk_tree_model_filter_get_model(GTK_TREE_MODEL_FILTER(gtk_tree_view_get_model(Pan.article_tree))));
	gboolean map_exp=header_pane_is_threaded && is_tree && (group==my_group);
	gboolean map_cursor=my_group==group;
	
	if(map_exp)
	{
		exp=g_ptr_array_sized_new(20);
		gtk_tree_view_map_expanded_rows(Pan.article_tree,add_expanded,exp);
	}
	
	if(map_cursor)
	{
		GtkTreePath *cp=NULL;
		GtkTreeIter ci;
		// get current cursor location
		gtk_tree_view_get_cursor(Pan.article_tree,&cp,&fc);
		if(cp)
		{
			gtk_tree_model_get_iter(gtk_tree_view_get_model(Pan.article_tree),&ci,cp);
			gtk_tree_model_get(gtk_tree_view_get_model(Pan.article_tree),ARTICLE_COL_OBJECT,&cursor,-1);
			gtk_tree_path_free(cp);
		}
	}
	
	build_trees(group,articles,tree,root);
	
	if(header_pane_is_threaded)
		model=make_tree_store(root,tree);
	else
		model=make_list_store(articles,tree);
	
	filter=GTK_TREE_MODEL_FILTER(model);
	
	my_group=group;
	if(article_to_it)
		g_tree_destroy(article_to_it);
	if(article_tree)
		g_node_destroy(article_tree);
	article_to_it=tree;
	article_tree=root;
	my_articles=articles;
	
	/* test them against the header pane's filter */
	if (group!=NULL)
	{
		Filter * filterx = NULL;
		FilterShow show = 0;
		//status_item_emit_status(status, _("Filtering Articles"));
		article_toolbar_set_group(group);
		if (articles->len)
		{
			/* article_toolbar_get_filter(&filterx, &show);
			apply_filter_tests(filterx, show, articles);
			g_object_unref(G_OBJECT(filterx)); */
			refilter_threads(articles);
		}
	}
	
	gtk_tree_view_set_model(Pan.article_tree,model);
	g_object_unref(model);
	
	if(map_exp)
	{
		int i;
		GtkTreeIter *c,f;
		GtkTreePath *p;
		for(i=0;i<exp->len;++i)
		{
			c=&((iter_tree*)g_tree_lookup(tree,g_ptr_array_index(exp,i)))->iter;
			if(c)
			{
				gtk_tree_model_filter_convert_child_iter_to_iter(filter,&f,c);
				p=gtk_tree_model_get_path(GTK_TREE_MODEL(filter),&f);
				gtk_tree_view_expand_row(Pan.article_tree,p,FALSE);
				gtk_tree_path_free(p);
			}
		}
		g_ptr_array_free(exp,TRUE);
	}

	if(sel)
		select_these_articles(sel,NULL);
	
	if(map_cursor && cursor)
	{
		iter_tree *it=g_tree_lookup(tree,cursor);
		GtkTreeIter f;
		GtkTreePath *cp;
		if(it)
		{
			gtk_tree_model_filter_convert_child_iter_to_iter(filter,&f,&it->iter);
			cp=gtk_tree_model_get_path(GTK_TREE_MODEL(filter),&f);
			gtk_tree_view_set_cursor(Pan.article_tree,cp,fc,FALSE);
			gtk_tree_path_free(cp);
		}
	}
	
	debug_exit("articlelist_repopulate_nolock");
}



/***
****
****
****  UPDATING EXISTING NODES: THE DIRTY WORK
****
****
***/

static void articles_changed_cb(PanBaseSignals *ps, Article **a, guint q, ArticleChangeEnum type, gpointer ud)
{
	// cheat: just make the treeview redraw
	// FIXME: need to handle new & unread counts here
	// for treestore only
	switch(type)
	{
		case ARTICLE_NEW:
		case ARTICLE_UNREAD:
		case ARTICLE_OTHER:
			gtk_widget_queue_draw(GTK_WIDGET(Pan.article_tree));
			break;
	}
}

static void refilter_threads(GPtrArray *articles)
{
	gint32 i;

	if (articles->len == 0u)
		return;

	/* refilter the articles, and in doing so,
	 * make a list of articles whose `passes' state has changed */
	if (1)
	{
		Filter * filter = NULL;
		FilterShow show = 0;
		article_toolbar_get_filter(&filter, &show);
		apply_filter_tests(filter, show, articles);
		g_object_unref(G_OBJECT(filter));
	}

	/* update the subthread_scores */
	if (1)
	{
		Article *a,*p;
		gint16 s;
		for (i=0; i<articles->len; ++i) {
			a=g_ptr_array_index(articles,i);
			a->subthread_score = -9999;
		}
		for(i=articles->len-1;i>-1;--i)
		{
			a=g_ptr_array_index(articles,i);
			p=article_get_parent(a,my_group);
			s=article_get_score(a,my_group->id);
			if(p)
				if(s>p->subthread_score)
					p->subthread_score=s;
		}
	}

}

static void scorefile_invalidated_cb(Signals *s, gboolean do_rescore, gpointer user_data_unused)
{
	if (my_group!=NULL && do_rescore)
	{
		gdk_window_freeze_updates(GTK_WIDGET(Pan.article_tree)->window);
		g_signal_handler_block(pan_base_signals_get(),articles_changed_id);
		
		ensure_articles_scored(my_group, (Article**)my_articles->pdata,my_articles->len,NULL);
		refilter_threads(my_articles);
		gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(gtk_tree_view_get_model(Pan.article_tree)));
		
		g_signal_handler_unblock(pan_base_signals_get(),articles_changed_id);
		gdk_window_thaw_updates(GTK_WIDGET(Pan.article_tree)->window);
	}
}

/***
****
****  SELECTION / KEYPRESS / BUTTONPRESS HANDLING
****
***/

/**
 * The following buttonpress, keypress, and selection callbacks are all here
 * to interact with each other to do one thing: force this damn GtkCTree to
 * have the UI behavior that we want, which is described in 
 * articlelist_selection_changed()
 */
static gboolean articlelist_key_press(GtkWidget      * widget,
									   GdkEventKey    * event,
									   gpointer         data)
{
	gboolean retval = FALSE;
	const gulong modmask=gtk_accelerator_get_default_mod_mask();
	const gulong mod = event->state & modmask;

	switch (event->keyval)
	{
		case GDK_leftarrow:
		case GDK_Left:
			articlelist_collapse_selected_threads();
			retval = TRUE;
			break;

		case GDK_rightarrow:
		case GDK_Right:
			articlelist_expand_selected_threads();
			retval = TRUE;
			break;
// FIXME: are these still needed?
		case GDK_Home:
			if (mod == GDK_SHIFT_MASK) {
				/* left shift-home behave as shift-control-home */
				event->state |= GDK_CONTROL_MASK;
			}
			else if (!mod) {
				/* go to the first thread */
				/* node = gtk_ctree_node_nth (_article_ctree, 1);
				if (node != NULL)
					select_node_nolock (node); */
				break;
			}

		case GDK_End:
			if (mod == GDK_SHIFT_MASK) {
				/* left shift-end behave as shift-control-end */
				event->state |= GDK_CONTROL_MASK;
			}
			else if (!mod) {
				/* go to the last thread */
				/* node = gtk_ctree_node_nth (_article_ctree, 0);
				if (node != NULL)
					node = gtk_ctree_last (_article_ctree, node);
				if (node != NULL)
					select_node_nolock (node); */
				break;
			}
	}

	return retval;
}

static gboolean articlelist_button_press(GtkWidget* widget, GdkEventButton* bevent)
{
	gboolean retval = FALSE;

	if( bevent->type==GDK_BUTTON_PRESS && bevent->button==3)
	{
		articlelist_menu_popup_nolock(bevent);
		retval = TRUE;
	}

	return retval;
}
/***
****
****   EXPAND / COLLAPSE THREADS
****
***/

static void collapse_expand_threads(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
	gboolean collapse=GPOINTER_TO_INT(data);
	
	if(collapse)
		gtk_tree_view_collapse_row(Pan.article_tree,path);
	else
		gtk_tree_view_expand_row(Pan.article_tree,path,FALSE);
}

void articlelist_expand_selected_threads(void)
{
	debug_enter("articlelist_expand_selected_threads");
	
	GtkTreeSelection *sel=gtk_tree_view_get_selection(Pan.article_tree);
	int cnt=gtk_tree_selection_count_selected_rows(sel);
	
	if(cnt>FREEZE_THRESHOLD)
		gdk_window_freeze_updates(GTK_WIDGET(Pan.article_tree)->window);
	
	gtk_tree_selection_selected_foreach(sel,collapse_expand_threads,GINT_TO_POINTER(FALSE));
	
	if(cnt>FREEZE_THRESHOLD)
		gdk_window_thaw_updates(GTK_WIDGET(Pan.article_tree)->window);
	debug_exit("articlelist_expand_selected_threads");
}

void articlelist_collapse_selected_threads(void)
{
	debug_enter("articlelist_collapse_selected_threads");
	
	GtkTreeSelection *sel=gtk_tree_view_get_selection(Pan.article_tree);
	int cnt=gtk_tree_selection_count_selected_rows(sel);
	
	if(cnt>FREEZE_THRESHOLD)
		gdk_window_freeze_updates(GTK_WIDGET(Pan.article_tree)->window);
	
	gtk_tree_selection_selected_foreach(sel,collapse_expand_threads,GINT_TO_POINTER(TRUE));
	
	if(cnt>FREEZE_THRESHOLD)
		gdk_window_thaw_updates(GTK_WIDGET(Pan.article_tree)->window);
	debug_exit("articlelist_collapse_selected_threads");
}

static void article_toolbar_user_changed_filter_cb(Signals *s, GPtrArray* foo, gpointer bar)
{
	header_pane_refresh(REFRESH_FILTER);
}

/***
****
***/

/*--------------------------------------------------------------------
 * generate the listing of articles, for the "Articles" tab
 *--------------------------------------------------------------------*/
static void tree_destroy_cb(void)
{
	debug_enter("tree_destroy_cb");
	
	g_signal_handlers_disconnect_by_func(signals_get(),
			article_toolbar_user_changed_filter_cb, NULL);
	g_signal_handlers_disconnect_by_func(signals_get(),
			scorefile_invalidated_cb, NULL);
	
	if(my_group)
	{
		g_signal_handler_disconnect(my_group,group_headers_retreived_id);
		my_group=NULL;
		group_headers_retreived_id=0;
	}
	g_ptr_array_free(my_articles,TRUE);
	my_articles=NULL;
	g_tree_destroy(article_to_it);
	article_to_it=NULL;
	g_node_destroy(article_tree);
	article_tree=NULL;
	g_object_unref(threads_popup_factory);
	threads_popup_factory=NULL;
	g_object_unref(threads_popup_menu);
	threads_popup_menu=NULL;
	g_mem_chunk_destroy(iter_tree_chunk);
	iter_tree_chunk=NULL;
	debug_exit("tree_destroy_cb");
}

gpointer create_articlelist_tree(void)
{
	int i;
	GtkTreeView * tree;
	GtkWidget * vbox;

	/* callbacks */
/* 	pan_callback_add (acache_get_bodies_added_callback(),
	                  refresh_articles_from_message_id_array_cb, NULL);
	pan_callback_add (acache_get_bodies_removed_callback(),
	                  refresh_articles_from_message_id_array_cb, NULL);
	pan_callback_add (queue_get_message_id_status_changed(),
	                  refresh_articles_from_message_id_array_cb, NULL);
	pan_callback_add (article_get_articles_changed_callback(),
	                  articlelist_articles_changed_cb, NULL);
	pan_callback_add (server_get_groups_removed_callback(),
	                  articlelist_groups_removed_cb, NULL);
	pan_callback_add (articlelist_get_selection_changed_callback(),
	                  articlelist_selection_changed_cb, NULL);
 */
//	signals_connect__filters_changed(signals_get(),
	signals_connect__article_toolbar_user_changed_filter(signals_get(),
			article_toolbar_user_changed_filter_cb, NULL);
	signals_connect__scorefile_invalidated(signals_get(),
			scorefile_invalidated_cb, NULL);

	//article_to_iter=make_hash_table();
	//my_articles=g_ptr_array_new();
	
	vbox=header_pane_renderer_new();
	tree=Pan.article_tree;
	if(!iter_tree_chunk);
		iter_tree_chunk=g_mem_chunk_create(iter_tree,1024,G_ALLOC_AND_FREE);
	/* create the right click popup menu */
	threads_popup_menu = menu_create_items (threads_popup_entries,
	                                        threads_popup_entries_qty,
	                                        "<ThreadView>",
	                                        &threads_popup_factory,
	                                        NULL);
	
	GtkTreeModel *m=make_tree_store(NULL,NULL);
	gtk_tree_view_set_model(tree,m);
	g_object_unref(m);
	
	/* connect signals */
	g_signal_connect(tree,"row-expanded",(GCallback)row_expanded_cb,NULL);
	g_signal_connect(tree,"row-collapsed",(GCallback)row_collapsed_cb,NULL);
	g_signal_connect(tree,"destroy",(GCallback)tree_destroy_cb,NULL);
	g_signal_connect(tree,"popup-menu",(GCallback)popup_menu_cb,NULL);
	g_signal_connect(tree,"button-press-event",(GCallback)articlelist_button_press,NULL);
	//g_signal_connect(tree, "key_press_event",(GCallback) articlelist_key_press, NULL);
	g_signal_connect(tree,"row-activated",(GCallback)activate_article_cb,NULL);
	articles_changed_id=pan_base_signals_connect__articles_changed(pan_base_signals_get(),
								articles_changed_cb,NULL);
	
/* 	
	g_signal_connect (tree, "tree-select-row", G_CALLBACK(tree_select_row_cb), NULL);
	g_signal_connect (tree, "tree-unselect-row", G_CALLBACK(tree_select_row_cb), NULL);
	g_signal_connect (tree, "key_release_event", G_CALLBACK(articlelist_key_release), NULL);
	g_signal_connect (tree, "style_set", G_CALLBACK(style_set_cb), NULL);

	pan_callback_add (group_get_articles_removed_callback(), group_articles_removed_cb, NULL);
	pan_callback_add (group_get_articles_added_callback(), group_articles_added_cb, NULL);
 */

	return vbox;
}

void articlelist_update_columns(void)
{
	header_pane_renderer_update_columns();
}
/**
***
***  ARTICLE THREADING
***
**/


void articlelist_set_threaded(gboolean threaded_on)
{
	if (threaded_on != header_pane_is_threaded)
	{
		header_pane_is_threaded = threaded_on;
		header_pane_refresh(REFRESH_TREE);
	}
}

/****
*****
*****  COPY TO FOLDER
*****
****/

static void
header_pane_copy_articles_to_folder(const GPtrArray *articles, gpointer folder_gpointer)
{
	Group * folder = GROUP(folder_gpointer);
	
	article_copy_articles_to_folder(folder, (Article**)articles->pdata, articles->len);
}

static void copy_to_folder_cb(gpointer user_data, int index, GtkWidget * w)
{
	GPtrArray * folders;
	Group * folder;

	/* get the target folder */
	folders = vserver_get_groups(VSERVER_FOLDERS);
	folder = GROUP( g_ptr_array_index(folders, index) );

	header_pane_forall_active((ArticleForall)header_pane_copy_articles_to_folder, folder);

	/* cleanup */
	int i;
	for(i=0;i<folders->len;++i)
		g_object_unref( GROUP( g_ptr_array_index(folders,i) ) );
	g_ptr_array_free(folders, TRUE);
}

/***
****  POPUP MENU
***/

static void refresh_folder_popup_menu(void)
{
	int i;
	GPtrArray * folders;
	static GtkItemFactoryEntry * entries = NULL;
	static int qty = 0;
	debug_enter ("refresh_folder_popup_menu");

	/* remove any old server buttons */
	if (entries != NULL) {
		gtk_item_factory_delete_entries (threads_popup_factory, qty, entries);
		for (i=0; i<qty; ++i)
			g_free (entries[i].path);
		g_free (entries);
		entries = NULL;
		qty = 0;
	}

	/* get a list of folders */
	folders = vserver_get_groups(VSERVER_FOLDERS);

	/* build the ItemFactoryEntries */
	entries = g_new0 (GtkItemFactoryEntry, folders->len);
	for (i=qty=0; i<folders->len; ++i)
	{
		Group * folder = GROUP(g_ptr_array_index(folders,i));
		const PString *name = group_get_name(folder);

		/* build an entry: the callback arg is the index into the folders array */
		entries[qty].path = g_strdup_printf ("/_Copy to Folder/%*.*s", name->len, name->len, name->str);
		entries[qty].item_type = NULL;
		entries[qty].callback = copy_to_folder_cb;
		entries[qty].callback_action = i;

		++qty;
		g_object_unref(folder);
	}

	/* add the servers */
	gtk_item_factory_create_items (threads_popup_factory, qty, entries, NULL);

	/* cleanup */
	g_ptr_array_free (folders, TRUE);

	debug_exit ("refresh_folder_popup_menu");
}

static void articlelist_menu_popup_nolock(GdkEventButton* bevent)
{
	GtkItemFactory * gif = threads_popup_factory;
	int b=0,t;
	const gboolean have_article = header_pane_has_selection();

	refresh_folder_popup_menu();

	menu_set_sensitive (gif, "/Read Article", have_article);
	menu_set_sensitive (gif, "/Save Article As...", have_article);
	menu_set_sensitive (gif, "/Save Article Attachments", have_article);
	menu_set_sensitive (gif, "/Manual Decode...", have_article);
	menu_set_sensitive (gif, "/Download Flagged", have_article);
	menu_set_sensitive (gif, "/Flag", have_article);
	menu_set_sensitive (gif, "/Unflag", have_article);
	menu_set_sensitive (gif, "/Watch Thread", have_article);
	menu_set_sensitive (gif, "/Ignore Thread", have_article);
	menu_set_sensitive (gif, "/View Article's Scores", have_article);
	menu_set_sensitive (gif, "/Create Score", have_article);
	menu_set_sensitive (gif, "/Add Set to Selection", have_article);
	menu_set_sensitive (gif, "/Copy to Folder", have_article);
	menu_set_sensitive (gif, "/Delete", have_article);
// FIXME: menu position function when bevent==NULL
	if(bevent)
	{
		b=bevent->button;
		t=bevent->time;
	}
	else
		t=gtk_get_current_event_time();
	gtk_menu_popup(GTK_MENU (threads_popup_menu),
			NULL, NULL, NULL, NULL,
			b,t);
}

static gboolean popup_menu_cb(GtkWidget *wig)
{
	articlelist_menu_popup_nolock(NULL);
	return TRUE;
}

static void menu_cb(gpointer user_data, int iaction, GtkWidget * w)
{
        const PanAction action = (PanAction) iaction;
        pan_action_do(action);
}
static int threads_popup_entries_qty = 20;
static GtkItemFactoryEntry threads_popup_entries[] =
{
	/* r */ {N_("/_Read Article"), NULL, articlelist_read_selected, 0, NULL},
	/*   */ {N_("/---"), NULL, NULL, 0, "<Separator>"},
	/* s */ {N_("/_Save Article As..."), NULL, menu_cb, ACTION_SAVE_AS, NULL},
	/* t */ {N_("/Save Article A_ttachments"), NULL, menu_cb, ACTION_SAVE, NULL},
	/* m */ {N_("/_Manual Decode..."), NULL, menu_cb, ACTION_MANUAL_DECODE, NULL},
	/*   */ {N_("/---"), NULL, NULL, 0, "<Separator>"},
	/* o */ {N_("/D_ownload Flagged"), NULL, flagset_flush, 0, NULL},
	/* f */ {N_("/_Flag"), NULL, articlelist_selected_flag_for_dl_nolock, 0, NULL},
	/* u */ {N_("/_Unflag"), NULL, articlelist_selected_unflag_for_dl_nolock, 0, NULL},
	/*   */ {N_("/---"), NULL, NULL, 0, "<Separator>"},
        /* w */ {N_("/_Watch Thread"), NULL, menu_cb, ACTION_SCORE_WATCH_THREAD, NULL},
        /* i */ {N_("/_Ignore Thread"), NULL, menu_cb, ACTION_SCORE_IGNORE_THREAD, NULL},
        /* v */ {N_("/_View Article's Scores"), NULL, menu_cb, ACTION_SCORE_VIEW, NULL},
        /*   */ {N_("/_Create Score"), NULL, menu_cb, ACTION_ADD_TO_SCOREFILE, NULL},
	/*   */ {N_("/---"), NULL, NULL, 0, "<Separator>"},
	/* e */ {N_("/Add S_et to Selection"), NULL, articlelist_add_set_to_selection_nolock, 0, NULL},
	/*   */ {N_("/---"), NULL, NULL, 0, "<Separator>"},
	/* c */ {N_("/_Copy to Folder"), NULL, NULL, 0, "<Branch>"},
	/*   */ {N_("/---"), NULL, NULL, 0, "<Separator>"},
	/* d */ {N_("/_Delete"), NULL, menu_cb, ACTION_ARTICLE_DELETE, NULL}
};

/***
****
****  REFRESH
****
***/
static gboolean _pane_is_about_to_be_rebuilt = FALSE;

static int refresh_mainthread_end(gpointer p)
{
	ArgSet * argset;
	Group * group;
	StatusItem * status;
	GPtrArray * articles;
    int actions;
	debug_enter ("refresh_mainthread_end");

	/* pump out the arguments */
	argset = (ArgSet*) p;
	group = (Group*) argset_get(argset, 0);
	articles = (GPtrArray*) argset_get(argset, 1);
	status = (StatusItem*) argset_get(argset, 2);
	actions = GPOINTER_TO_INT (argset_get(argset, 3));

	/* if we've changed anything at all, we need to update the GUI */
	if (actions & REFRESH_TREE)
	{
		GPtrArray * tmp = _articlelist_get_selected_articles_nolock();
		articlelist_repopulate_nolock(group, articles, tmp);
		g_ptr_array_free(tmp, TRUE);
		articles=NULL;
	}
	else
	{
		GtkTreeModelFilter *fltr=GTK_TREE_MODEL_FILTER(gtk_tree_view_get_model(Pan.article_tree));
		gtk_tree_model_filter_refilter(fltr);
	}

	/* cleanup - unref the safe-keeping refs */
	if(articles)
		g_ptr_array_foreach(articles,gpa_unref,NULL);

	/* cleanup - done with status-item */
	status_item_set_active(status, FALSE);
	g_object_unref(G_OBJECT(status));
	
	/* cleanup - free the tmp stuff */
	argset_free(argset);

	debug_exit("refresh_mainthread_end");
	return 0;
}

static gboolean refresh_worker(gpointer p)
{
	ArgSet         * argset;
	Group          * group;
	StatusItem     * status;
	GPtrArray      * articles;
	int              actions;
	debug_enter("refresh_worker");

	/* pump out the arguments */
	argset = (ArgSet*) p;
	group = (Group*) argset_get(argset, 0);
	articles = (GPtrArray*) argset_get(argset, 1);
	status = (StatusItem*) argset_get(argset, 2);
	actions = GPOINTER_TO_INT (argset_get(argset, 3));

	g_signal_handler_block(pan_base_signals_get(),articles_changed_id);

	/* make sure the scoring is up-to-date */
	if (actions & REFRESH_SCORE) {
		status_item_emit_status(status, _("Scoring Articles"));
		ensure_articles_scored(group, (Article**)articles->pdata,articles->len, NULL);
	}

	/* filter the articles */
	if (actions & (REFRESH_FILTER | REFRESH_SCORE)) {
		status_item_emit_status(status, _("Filtering Articles"));
		if (articles->len)
			refilter_threads(articles);
	}

	g_signal_handler_unblock(pan_base_signals_get(),articles_changed_id);

	/* hard work done; back to the main thread */
	status_item_emit_status(status, _("Updating Header Pane..."));
	//??? g_usleep (75000);
	g_idle_add(refresh_mainthread_end, argset);

	debug_exit("refresh_worker");
	return FALSE;
}

static int refresh_mainthread_begin(gpointer actions_gpointer)
{
	Group * group;
	GPtrArray *articles;
	char buf[512];
	StatusItem * status;
	debug_enter("refresh_mainthread_begin");

	_pane_is_about_to_be_rebuilt = FALSE;

	group = my_group;
	if(group==NULL)
	{
		debug_exit("refresh_mainthread_begin");
		return 0;
	}

	/* create the status-item */
	g_snprintf (buf, sizeof(buf), _("Refreshing Group \"%s\""), group_get_name(group)->str);
	status = status_item_new_with_description(buf);
	status_item_set_active(status, TRUE);
	
	if(GPOINTER_TO_INT(actions_gpointer) & REFRESH_TREE)
		articles=group_get_all_articles(group,status);
	else
	{
		articles=my_articles;
		g_ptr_array_foreach(articles,gpa_ref,NULL);
	}
	
	g_idle_add(refresh_worker, argset_new4 (group, articles, status, actions_gpointer));

	debug_exit("refresh_mainthread_begin");
	return 0;
}

static void header_pane_refresh(int actions)
{
	if (!_pane_is_about_to_be_rebuilt)
	{
		_pane_is_about_to_be_rebuilt = TRUE;

		g_idle_add(refresh_mainthread_begin, GINT_TO_POINTER(actions));
	}
}

/***
****
****  SET THE GROUP
****
***/

static void group_headers_retreived_cb(Group *ps, gpointer ud)
{
	header_pane_refresh(REFRESH_ALL);
}

static gboolean fire_group_changed(gpointer group)
{
	signals_articlelist_group_changed(signals_get(),Pan.article_tree,GROUP(group));
	return FALSE;
}

typedef struct
{
	Group *group;
	GHashTable *a2i;
	StatusItem *status;
	GPtrArray *articles;
} updt_s;

static void set_group_field(updt_s *p)
{
	debug_enter("set_group_field");
	
	g_signal_handler_block(pan_base_signals_get(),articles_changed_id);
	
	if(my_group && my_group!=p->group && mark_read_on_group_exit)
		g_ptr_array_foreach(my_articles,gpa_mark_read,NULL);
	
	g_signal_handler_unblock(pan_base_signals_get(),articles_changed_id);
	
	articlelist_repopulate_nolock(p->group,p->articles,NULL);
	
	g_idle_add(fire_group_changed,p->group);
	
	debug_exit("set_group_field");
}

static gboolean clear_group(gpointer unused)
{
	if(my_group==NULL)
		return 0;
	
	g_signal_handler_disconnect(my_group,group_headers_retreived_id);
	group_headers_retreived_id=0;
	
	updt_s p;
	p.group=NULL;
	//p.a2i=make_hash_table();
	p.articles=g_ptr_array_new();
	
	/* clear the model */
	set_group_field(&p);
	
	return 0;
}

static gboolean open_download_dialog(gpointer data)
{
	/* This is a bit ugly */
	group_action_selected_download_dialog();
	return 0;
}

static gboolean set_group_mainthread_end(gpointer p)
{
	debug_enter("set_group_mainthread_end");
	
	updt_s *args=(updt_s*)p;
	StatusItem *status=args->status;
	gboolean enter=my_group!=args->group;
	
	/* no longer using the status */
	status_item_set_active(status, FALSE);
	g_object_unref(G_OBJECT(status));
	
	//articlelist_repopulate_nolock(args->group,args->articles,NULL);
	set_group_field(args);
	g_ptr_array_foreach(args->articles,gpa_unref,NULL);
	g_free(args);
	group_headers_retreived_id=group_connect__headers_retrieved(my_group, group_headers_retreived_cb,NULL);
	
	/* Now that the articlelist is fully loaded,
	   get new headers, if required. */
	if (fetch_new_on_group_enter && !group_is_folder(my_group) && enter)
	{
		Task * o = TASK(task_headers_new(my_group, HEADERS_NEW));
		GSList * l = g_slist_append(NULL, o);
		queue_insert_tasks(l, 0);
		g_slist_free(l);
	}
	
	debug_exit("set_group_mainthread_end");
	return 0;
}

static gboolean set_group_worker(gpointer p)
{
	debug_enter("set_group_worker");
	updt_s *args=(updt_s*)p;
	GPtrArray *articles;
	Group *group=args->group;
	StatusItem *status=args->status;
	
	GTimer *timer=g_timer_new();
	g_timer_start(timer);
	articles=group_get_all_articles(group,status);
	args->articles=articles;
	g_timer_stop(timer);
	log_add_va(LOG_INFO,"Loaded %d articles int %.4f seconds.",articles->len,g_timer_elapsed(timer,NULL));
	
	g_signal_handler_block(pan_base_signals_get(),articles_changed_id);
	
	/* massage them into shape */
	status_item_emit_status(status, _("Scoring Articles"));
	ensure_articles_scored(group, (Article**)articles->pdata,articles->len,NULL);
	
	
	g_signal_handler_unblock(pan_base_signals_get(),articles_changed_id);
	
	status_item_emit_status(status, _("Updating Header Pane..."));
	g_idle_add(set_group_mainthread_end, args);
	
	debug_exit("set_group_worker");
	return FALSE;
}

static gboolean set_group_mainthread_begin(gpointer data)
{
	debug_enter("set_group_mainthread_begin");
	char buf[512];
	Group *group=GROUP(data);
	updt_s *args=g_new0(updt_s,1);
	
	args->group=group;
	//args->a2i=make_hash_table();
	
	/* create the status-item */
	g_snprintf(buf, sizeof(buf), _("Loading group \"%s\""), group_get_name(group)->str);
	args->status = status_item_new_with_description(buf);
	status_item_set_active(args->status, TRUE);
	
	g_idle_add(set_group_worker, args);
	
	if(group_headers_retreived_id)
	{
		g_signal_handler_disconnect(my_group,group_headers_retreived_id);
		group_headers_retreived_id=0;
	}
	
	debug_exit("set_group_mainthread_begin");
	return 0;
}

void articlelist_set_group(Group * group)
{
	if (group == NULL && my_group!=NULL)
		g_idle_add(clear_group, NULL);
	else if (group !=NULL && !group_is_folder(group)
	         && !group->article_qty
	         )//&& !group->article_high)
		g_idle_add(open_download_dialog, NULL);
	else if(group!=my_group)
	{
		g_idle_add(clear_group, NULL);
		g_idle_add(set_group_mainthread_begin, group);
	}
}

