/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>

#include <glib.h>
#include <glib/gi18n-lib.h>

#include <gmime/gmime.h>
#include <gmime/md5-utils.h>

#include <pan/base/acache.h>
#include <pan/base/pan-base-signals.h>
#include <pan/base/argset.h>
#include <pan/base/base-prefs.h>
#include <pan/base/debug.h>
#include <pan/base/serverlist.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/util-file.h>
#include <pan/base/util-mime.h>
#include <pan/base/log.h>
#include <pan/base/db.h>
#include <pan/base/group.h>

static gint64 cache_size=0;



static void acache_get_path(sqlite3 *db, char *buf, int buflen, const char *key);

static void acache_add_path(sqlite3 *db, const char * key, const char * path);

/**
 * This is the name of the default directory
 */
const char * ACACHE_DEFAULT_KEY = "cache";


/***
****
****  DIRECTORIES & FILENAMES
****
***/

static char*
acache_get_message_base (void)
{
	return g_build_filename (get_data_dir(), "messages", NULL);
}
static char*
acache_get_folder_base (void)
{
	char * base = acache_get_message_base ();
	char * retval = g_build_filename (base, "folders", NULL);
	g_free (base);
	return retval;
}

/**
 * Some characters in message-ids don't work well in file names,
 * so we transform them to a safer name.
 */
static char*
acache_message_id_to_basename (char * buf, int len, const PString * message_id)
{
	const char * in;
	char * out;
	debug_enter ("acache_message_id_to_basename");

	/* sanity clause */
	g_return_val_if_fail (message_id!=NULL, NULL);
	g_return_val_if_fail (message_id->str!=NULL, NULL);
	g_return_val_if_fail (buf!=NULL, NULL);
	g_return_val_if_fail (len>0, NULL);

	/* some characters in message-ids are illegal on older Windows boxes,
	 * so we transform those illegal characters using URL encoding */
	for (in=message_id->str, out=buf; *in; ++in) {
		switch (*in) {
			case '%': /* this is the escape character */
			case '"': case '*': case '/': case ':': case '?': case '|':
			case '\\': /* these are illegal on vfat, fat32 */
				g_snprintf (out, len-(out-buf), "%%%02x", (int)*in);
				out += 3;
				break;
			case '<': case '>': /* these are illegal too, but rather than encoding
			                       them, follow the convention of omitting them */
				break;
			default:
				*out++ = *in;
		}
	}
	g_snprintf (out, len-(out-buf), ".msg");

	debug_exit ("acache_message_id_to_basename");
	return buf;
}

/**
 * Message-IDs are transformed via acache_message_id_to_basename()
 * to play nicely with some filesystems, so to extract the Message-ID
 * from a basename we need to reverse the transform.
 *
 * @return string length, or 0 on failure
 */
static int
acache_basename_to_message_id (char * buf, int len, const char * basename)
{
	const char * in;
	char * out;
	char * pch;
	char tmp_basename[PATH_MAX];
	debug_enter ("acache_basename_to_message_id");

	/* sanity clause */
	g_return_val_if_fail (is_nonempty_string(basename), 0);
	g_return_val_if_fail (buf!=NULL, 0);
	g_return_val_if_fail (len>0, 0);

	/* remove the trailing ".msg" */
	g_strlcpy (tmp_basename, basename, sizeof(tmp_basename));
	if ((pch = g_strrstr (tmp_basename, ".msg")))
		*pch = '\0';
	g_strstrip (tmp_basename);

	/* transform */
	out = buf;
	/* *out++ = '<'; */
	for (in=tmp_basename; *in; ++in) {
		if (in[0]!='%' || !g_ascii_isxdigit(in[1]) || !g_ascii_isxdigit(in[2]))
			*out++ = *in;
		else {
			char buf[3];
			buf[0] = *++in;
			buf[1] = *++in;
			buf[2] = '\0';
			*out++ = (char) strtoul (buf, NULL, 16);
		}
	}
	/* *out++ = '>'; */
	*out = '\0';

	/* cleanup */
	debug_exit ("acache_basename_to_message_id");
	return out - buf;
}

static char*
acache_get_filename (char * buf, int len, const char * path, const PString * message_id)
{
	char basename[PATH_MAX] = { '\0' };

	/* sanity clause */
	g_return_val_if_fail (buf!=NULL, NULL);
	g_return_val_if_fail (len>0, NULL);
	g_return_val_if_fail (pstring_is_set (message_id),NULL);

	/* build the filename */
	*buf = '\0';
	if (acache_message_id_to_basename (basename, sizeof(basename), message_id)) {
		g_snprintf (buf, len, "%s%c%s", path, G_DIR_SEPARATOR, basename);
		pan_file_normalize_inplace (buf);
	}

	return is_nonempty_string(buf) ? buf : NULL;
}

/***
****
****
****
***/

static int
get_message_id_from_file (char * msgid, int msgid_len, const char * filename)
{
	const char * basename;
	int retval = 0;

	/* sanity clause */
	g_return_val_if_fail (is_nonempty_string(filename), FALSE);
	g_return_val_if_fail (msgid!=NULL, FALSE);
	g_return_val_if_fail (msgid_len>0, FALSE);
	
	/* get the message-id */
	*msgid = '\0';
	basename = g_basename (filename);
	if (basename != NULL)
		retval = acache_basename_to_message_id (msgid, msgid_len, basename);

	return retval;
}

void acache_build_cache()
{
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	char * message_base;
	char * folder_base;
	char * path;
	GDir * dir;
	GError * err;
	int sq_res;
	
	debug_enter("acache_build_cache");
	
	
	message_base = acache_get_message_base();
	folder_base = acache_get_folder_base();
	
	/* ensure pan.sent exists */
	path = g_build_filename(folder_base, g_quark_to_string(PAN_SENT()), NULL);
	pan_file_ensure_path_exists(path);
	replace_gstr(&path, NULL);

	/* ensure pan.sendlater exists */
	path = g_build_filename(folder_base, g_quark_to_string(PAN_SENDLATER()), NULL);
	pan_file_ensure_path_exists(path);
	replace_gstr(&path, NULL);

	/* add all the folder directories */
	err = NULL;
	dir = g_dir_open(folder_base, 0, &err);
	if (err != NULL)
	{
		log_add_va(LOG_ERROR, _("Error opening directory \"%s\": %s)"), folder_base, err->message);
		g_error_free(err);
	}
	else
	{
		const char * key;
		
		/* add the folders */
		while ((key = g_dir_read_name(dir)))
		{
			/* add the folder */
			acache_add_folder(key);
		}
		
		g_dir_close(dir);

	}

	/* ensure the cache directory exist */
	path = g_build_filename(message_base, ACACHE_DEFAULT_KEY, NULL);
	acache_add_path(db, ACACHE_DEFAULT_KEY, path);
	replace_gstr(&path, NULL);
	sq_res=sqlite3_prepare(db,"SELECT sum(size) FROM cache,cache_paths AS cp "
						"WHERE key='cache' AND keyid=cp.id;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_step(stmt);
	cache_size=sqlite3_column_int64(stmt,0);
	sqlite3_finalize(stmt);

	/* cleanup */
	g_free(folder_base);
	g_free(message_base);
	
	debug_exit("acache_build_cache");
}

void acache_init(void)
{
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	int sq_res;
	
	sq_res=sqlite3_prepare(db,"SELECT SUM(size) FROM cache_paths AS cp, cache "
						"WHERE key=? AND keyid=cp.id;", -1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_text(stmt,1,ACACHE_DEFAULT_KEY,-1,SQLITE_STATIC);
	sqlite3_step(stmt);
	cache_size=sqlite3_column_int64(stmt,0);
	sqlite3_finalize(stmt);
}

/***
****
****  EXPIRE
****
***/
// TODO: this can be put in 1 loop
static void acache_remove_entries(sqlite3 *db, const char *key, const char **names,
	const gint64 *ids, int len)
{
	sqlite3_stmt *stmt;
	char filename[PATH_MAX], path[PATH_MAX];
	const char *name;
	int i, sq_res;
	
	acache_get_path(db,path,PATH_MAX,key);
	
	//sqlite3_prepare(db,"SELECT file FROM cache WHERE id=?;",-1,&stmt,NULL);
	for(i=0;i<len;i++)
	{
		//sqlite3_bind_int64(stmt,1,ids[i]);
		//sqlite3_step(stmt);
		//name=sqlite3_column_text(stmt,0);
		name=names[i];
		g_snprintf(filename,PATH_MAX,"%s%c%s",path,G_DIR_SEPARATOR,name);
		//sqlite3_reset(stmt);
		debug1(DEBUG_ACACHE, "unlinking [%s]", filename);
		pan_unlink(filename);
	}
	//sqlite3_finalize(stmt);
	
	/* remove it from the lookup table */
	if(len>4)
		sqlite3_exec(db,"BEGIN TRANSACTION;",NULL,NULL,NULL);
	
	sq_res=sqlite3_prepare(db,"DELETE FROM cache WHERE id=?;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	
	for(i=0;i<len;i++)
	{
		sqlite3_bind_int64(stmt,1,ids[i]);
		sqlite3_step(stmt);
		sqlite3_reset(stmt);
	}
	sqlite3_finalize(stmt);
	
	if(len>4)
		sqlite3_exec(db,"COMMIT TRANSACTION;",NULL,NULL,NULL);
}

void
acache_expire_messages (const char                * path_key,
                        const char * const     * hashs,
                        int                         hash_qty)
{
	/* sanity clause */
	g_return_if_fail (is_nonempty_string (path_key));
	g_return_if_fail (hash_qty > 0);
	g_return_if_fail (hashs != NULL);
	
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	char *sql;
	int res;
	Article *article;
	GPtrArray *t_art=g_ptr_array_sized_new(hash_qty);
	char **names=g_new0(char*, hash_qty);
	int *refcnts=g_new0(int,hash_qty);
	gint64 *ids=g_new0(gint64,hash_qty);
	const gboolean cache=(strcmp(path_key,ACACHE_DEFAULT_KEY)==0)?TRUE:FALSE;
	int cnt=0;
	
	/* do the work */
	int i;

	sql="SELECT refcnt,size,cache.id,a.id,file FROM cache, cache_paths AS cp, article AS a "
		"WHERE cache.hash=? AND key=? AND keyid=cp.id AND a.hash=cache.hash;";
	res=sqlite3_prepare(db,sql,-1,&stmt,NULL);
	SQL_PREP_VALID(res,db);
	for (i=0; i<hash_qty; ++i)
	{
		sqlite3_bind_blob(stmt,1,hashs[i],16,SQLITE_STATIC);
		sqlite3_bind_text(stmt,2,path_key,-1,SQLITE_STATIC);
		res=sqlite3_step(stmt);
		if(res!=SQLITE_ROW)
		{
			sqlite3_reset(stmt);
			continue;
		}
		refcnts[cnt]=sqlite3_column_int(stmt,0);
		ids[cnt]=sqlite3_column_int64(stmt,2);
		names[cnt]=g_strdup( sqlite3_column_text(stmt,4) );
		if G_LIKELY(cache)
			cache_size-=sqlite3_column_int(stmt,1);
		article=article_get_by_id( sqlite3_column_int64(stmt,3) );
		if(article)
			g_ptr_array_add(t_art,article);
		cnt++;
		sqlite3_reset(stmt);
	}
	sqlite3_finalize(stmt);

	/* notify listeners that this message is going away */
	if(t_art->len)
	pan_base_signals_bodies_removed( pan_base_signals_get(), (const Article**) t_art->pdata, t_art->len);

	acache_remove_entries(db,path_key,(const char**)names,ids,cnt);
	
	g_free(refcnts);
	g_free(ids);
	for(i=0;i<cnt;i++)
		g_free(names[i]);
	g_free(names);
	//articles not refed by get_by_id
	g_ptr_array_free(t_art,TRUE);
}

static int
acache_expire_to_size(double max_megs)
{
	int files_removed = 0;
	const gint64 cache_max = (size_t) max_megs * (1024 * 1024);
	debug_enter("acache_expire_to_size");

	/* sanity clause */
	debug3(DEBUG_ACACHE, "expiring to %lu MB; current size is %lu MB (%ld Bytes free)", 
		(gint32)max_megs, cache_size/(1024*1024), cache_max-cache_size);
		
	/* short circuit */
	if(cache_max>cache_size)
	{
		debug_exit("acache_expire_to_size");
		return 0;
	}
	
	gint64 size_temp=cache_size;
	GArray *ids=g_array_sized_new(FALSE,TRUE,sizeof(gint64),100);
	GPtrArray *names=g_ptr_array_sized_new(100);
	GPtrArray *t_art=g_ptr_array_sized_new(100);
	Article *article;
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	gint64 ti;
	char *sql;
	const char *tmp;
	int res;

	sql="SELECT cache.id,size,file,a.id FROM cache_paths AS cp, cache, article AS a "
		"WHERE cp.key=? AND cache.keyid=cp.id AND cache.refcnt=0 AND a.hash=cache.hash "
		"ORDER BY cache.date;";
	
	res=sqlite3_prepare(db,sql,-1,&stmt,NULL);
	SQL_PREP_VALID(res,db);
	sqlite3_bind_text(stmt,1,ACACHE_DEFAULT_KEY,-1,SQLITE_STATIC);

	while(size_temp>cache_max)
	{
		res=sqlite3_step(stmt);
		if G_UNLIKELY(res!=SQLITE_ROW)
			break;
		ti=sqlite3_column_int64(stmt,0);
		g_array_append_val(ids,ti);
		size_temp-=sqlite3_column_int(stmt,1);
		tmp=sqlite3_column_text(stmt,2);
		g_ptr_array_add(names,g_strdup(tmp));
		article=article_get_by_id( sqlite3_column_int64(stmt,3) );
		if(article)
			g_ptr_array_add(t_art,article);
		sqlite3_reset(stmt);
	}
	sqlite3_finalize(stmt);
	
	pan_base_signals_bodies_removed(pan_base_signals_get(), (const Article**)t_art->pdata, t_art->len);
	
	files_removed=ids->len;
	acache_remove_entries(db,ACACHE_DEFAULT_KEY,(const char**)(names->pdata),(gint64*)(ids->data),ids->len);
	
	int i;
	for(i=0;i<names->len;i++)
		g_free(g_ptr_array_index(names,i));
	g_ptr_array_free(names,TRUE);
	g_array_free(ids,TRUE);
	//articles not refed by get_by_id
	g_ptr_array_free(t_art,TRUE);
	
	/* done */
	debug_exit ("acache_expire_to_size");
	return files_removed;
}

int 
acache_expire(void)
{
	const double desired_size = acache_max_megs * 0.8;
	return acache_expire_to_size(desired_size);
}

// TODO: optimize by specail casing expire_all
int
acache_expire_all(void)
{
	int retval;

	g_message(_("Flushing article cache... "));
	retval = acache_expire_to_size(0.0);
	g_message(_("%d files erased."), retval);
	fflush(NULL);

	return retval;
}

/***
****
****  CHECKIN / CHECKOUT
****
***/

void
acache_checkout(const char               * key,
                 const Article    	  ** articles,
                 int                        article_qty)
{
	debug_enter("acache_checkout");

	/* sanity clause */
	g_return_if_fail(is_nonempty_string(key));
	g_return_if_fail(articles!=NULL);
	g_return_if_fail(article_qty >= 1);

	/* ref the entries */
	const char *sql="UPDATE cache SET refcnt=refcnt+1 WHERE hash=(SELECT hash FROM article WHERE id=?) "
					"AND keyid=(SELECT id FROM cache_paths WHERE key=?) ;";
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	int x, sq_res;
	if(article_qty>4)
		BEGIN_TRANSACTION(db,stmt);
	
	sq_res=sqlite3_prepare(db,sql,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	for(x=0;x<article_qty;x++)
	{
		sqlite3_bind_int64(stmt,1,articles[x]->id);
		sqlite3_bind_text(stmt,2,key,-1,SQLITE_STATIC);
		sqlite3_step(stmt);
		sqlite3_reset(stmt);
	}
	sqlite3_finalize(stmt);
	
	if(article_qty>4)
		END_TRANSACTION(db,stmt);


	debug_exit("acache_checkout");
}

void
acache_checkin(const char               * key,
                 const Article    	  ** articles,
                 int                        article_qty)
{
	debug_enter("acache_checkin");

	/* sanity clause */
	g_return_if_fail(is_nonempty_string(key));
	g_return_if_fail(articles!=NULL);
	g_return_if_fail(article_qty >= 1);

	/* ref the entries */
	const char *sql="UPDATE cache SET refcnt=refcnt-1 WHERE hash=(SELECT hash FROM article WHERE id=?) "
					"AND keyid=(SELECT id FROM cache_paths WHERE key=?) ;";
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	int x, sq_res;
	if(article_qty>4)
		BEGIN_TRANSACTION(db,stmt);
	
	sq_res=sqlite3_prepare(db,sql,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	for(x=0;x<article_qty;x++)
	{
		sqlite3_bind_int64(stmt,1,articles[x]->id);
		sqlite3_bind_text(stmt,2,key,-1,SQLITE_STATIC);
		sqlite3_step(stmt);
		sqlite3_reset(stmt);
	}
	sqlite3_finalize(stmt);
	
	if(article_qty>4)
		END_TRANSACTION(db,stmt);


	debug_exit("acache_checkin");
}

/***
****
****  ADD FILES
****
***/
static unsigned long
get_today(void)
{
	unsigned long mm, yy, dd;
	struct tm tm_struct;
	const time_t now = time (NULL);

	pan_localtime_r (&now, &tm_struct);
	yy = tm_struct.tm_year;
	mm = tm_struct.tm_mon;
	dd = tm_struct.tm_mday;

	return yy * 10000 + mm * 100 + dd;
}

void
acache_set_message (const char        * key,
                    const PString     * message_id,
                    const char        * message,
                    guint               message_len)
{
	/* sanity clause */
	g_return_if_fail (is_nonempty_string(key));
	g_return_if_fail (pstring_is_set(message_id));
	g_return_if_fail(message_id->str[0]!='<');
	g_return_if_fail (is_nonempty_string(message));
	g_return_if_fail (message_len > 0);

	FILE * fp = NULL;
	char filename[PATH_MAX], apath[PATH_MAX];
	sqlite3 *db=get_db();

	/* get the path */
	acache_get_path(db,apath,PATH_MAX,key);
	g_return_if_fail(apath[0]!=0);

	/* find out where to write the message */
	if (acache_get_filename(filename, sizeof(filename), apath, message_id))
		fp = pan_fopen(filename, "wb+");

	if (fp != NULL)
	{
		size_t bytes_written = fwrite(message, sizeof(char), message_len, fp);
		fclose(fp);

		if (bytes_written < message_len)
		{
			/* couldn't save the whole message try makeing space*/
			pan_unlink(filename);
			double x=cache_size*.8;
			acache_expire_to_size(x);
			fp = pan_fopen(filename, "wb+");
			bytes_written = fwrite(message, sizeof(char), message_len, fp);
			fclose(fp);
			
			if (bytes_written < message_len)
			{
				/* couldn't save the whole message */
				char * path = g_path_get_dirname(filename);
				log_add_va(LOG_ERROR, _("Error saving article \"%s\" (is %s full?)"), message_id->str, path);
				g_free(path);
			}
		}
		
		if(bytes_written >= message_len)
		{
			sqlite3_stmt *stmt;
			char *sql="INSERT INTO cache(hash,file,size,date,keyid,refcnt) VALUES(?,?,?,?,?,0);";
			char *tmp=g_path_get_basename(filename);
			int keyid, sq_res;
			char hash[16];

			sq_res=sqlite3_prepare(db,"SELECT id FROM cache_paths WHERE key=?;",-1,&stmt,NULL);
			SQL_PREP_VALID(sq_res,db);
			sqlite3_bind_text(stmt,1,key,-1,SQLITE_STATIC);
			sqlite3_step(stmt);
			keyid=sqlite3_column_int(stmt,0);
			sqlite3_finalize(stmt);
			
			sq_res=sqlite3_prepare(db,sql,-1,&stmt,NULL);
			SQL_PREP_VALID(sq_res,db);
			/* file an entry for this message */
			md5_get_digest(message_id->str, message_id->len, hash);
			sqlite3_bind_blob(stmt,1,hash,16,SQLITE_STATIC);
			sqlite3_bind_int(stmt,5,keyid);
			sqlite3_bind_text(stmt,2,tmp,-1,SQLITE_STATIC);
			sqlite3_bind_int(stmt,3,message_len);
			sqlite3_bind_int(stmt,4,get_today());
			sqlite3_step(stmt);
			sqlite3_finalize(stmt);
			if G_LIKELY(key==ACACHE_DEFAULT_KEY || strcmp(key,ACACHE_DEFAULT_KEY)==0)
				cache_size += message_len;
			g_free(tmp);

			/* notify everyone that this message has been added */
			if(key==ACACHE_DEFAULT_KEY || strcmp(key,ACACHE_DEFAULT_KEY)==0)
			{
				Article *art=NULL;
				sq_res=sqlite3_prepare(db,"SELECT id FROM article WHERE hash=?;",-1,&stmt,NULL);
				SQL_PREP_VALID(sq_res,db);
				sqlite3_bind_blob(stmt,1,hash,16,SQLITE_STATIC);
				if(sqlite3_step(stmt)==SQLITE_ROW)
					art=article_get_by_id( sqlite3_column_int64(stmt,0));
				else
					log_add_va(LOG_ERROR,"Article id not found for msgid=%s",message_id->str);
				sqlite3_finalize(stmt);
				if(art)
				{
					pan_base_signals_body_added(pan_base_signals_get(),art);
					g_object_unref(art);
				}
			}

			/* if the acache is too big, purge the old stuff */
			acache_expire();
		}
	}

}


/***
****
****  GETTING MESSAGES
****
***/

static GMimeStream*
acache_get_message_file_stream (const char * filename)
{
	GMimeStream * retval = NULL;

	/* sanity clause */
	g_return_val_if_fail (filename!=NULL, NULL);

	/* open the file */
	FILE * fp;
	errno = 0;
	fp = pan_fopen(filename, "rb");
	if (!fp)
		log_add_va (LOG_ERROR, _("Error opening file \"%s\": %s"), filename, g_strerror(errno));
	else
	{
		GMimeStream * file_stream = g_mime_stream_file_new (fp);
		retval = g_mime_stream_buffer_new (file_stream, GMIME_STREAM_BUFFER_BLOCK_READ);
		g_object_unref (file_stream);
	}

	return retval;
}

static GMimeStream*
acache_get_message_mem_stream (const char * filename)
{
	GMimeStream * retval = NULL;

	/* sanity clause */
	g_return_val_if_fail (filename!=NULL, NULL);

	/* open the file */
	gsize len = 0;
	char * buf = NULL;
	GError * err = NULL;

	if (g_file_get_contents (filename, &buf, &len, &err)) {
		retval = g_mime_stream_mem_new_with_buffer (buf, len);
		g_free (buf);
	} else {
		log_add_va (LOG_ERROR, _("Error reading file \"%s\": %s"), filename, err->message);
		g_error_free (err);
	}

	return retval;
}

GMimeMessage*
acache_get_message  (const char                * path_key,
                     const Article              ** articles,
                     int                         article_qty)
{
	/* sanity clause */
	g_return_val_if_fail(is_nonempty_string(path_key), NULL);
	g_return_val_if_fail(article_qty>0, NULL);
	g_return_val_if_fail(articles!=NULL, NULL);

	int i;
	int stream_qty;
	GMimeStream ** streams;
	GMimeMessage * retval = NULL;
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	char path[PATH_MAX], fn[PATH_MAX];
	const char *tmp, *sql;
	int res;
	gboolean is_cache;

	if G_LIKELY(path_key==ACACHE_DEFAULT_KEY || g_ascii_strcasecmp(ACACHE_DEFAULT_KEY,path_key)==0)
		is_cache=TRUE;
	else
		is_cache=FALSE;
	
	acache_get_path(db,path,PATH_MAX,path_key);
	
	if G_LIKELY(is_cache)
		sql="SELECT file,cache.id,size,key FROM article AS a , cache_paths AS cp, cache "
			"WHERE a.id=? AND key NOT LIKE 'pan.%' AND cache.hash=a.hash AND cache.keyid=cp.id;";
	else
		sql="SELECT file,cache.id,size FROM article AS a, cache_paths AS cp , cache "
			"WHERE a.id=? AND key=? AND cache.hash=a.hash AND cache.keyid=cp.id;";

	res=sqlite3_prepare(db,sql,-1,&stmt,NULL);
	SQL_PREP_VALID(res,db);
	
	if G_UNLIKELY(!is_cache)
		sqlite3_bind_text(stmt,2,path_key,-1,SQLITE_STATIC);
	
	/* get streams */
	streams = g_newa(GMimeStream*, article_qty);
	for (i=stream_qty=0; i<article_qty; ++i)
	{
		GMimeStream * stream;
		gboolean found;
		found=FALSE;
		sqlite3_bind_int64(stmt,1,articles[i]->id);
		while(sqlite3_step(stmt)==SQLITE_ROW && !found)
		{
			tmp=sqlite3_column_text(stmt,0);
			g_snprintf(fn,PATH_MAX,"%s%c%s",path,G_DIR_SEPARATOR,tmp);
			
			/* file has been deleted so remove it from db
			** and tell listeners */
			if G_UNLIKELY(!pan_file_exists(fn))
			{
				gint64 id=sqlite3_column_int64(stmt,1);
				sqlite3_stmt *s;
				const char *key;
				cache_size-=sqlite3_column_int(stmt,2);
				res=sqlite3_prepare(db,"DELETE FROM cache WHERE id=?;",-1,&s,NULL);
				SQL_PREP_VALID(res,db);
				sqlite3_bind_int64(s,1,id);
				sqlite3_step(s);
				sqlite3_finalize(s);
				
				if G_LIKELY(is_cache)
				{
					key=sqlite3_column_text(stmt,3);
					if (g_ascii_strcasecmp(ACACHE_DEFAULT_KEY,key)==0)
						pan_base_signals_bodies_removed(pan_base_signals_get(),&articles[i],1);
				}
			}
			else
				found=TRUE;
		}
		sqlite3_reset(stmt);
		if G_UNLIKELY(!found)
			break;
		
		stream = article_qty>2
			? acache_get_message_file_stream(fn)
			: acache_get_message_mem_stream(fn);
		if G_UNLIKELY(!stream)
			break;
		
		streams[stream_qty++] = stream;
	}
	sqlite3_finalize(stmt);
	
	/* build the message */
	if (stream_qty == article_qty)
		retval = pan_g_mime_parser_construct_message(streams, stream_qty);
	
	/* cleanup */
	for (i=0; i<stream_qty; ++i)
		g_object_unref(streams[i]);
	
	return retval;
}


gboolean
acache_has_message(const char * path_key, gint64 aid)
{
	/* sanity clause */
	g_return_val_if_fail(is_nonempty_string(path_key), FALSE);
	g_return_val_if_fail(aid!=0, FALSE);

	gboolean retval = FALSE;
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	int res;
	char *sql;
	gboolean is_cache;

	if G_LIKELY(path_key==ACACHE_DEFAULT_KEY || g_ascii_strcasecmp(ACACHE_DEFAULT_KEY,path_key)==0)
	{
		is_cache=TRUE;
		path_key="pan.%";
	}
	else
		is_cache=FALSE;

	/* lookup the message */
	if G_LIKELY(is_cache)
		sql="SELECT count(*) FROM article AS a, cache_paths AS cp, cache "
			"WHERE a.id=? AND key NOT LIKE ? AND cache.hash=a.hash AND cache.keyid=cp.id;";
	else
		sql="SELECT count(*) FROM article AS a, cache_paths AS cp , cache "
			"WHERE a.id=? AND key=? AND cache.hash=a.hash AND cache.keyid=cp.id;";
	res=sqlite3_prepare(db,sql,-1,&stmt,NULL);
	SQL_PREP_VALID(res,db);
	sqlite3_bind_int64(stmt,1,aid);
	sqlite3_bind_text(stmt,2,path_key,-1,SQLITE_STATIC);
	sqlite3_step(stmt);
	res=sqlite3_column_int(stmt,0);
	sqlite3_finalize(stmt);
	
	retval=(res==0)?FALSE:TRUE;
	return retval;
}

/***
****
****  PATHS
****
***/

static void acache_get_path(sqlite3 *db, char *buf, int buflen, const char *key)
{
	sqlite3_stmt *stmt;
	const char *t1;
	int res;
	
	buf[0]='\0';
	res=sqlite3_prepare(db,"SELECT path FROM cache_paths WHERE key=?;",-1,&stmt,NULL);
	SQL_PREP_VALID(res,db);
	sqlite3_bind_text(stmt,1,key,-1,SQLITE_STATIC);
	res=sqlite3_step(stmt);
	if G_LIKELY(res==SQLITE_ROW)
	{
		t1=sqlite3_column_text(stmt,0);
		if (t1)
			g_snprintf(buf,buflen,"%s",t1);
	}
	sqlite3_finalize(stmt);
}

static void
acache_foreach(const char *path,
						  AcachePathForeachFunc func,
                          gpointer user_data)
{
	GMimeStream * stream = acache_get_message_mem_stream (path);
	GMimeMessage * message = pan_g_mime_parser_construct_message (&stream, 1);

	/* call the user's function */
	(func)(message, user_data);

	/* cleanup */
	g_object_unref (message);
	g_object_unref (stream);
}

void
acache_path_foreach(const char             * key,
                     AcachePathForeachFunc    func,
                     gpointer                 user_data)
{
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	char path[PATH_MAX], fn[PATH_MAX];
	int sq_res;
	const char *t1;
	debug_enter("acache_path_foreach");

	/* get the scratch directory path */
	debug1 (DEBUG_ACACHE, "Calling acache_path_foreach for key \"%s\"", key);
	
	acache_get_path(db,path,PATH_MAX,key);
	
	sq_res=sqlite3_prepare(db,"SELECT file FROM cache WHERE key=?;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_text(stmt,1,key,-1,SQLITE_STATIC);
	while(sqlite3_step(stmt)!=SQLITE_DONE)
	{
		t1=sqlite3_column_text(stmt,0);
		g_snprintf(fn,PATH_MAX,"%s%c%s",path,G_DIR_SEPARATOR,t1);
		acache_foreach(fn,func,user_data);
	}
	sqlite3_finalize(stmt);

	debug_exit("acache_path_foreach");
}
// FIXME: need remove_folder
void
acache_add_folder(const char * key)
{
	char * folder_base = acache_get_folder_base();
	char * path = g_build_filename(folder_base, key, NULL);
	sqlite3 *db=get_db();
	
	acache_add_path(db, key, path);

	g_free(path);
	g_free(folder_base);
}

static void
acache_add_path(sqlite3 *db, const char * key, const char * path)
{
	int qty = 0;
	gint64 size=0;
	GDir * dir;
	GError * err;
	sqlite3_stmt *stmt;
	int sq_res;
	debug_enter ("acache_add_path");

	/* sanity clause */
	g_return_if_fail (is_nonempty_string(key));
	g_return_if_fail (is_nonempty_string(path));
	g_return_if_fail (pan_file_ensure_path_exists(path));

	sq_res=sqlite3_prepare(db,"DELETE FROM cache WHERE keyid=(SELECT id FROM cache_paths WHERE key=?);",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_text(stmt,1,key,-1,SQLITE_STATIC);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);

	
	/* make the path entry */
	debug2 (DEBUG_ACACHE, "Adding path: key \"%s\", path \"%s\"", key, path);

	/* walk through the directory */
	err = NULL;
	dir = g_dir_open (path, 0, &err);
	if (err != NULL)
	{
		log_add_va (LOG_ERROR, _("Error opening directory \"%s\": %s)"), path, err->message);
		g_error_free (err);
	}
	else
	{
		char *sql;
		const char * fname;
		char str[2048];
		char filename[PATH_MAX];
		int keyid;

		sq_res=sqlite3_prepare(db,"INSERT OR IGNORE INTO cache_paths(key,path) VALUES(?,?);",-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		sqlite3_bind_text(stmt,1,key,-1,SQLITE_STATIC);
		sqlite3_bind_text(stmt,2,path,-1,SQLITE_STATIC);
		sqlite3_step(stmt);
		sqlite3_finalize(stmt);
		sq_res=sqlite3_prepare(db,"SELECT id FROM cache_paths WHERE key=?;",-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		sqlite3_bind_text(stmt,1,key,-1,SQLITE_STATIC);
		sqlite3_step(stmt);
		keyid=sqlite3_column_int(stmt,0);
		sqlite3_finalize(stmt);

		sqlite3_exec(db,"BEGIN TRANSACTION;",NULL,NULL,NULL);
		
		sql="INSERT INTO cache(hash,file,size,date,keyid,refcnt) VALUES(?,?,?,?,?,0);";
		sq_res=sqlite3_prepare(db,sql,-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		
		while ((fname = g_dir_read_name (dir)))
		{
			struct stat stat_p;
			char hash[16];

			g_snprintf (filename, sizeof(filename), "%s%c%s", path, G_DIR_SEPARATOR, fname);
			if (stat (filename, &stat_p) == 0)
			{
				int len = get_message_id_from_file (str, sizeof(str), filename);
				md5_get_digest(str,len,hash);

				if (len != 0)
				{
					sqlite3_bind_blob(stmt,1,hash,16,SQLITE_STATIC);
					sqlite3_bind_text(stmt,2,fname,-1,SQLITE_STATIC);
					sqlite3_bind_int(stmt,3,stat_p.st_size);
					sqlite3_bind_int(stmt,4,stat_p.st_mtime);
					sqlite3_bind_int(stmt,5,keyid);
					sqlite3_step(stmt);
					sqlite3_reset(stmt);
					size+=stat_p.st_size;
					++qty;
				}
			}
		}
		sqlite3_finalize(stmt);
		sqlite3_exec(db,"COMMIT TRANSACTION;",NULL,NULL,NULL);
		
		g_dir_close(dir);
	}
	
	
	log_add_va (LOG_INFO, _("Directory \"%s\" contains %.1f MB in %d files"),
		path, (double)size/(1024*1024), qty);

	debug_exit ("acache_add_path");
}

