/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#include <glib.h>
#include <glib/gi18n-lib.h>

#include <pan/base/db.h>
#include <pan/base/article-thread.h>
#include <pan/base/article.h>
#include <pan/base/debug.h>


/**
 * Skip the "Re: " part of a subject header, if any
 * @param subject
 * @return the non-"Re:" portion of the subject header
 */
#define skip_reply_leader(a) \
	(((a!=NULL) && \
	  (a[0]=='R' || a[0]=='r') && \
	  (a[1]=='E' || a[1]=='e') && \
	  (a[2]==':') && \
	  (a[3]==' ')) ? a+4 : a)

/**
 * normalize_subject:
 * @subject: PString* UTF-8 subject %NULL terminated.
 * @pre_subj: gchar **
 * @norm_subj: gchar**.
 * @post_subj: gchar**
 *
 * All reply leaders are placed in @pre_subj. @norm_subj contains the text from 
 * @pre_subj up to the LAST multi-part indicator. @post_subj contains the text 
 * starting with the last multi-part indicator to end of @subject.
 * All subject components will be returned.
 *
 * Returns: gboolean - TRUE if succesful, FALSE on error
 */
gboolean normalize_subject(const PString *subject, gchar **pre_subj, gchar** norm_subj, gchar** post_subj)
{
	g_return_val_if_fail(subject!=NULL,FALSE);
	g_return_val_if_fail(pstring_is_set(subject),FALSE);
	g_return_val_if_fail(pre_subj!=NULL,FALSE);
	g_return_val_if_fail(norm_subj!=NULL,FALSE);
	g_return_val_if_fail(post_subj!=NULL,FALSE);
	const gchar *subj;
	subj=subject->str;
	const gchar *old=NULL,*temp;
	gboolean multipart=FALSE;
	debug_enter("normalize_subject");
	
	// rare short-circuit
	if G_UNLIKELY(!*subj)
	{
		*pre_subj=g_strdup("");
		*norm_subj=g_strdup("");
		*post_subj=g_strdup("");
		debug_exit("normalize_subject");
		return TRUE;
	}
	
	*post_subj=*pre_subj=NULL;
	// strip ALL reply leaders
	temp=subj;
	while(temp!=old)
	{
		old=temp;
		temp=skip_reply_leader(temp);
	}
	if(temp!=subj)
		*pre_subj=g_strndup(subj,temp-subj);
	else
		*pre_subj=g_strdup("");
	
	old=temp;
	
	gchar *free_me=g_utf8_strreverse(temp,-1);
	temp=free_me;
	// NOTE: this code doens not allow whitespace in the multipart designator
	// ie. [xx/xx] not [ xx / xx ]
	while(!multipart && *temp)
	{
		while(*temp && *temp!=']' && *temp!=')')
			temp=g_utf8_next_char(temp);
		// since we can't interpret foreign numbers this will do
		if(!*temp)
			break; // end of string
		++temp;
		while(*temp && g_ascii_isdigit(*temp))
			++temp;
		if(!*temp)
			break; // end of string
		if(*temp!='/' && *temp!='|')
			continue; // not multipart yet
		++temp;
		while(*temp && g_ascii_isdigit(*temp))
			++temp;
		if(!*temp)
			break;  // end of string
		if(*temp!='[' && *temp!='(')
			continue; // not multipart yet
		multipart=TRUE;
	}
	int post_len=temp-free_me+1;
	g_free(free_me);

	temp=subj+subject->len-post_len;
	if(multipart)
	{
		*post_subj=g_strndup(temp,post_len);
		*norm_subj=g_strndup(old,temp-old);
	}else
	{
		*post_subj=g_strdup("");
		*norm_subj=g_strndup(old,post_len);
	}
	
	debug_exit("normalize_subject");
	return TRUE;
}
//
// Threading code here
//

/**
 * Thread the articles specified in list
 */

/**
 * thread_group:
 * @group_id:
 *
 * Threads all the articles in the specified group.
 */
void thread_group(gint group_id)
{
	debug_enter("thread_group");
	
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt, *stmt2, *stmt3, *updt, *trans;
	int sq_res;
	gint64 a,p;
	gint32 norm_subj_id,date,parts;
	const char *pre_subj;
	
	// create temporary result table
	// required due to db locking
	sq_res=sqlite3_prepare(db,"CREATE TEMP TABLE thread_temp(aid INT, pid INT);",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	
	// unthread group
	sq_res=sqlite3_prepare(db,"UPDATE group_article SET parent=null WHERE group_id=?;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt,1,group_id);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);

	// add to temp table
	sq_res=sqlite3_prepare(db,"INSERT INTO thread_temp(pid,aid) VALUES(?,?);",
						-1,&updt,NULL);
	SQL_PREP_VALID(sq_res,db);
	
	// thread articles except part>2 of multi-part
	sq_res=sqlite3_prepare(db,"SELECT article_id,norm_subj,date,parts,subject "
							"FROM group_article, article , norm_subject AS ns "
							"WHERE group_id=? AND article.id=article_id AND ns.id=pre_subj;"
							,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt,1,group_id);
	// first reference in references & group
	sq_res=sqlite3_prepare(db,"SELECT article.id FROM refs,article,group_article "
							"WHERE refs.article_id=? AND hash=reference "
							"AND group_id=? AND group_article.article_id=article.id "
							"ORDER BY refs.id DESC;",  //if threading is wrong make this DESC
							-1,&stmt2,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt2,2,group_id);
	/* for threading by subject
	 * look for article in group with same normalized subject, fewer REs,
	 * younger than the reply and return them in descending order by date
	 */
	sq_res=sqlite3_prepare(db,"SELECT article.id FROM norm_subject AS ns, article,group_article "
							"WHERE subject=?1 AND norm_subj=?2 AND pre_subj=ns.id AND parts=?5 AND "
							"date<?3 AND group_id=?4 AND article_id=article.id "
							"ORDER BY date DESC;",-1,&stmt3,NULL);	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt3,4,group_id);
	
	//thread by references or subject
 	while(sqlite3_step(stmt)==SQLITE_ROW)
	{
		p=0;
		a=sqlite3_column_int64(stmt,0);
		sqlite3_bind_int64(stmt2,1,a);
		
		if (sqlite3_step(stmt2)==SQLITE_ROW)
		{
			p=sqlite3_column_int64(stmt2,0);
		}else //no refs or refed articles not in group
		{
 			norm_subj_id=sqlite3_column_int(stmt,1);
			date=sqlite3_column_int(stmt,2);
			parts=sqlite3_column_int(stmt,3);
			pre_subj=sqlite3_column_text(stmt,4);
			
			sqlite3_bind_int(stmt3,2,norm_subj_id);
			sqlite3_bind_text(stmt3,1,skip_reply_leader(pre_subj),-1,SQLITE_STATIC);
			sqlite3_bind_int(stmt3,3,date);
			sqlite3_bind_int(stmt3,5,parts);
			if(sqlite3_step(stmt3)==SQLITE_ROW)
				p=sqlite3_column_int64(stmt3,0);
			sqlite3_reset(stmt3);
 		}
		sqlite3_reset(stmt2);
		
		if(p!=0)
		{
			sqlite3_bind_int64(updt,1,p);
			sqlite3_bind_int64(updt,2,a);
			sqlite3_step(updt);
			sqlite3_reset(updt);
		}
	}
	
	sqlite3_finalize(stmt3);
	sqlite3_finalize(stmt2);
	sqlite3_finalize(stmt);
	sqlite3_finalize(updt);
	
	
	BEGIN_TRANSACTION(db,trans);

	// update parent info
	sq_res=sqlite3_prepare(db,"SELECT aid,pid FROM thread_temp;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	
	sq_res=sqlite3_prepare(db,"UPDATE group_article SET parent=? WHERE article_id=? AND group_id=? ;",-1,&updt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(updt,3,group_id);
	
	while(sqlite3_step(stmt)==SQLITE_ROW)
	{
		a=sqlite3_column_int64(stmt,0);
		p=sqlite3_column_int64(stmt,1);
		sqlite3_bind_int64(updt,1,p);
		sqlite3_bind_int64(updt,2,a);
		sqlite3_step(updt);
		sqlite3_reset(updt);
	}
	sqlite3_finalize(updt);
	sqlite3_finalize(stmt);
	
	END_TRANSACTION(db,trans);
	
	sq_res=sqlite3_prepare(db,"DROP TABLE thread_temp;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	
	debug_exit("thread_group");
}

