/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2004  K. Haley <haleykd@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <glib.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/base-prefs.h>
#include <pan/base/pan-config.h>
#include <pan/base/pstring.h>
#include <gmime/gmime.h>
#include <gmime/gmime-utils.h>
#include <gmime/md5-utils.h>
#include <pan/base/db.h>
#include <pan/base/article-thread.h>
#include <pan/base/db-add.h>
#include <pan/base/debug.h>
#include <pan/base/import.h>

static const char *DB_NAME_MAIN="pan.sqlite3";
static const char *DB_NAME_AUX="pan_article.sqlite3";
static const char *DB_NAME_CACHE="pan_cache.sqlite3";

static int busy_handler(void *x, int y)
{
	static gulong t=G_USEC_PER_SEC/100;
	g_usleep(t);
	return 1;
}

static void build_cache_db(const char *fn)
{
	debug_enter("build_cache_db");
	sqlite3 *db;
	int sq_res;
	char *errmsg;
	
	const char *mk_ctbl="CREATE TABLE cache_paths("
							"key TEXT UNIQUE NOT NULL, "
							"path TEXT NOT NULL, "
							"id INTEGER PRIMARY KEY AUTOINCREMENT);"
						"CREATE TABLE cache("
							"file TEXT NOT NULL, "
							"size INT NOT NULL, "
							"date INT NOT NULL, "
							"refcnt INT NOT NULL, "
							"keyid INT NOT NULL REFERENCES cache_paths(id), "
							"hash BLOB NOT NULL, "
							"id INTEGER PRIMARY KEY "
							");"
						"CREATE UNIQUE INDEX c_mk ON cache(hash,keyid);";
					
	if(sqlite3_open(fn,&db)!=SQLITE_OK)
		g_print("build_cache_db %s\nerror: %s\n",fn,sqlite3_errmsg(db));
	/*make table*/
	sq_res=sqlite3_exec(db,mk_ctbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	sqlite3_close(db);
	debug_exit("build_cache_db");
}

static void build_aux_db(const char *fn)
{
	debug_enter("build_aux_db");
	sqlite3 *db;
	int sq_res;
	char *errmsg,*t;
	
	if(sqlite3_open(fn,&db)!=SQLITE_OK)
		g_error("build_aux_db error: %s\n",sqlite3_errmsg(db));
	
	const char *mk_ntbl="CREATE TABLE norm_subject("
							"subject TEXT NOT NULL UNIQUE, "
							"id INTEGER PRIMARY KEY);";
	sq_res=sqlite3_exec(db,mk_ntbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	const char *mk_authtable="CREATE TABLE authors("
								"real TEXT NOT NULL, "
								"address TEXT NOT NULL, "
								"id INTEGER PRIMARY KEY);"
							"CREATE UNIQUE INDEX auth_idx ON authors(real,address);";
	/*make table*/
	sq_res=sqlite3_exec(db,mk_authtable,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);

	const char *mk_atbl="CREATE TABLE article("
							"date INT NOT NULL, "
							"lines INT NOT NULL DEFAULT 0, "
							"author INT NOT NULL REFERENCES authors(id), "
							"read INT NOT NULL DEFAULT 0, "
							"new INT NOT NULL DEFAULT 1, "
							"byte_qty INT NOT NULL DEFAULT 0, "
							"flagged INT NOT NULL DEFAULT 0, "
							"parts INT NOT NULL DEFAULT 0, "
							"part INT NOT NULL DEFAULT 0, "
							"refcnt INT NOT NULL DEFAULT 0, "
							"keep INT NOT NULL DEFAULT 0, "
							"base_id INT, "
							"pre_subj INT NOT NULL REFERENCES norm_subject(id), "
							"norm_subj INT NOT NULL REFERENCES norm_subject(id), "
							"post_subj TEXT NOT NULL DEFAULT '', "
							"hash BLOB UNIQUE, "
							"message_id TEXT NOT NULL, "
							"id INTEGER PRIMARY KEY);"
						"CREATE INDEX a_ref ON article(refcnt);"
						"CREATE INDEX a_thrd ON article(norm_subj,pre_subj,date);"
						"CREATE INDEX a_bin ON article(base_id,part);"
						"CREATE TRIGGER a_t_ins AFTER INSERT ON article "
							"FOR EACH ROW WHEN NEW.base_id NOTNULL BEGIN "
								"UPDATE article SET refcnt=refcnt+1, lines=lines+NEW.lines WHERE id=NEW.base_id;"
							"END;"
						"CREATE TRIGGER a_t_del AFTER DELETE ON article "
							"FOR EACH ROW WHEN OLD.base_id NOTNULL BEGIN "
								"UPDATE article SET refcnt=refcnt-1, lines=lines-OLD.lines WHERE id=OLD.base_id;"
								"DELETE FROM refs WHERE article_id=OLD.id;"
								"DELETE FROM article_header WHERE article_id=OLD.id;"
							"END;"
						"CREATE TRIGGER a_t_del2 AFTER DELETE ON article "
							"FOR EACH ROW WHEN OLD.base_id ISNULL BEGIN "
								"DELETE FROM refs WHERE article_id=OLD.id;"
								"DELETE FROM article_header WHERE article_id=OLD.id;"
							"END;";
	/*make table*/
	sq_res=sqlite3_exec(db,mk_atbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	const char *mk_rtbl="CREATE TABLE refs("
							"article_id INT NOT NULL REFERENCES article(id), "
							"reference BLOB NOT NULL, "
							"id INTEGER PRIMARY KEY AUTOINCREMENT);"
						"CREATE INDEX ref_aid ON refs(article_id);";
	/*make table*/
	sq_res=sqlite3_exec(db,mk_rtbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	const char *mk_gatbl="CREATE TABLE group_article("
							"article_id INT NOT NULL REFERENCES article(id), "
							"group_id INT NOT NULL REFERENCES groups(id), "
							"parent INT REFERENCES article(id), "
							"score INT NOT NULL DEFAULT 0, "
							"score_date INT NOT NULL DEFAULT 0, "
							"id INTEGER PRIMARY KEY);"
						"CREATE UNIQUE INDEX ga_ga "
							"ON group_article(group_id, article_id);"
						"CREATE INDEX ga_parent ON group_article(parent);"
						"CREATE TRIGGER ga_ins AFTER INSERT ON group_article "
							"FOR EACH ROW BEGIN "
								"UPDATE groups SET qty=qty+1, "
									"qty_read=qty_read+(SELECT read FROM article WHERE id=NEW.article_id) "
									"WHERE id=NEW.group_id;"
							"END;"
						"CREATE TRIGGER ga_del AFTER DELETE ON group_article "
							"FOR EACH ROW BEGIN "
								"UPDATE groups SET qty=qty-1, "
									"qty_read=qty_read-(SELECT read FROM article WHERE id=OLD.article_id) "
									"WHERE id=OLD.group_id;"
								"UPDATE group_article SET parent=OLD.parent "
									"WHERE group_id=OLD.group_id AND parent=OLD.article_id;"
							"END;";
	/*make table*/
	sq_res=sqlite3_exec(db,mk_gatbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	const char *mk_gsatbl="CREATE TABLE g_s_a("
							"group_server_id INT NOT NULL REFERENCES group_server(id), "
							"article_id INT NOT NULL REFERENCES article(id), "
							"uid INT NOT NULL, "
							"id INTEGER PRIMARY KEY);"
						"CREATE INDEX gsa_gsid ON g_s_a(group_server_id);"
						"CREATE INDEX gsa_aid ON g_s_a(article_id);"
						"CREATE TRIGGER gsa_del_trig AFTER DELETE ON g_s_a "
							"FOR EACH ROW BEGIN "
								"UPDATE article SET refcnt=refcnt-1 WHERE id=OLD.article_id;"
							"END;"
						"CREATE TRIGGER gsa_add_trig AFTER INSERT ON g_s_a "
							"FOR EACH ROW BEGIN "
								"UPDATE article SET refcnt=refcnt+1 WHERE id=NEW.article_id;"
							"END;";
	/*make table*/
	sq_res=sqlite3_exec(db,mk_gsatbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	const char *sql="CREATE TABLE headers("
						"hash BLOB NOT NULL, "
						"header TEXT NOT NULL, "
						"htext TEXT NOT NULL, "
						"id INTEGER PRIMARY KEY);"
					"CREATE UNIQUE INDEX hdr_txt_idx ON headers( header,hash);";
	sq_res=sqlite3_exec(db,sql,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	sql="CREATE TABLE article_header("
			"hid INT NOT NULL REFERENCES headers(id), "
			"aid INT NOT NULL REFERENCES article(id), "
			"server_id INT REFERENCES server(id), "
			"id INTEGER PRIMARY KEY);"
		"CREATE INDEX ah_as ON article_header(aid,server_id);";
	sq_res=sqlite3_exec(db,sql,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	sqlite3_close(db);
	debug_exit("build_aux_db");
}

static void build_main_db(const char *fn)
{
	debug_enter("build_main_db");
	sqlite3 *db;
	int sq_res;
	char *errmsg;
	
	if(sqlite3_open(fn,&db)!=SQLITE_OK)
		g_print("build_main_db error: %s\n",sqlite3_errmsg(db));
	
	const char *mk_ptbl="CREATE TABLE pan("
							"key TEXT UNIQUE, "
							"value, "
							"id INTEGER PRIMARY KEY);"
						"INSERT INTO pan(key,value) VALUES('version',2);";
	/*make table*/
	sq_res=sqlite3_exec(db,mk_ptbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	const char *mk_stbl="CREATE TABLE server("
							"name TEXT UNIQUE, "
							"hostname TEXT NOT NULL, "
							"port INTEGER NOT NULL DEFAULT 25, "
							"username TEXT, "
							"password TEXT, "
							"rank INTEGER NOT NULL DEFAULT 0, "
							"max_conn INTEGER NOT NULL DEFAULT 2, "
							"idle_timeout INTEGER NOT NULL DEFAULT 0, "
							"need_auth INT NOT NULL DEFAULT 0, "
							"newsrc INT NOT NULL DEFAULT 0, "
							"last_gl_update INTEGER NOT NULL DEFAULT 0, "
							"newsrc_fn TEXT, "
							"id INTEGER PRIMARY KEY);";
	
	/*make table*/
	sq_res=sqlite3_exec(db,mk_stbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	const char *mk_gtbl="CREATE TABLE groups("
							"name TEXT UNIQUE NOT NULL, "
							"subscribed INT NOT NULL DEFAULT 0, "
							"folder INT NOT NULL DEFAULT 0, "
							"qty INT NOT NULL DEFAULT 0, "
							"qty_read INT NOT NULL DEFAULT 0, "
							"filter_name TEXT DEFAULT '', "
							"filter_show INT NOT NULL DEFAULT 0, "
							"filter_bits INT NOT NULL DEFAULT 0, "
							"sort INT NOT NULL DEFAULT 0, "
							"sort_old INT NOT NULL DEFAULT 0, "
							"new INT NOT NULL DEFAULT 1, "
							"loaded_since_fetch INT NOT NULL DEFAULT 1, "
							"description TEXT DEFAULT '', "
							"identity TEXT DEFAULT '', "
							"download_dir TEXT DEFAULT '', "
							"charset TEXT DEFAULT '', "
							"id INTEGER PRIMARY KEY);";
						"CREATE INDEX g_sub ON groups(subscribed);"
						"CREATE INDEX g_new ON groups(new);"
						"CREATE INDEX g_folder ON groups(folder);"
						"CREATE INDEX g_name ON groups(name);";
	/*make table*/
	sq_res=sqlite3_exec(db,mk_gtbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	const char *mk_gstbl="CREATE TABLE group_server("
							"group_id INT NOT NULL REFERENCES groups(id), "
							"server_id INT  REFERENCES server(id), "
							"article_min INT NOT NULL DEFAULT 0, "
							"article_max INT NOT NULL DEFAULT 0, "
							"article_max_old INT NOT NULL DEFAULT 0, "
							"permision TEXT NOT NULL, "
							"id INTEGER PRIMARY KEY);"
						"CREATE INDEX gs_gid ON group_server(group_id);"
						// QUESTION: only used during server delete?
						//"CREATE INDEX gs_sid ON group_server(server_id);"
						;
	/*make table*/
	sq_res=sqlite3_exec(db,mk_gstbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	const char *mk_vtbl="CREATE TABLE vserver(name TEXT UNIQUE, id INTEGER PRIMARY KEY);";
	/*make table*/
	sq_res=sqlite3_exec(db,mk_vtbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	const char *mk_vgtbl="CREATE TABLE vserver_group("
							"group_id INT NOT NULL REFERENCES groups(id), "
							"vserver INT NOT NULL  REFERENCES vserver(id), "
							"id INTEGER PRIMARY KEY);"
						"CREATE INDEX vg_vs ON vserver_group(vserver);"
						"CREATE TRIGGER vg_ins AFTER INSERT ON vserver_group "
							"FOR EACH ROW BEGIN "
								"UPDATE groups SET subscribed=subscribed+1 WHERE id=new.group_id;"
							"END;"
						"CREATE TRIGGER vg_del AFTER DELETE ON vserver_group "
							"FOR EACH ROW BEGIN "
								"UPDATE groups SET subscribed=subscribed-1 WHERE id=old.group_id;"
							"END;";
;
	/*make table*/
	sq_res=sqlite3_exec(db,mk_vgtbl,NULL,NULL,&errmsg);
	SQL_EXEC_VALID(sq_res,errmsg);
	
	sqlite3_close(db);
	debug_exit("build_main_db");
}

static void migrate_folders(void);
static void thread_groups(void);

static void update_v1_v2(sqlite3 *db)
{
	const char *new_sql="DELETE TRIGGER ga_ins;"
						"DELETE TRIGGER ga_del;"
						"UPDATE pan SET value=2 WHERE key='version';"
						"CREATE TRIGGER ga_ins AFTER INSERT ON group_article "
							"FOR EACH ROW BEGIN "
								"UPDATE groups SET qty=qty+1, "
									"qty_read=qty_read+(SELECT read FROM article WHERE id=NEW.article_id) "
									"WHERE id=NEW.group_id;"
							"END;"
						"CREATE TRIGGER ga_del AFTER DELETE ON group_article "
							"FOR EACH ROW BEGIN "
								"UPDATE groups SET qty=qty-1, "
									"qty_read=qty_read-(SELECT read FROM article WHERE id=OLD.article_id) "
									"WHERE id=OLD.group_id;"
								"UPDATE group_article SET parent=OLD.parent "
									"WHERE group_id=OLD.group_id AND parent=OLD.article_id;"
							"END;";

}

gboolean build_db(void)
{
	debug_enter("build_db");
	const char *dd=get_data_dir();
	char *main_db_fn=g_build_filename(dd,DB_NAME_MAIN,NULL);
	char *cache_db_fn=g_build_filename(dd,DB_NAME_CACHE,NULL);
	gboolean ret=FALSE;
	gboolean main=pan_file_exists(main_db_fn);
	gboolean cache=pan_file_exists(cache_db_fn);
	
	if(!cache)
		build_cache_db(cache_db_fn);
	
	if(!main)
	{
		build_main_db(main_db_fn);
		build_aux_db(main_db_fn);
		migrate_folders();
		ret=TRUE;
	}
	
	g_free(main_db_fn);
	g_free(cache_db_fn);
	if(!cache)
		acache_build_cache();
	
	{
		sqlite3 *db=get_db();
		sqlite3_stmt *stmt;
		int sq_res;
		
		sq_res=sqlite3_prepare(db,"SELECT value FROM pan WHERE key='version';",-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		sqlite3_step(stmt);
		sq_res=sqlite3_column_int(stmt,0);
		sqlite3_finalize(stmt);
		if(sq_res!=2)
			update_v1_v2(db);
	}
	
	if(ret)
	{
		import_db();
		thread_groups();
	}
	
	debug_exit("build_db");
	return ret;
}

static sqlite3 *gdb=NULL;

void close_db(void)
{
	debug_enter("close_db");
	int res;
	
	res=sqlite3_close(gdb);
	if G_UNLIKELY(res==SQLITE_ERROR)
		g_warning("Error closing database. DB either closed or invalid.");
	if G_UNLIKELY(res==SQLITE_BUSY)
		g_warning("Error closing database, not all prepared statements were finalized.");
	gdb=NULL;
	debug_exit("close_db");
}

sqlite3 * get_db(void)
{
	static char *sql=NULL,*main_db_fn;
	
	if G_LIKELY(gdb)
		return gdb;
	
	if(!sql)
	{
		const char *dd=get_data_dir();
		main_db_fn=g_build_filename(dd,DB_NAME_MAIN,NULL);
		char *aux_db_fn=g_build_filename(dd,DB_NAME_AUX,NULL);
		char *cache_db_fn=g_build_filename(dd,DB_NAME_CACHE,NULL);
		/* sql=sqlite3_mprintf("ATTACH '%q' AS articledb;ATTACH '%q' AS cachedb;",
			aux_db_fn,cache_db_fn); */
		sql=sqlite3_mprintf("ATTACH '%q' AS cachedb;",cache_db_fn);
		g_free(aux_db_fn);
		g_free(cache_db_fn);
	}
	
	if(!gdb)
	{
		sqlite3_open(main_db_fn,&gdb);
		char *msg=NULL;
		int res=sqlite3_exec(gdb,sql,NULL,NULL,&msg);
		SQL_EXEC_VALID(res,msg);
		sqlite3_busy_handler(gdb,busy_handler,NULL);
	}
	return gdb;
}

void clean_db(void)
{
	debug_enter("clean_db");
	const char *sql;
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt,*stmt2;
	gint32 gid,qty,qty_r,i;
	gint sqres;
	
	BEGIN_TRANSACTION(db,stmt);

	sql="DELETE FROM article WHERE refcnt=0;";
	sqres=sqlite3_prepare(db,sql,-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	sql="DELETE FROM group_article WHERE article_id IN (SELECT id FROM article WHERE hash ISNULL AND refcnt=0);";
	sqres=sqlite3_prepare(db,sql,-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	// do this twice due to multi-part
	sql="DELETE FROM article WHERE refcnt=0;";
	sqres=sqlite3_prepare(db,sql,-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);

	END_TRANSACTION(db,stmt);
	
	sqres=sqlite3_prepare(db,"CREATE TEMP TABLE taid AS "
								"SELECT UNIQUE author FROM article "
								"ORDER BY author;",-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	
	sqres=sqlite3_prepare(db,"CREATE TEMP TABLE tnsid(id INTEGER PRIMARY KEY);",-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	sqres=sqlite3_prepare(db,"INSERT INTO tnsid(id) SELECT UNIQUE norm_subj FROM article;",-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	sqres=sqlite3_prepare(db,"INSERT OR IGNORE INTO tnsid(id) SELECT UNIQUE pre_subj FROM article;",-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	
	
	BEGIN_TRANSACTION(db,stmt);
	
	sqres=sqlite3_prepare(db,"DELETE FROM authors WHERE id NOT IN (SELECT id FROM taid);",-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	
	sqres=sqlite3_prepare(db,"DELETE FROM norm_subject WHERE id NOT IN (SELECT id FROM tnsid);",-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	
	sqres=sqlite3_prepare(db,"DELETE FROM headers WHERE id NOT IN (SELECT UNIQUE hid FROM article_header);",-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	
	END_TRANSACTION(db,stmt);
	
	sqres=sqlite3_prepare(db,"DROP TABLE taid;",-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	sqres=sqlite3_prepare(db,"DROP TABLE tnsid;",-1,&stmt,NULL);
	SQL_PREP_VALID(sqres,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);

	debug_exit("clean_db");
}


typedef struct
{
	char *path;
	char *name;
} fldr_s;

static void migrate_folders(void)
{
	debug_enter("migrate_folders");

	sqlite3 *db=get_db();
	sqlite3_stmt *stmt,*stmt_2;
	GDir *dir;
	const char *name;
	char *fn,*folders,*fn_utf8;
	GList *dirs=NULL;
	fldr_s *fldr;
	gsize rd,wr;
	gint sq_res;
	char *errmsg;
	
	/* get list of folders */
	const char *sql_1="INSERT OR IGNORE INTO groups(name, description, folder, subscribed, new) "
						"VALUES(?,'',1,0,0);";
	BEGIN_TRANSACTION(db,stmt);
	sq_res=sqlite3_prepare(db,sql_1,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	folders=g_build_filename(get_data_dir(),"messages","folders",NULL);
	dir=g_dir_open(folders,0,NULL);
	while((name=g_dir_read_name(dir)))
	{
		// FIXME: is this correct for glib 2.6+
#ifdef G_OS_WIN32
		fn_utf8=g_strdup(name);
		wr=strlen(name);
#else
		fn_utf8=g_filename_to_utf8(name,-1,&rd,&wr,NULL);
#endif
		fn=g_build_filename(folders,fn_utf8,NULL);
		if(g_file_test(fn,G_FILE_TEST_IS_DIR))
		{
			fldr=g_new(fldr_s,1);
			fldr->path=fn;
			fldr->name=fn_utf8;
			dirs=g_list_append(dirs,fldr);
			sqlite3_bind_text(stmt,1,fn_utf8,wr,SQLITE_STATIC);
			sqlite3_step(stmt);
			sqlite3_reset(stmt);
		}
		else
		{
			g_free(fn);
			g_free(fn_utf8);
		}
	}
	sqlite3_finalize(stmt);
	g_dir_close(dir);
	g_free(folders);
	END_TRANSACTION(db,stmt);
	
	GList *i=dirs;
	GMimeParser *parser=NULL;
	GMimeStream *stream=NULL;
	GMimeMessage *msg=NULL;
	article_add_s art;
	
	BEGIN_TRANSACTION(db,stmt);
		
	const char *sql_4="SELECT id FROM groups WHERE name=?;";
	const char *sql_5="INSERT OR IGNORE INTO group_server(group_id, server_id, "
						"article_min, article_max, article_max_old, "
						"permision)"
						"SELECT id,null,0,0,0,'y' FROM groups WHERE folder=1;";
	const char *sql_6="SELECT id FROM group_server WHERE group_id=? AND server_id=null ;";
	/* map to sid null */
	sq_res=sqlite3_prepare(db,sql_5,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	int gid,gsid;
	while(i)
	{
		fldr=i->data;
		/* get gid  */
		sq_res=sqlite3_prepare(db,sql_4,-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		sqlite3_bind_text(stmt,1,fldr->name,-1,SQLITE_STATIC);
		sqlite3_step(stmt);
		gid=sqlite3_column_int(stmt,0);
		sqlite3_finalize(stmt);
		/* get gsid */
		sq_res=sqlite3_prepare(db,sql_6,-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		sqlite3_bind_int(stmt,1,gid);
		sqlite3_step(stmt);
		gsid=sqlite3_column_int(stmt,0);
		sqlite3_finalize(stmt);
		
		FILE *file;
		dir=g_dir_open(fldr->path,0,NULL);
		while(name=g_dir_read_name(dir))
		{
			fn=g_build_filename(fldr->path,name,NULL);
			if(g_file_test(fn,G_FILE_TEST_IS_REGULAR))
			{
				stream=NULL;
				parser=NULL;
				msg=NULL;
				
				file=fopen(fn,"rb");
				stream=g_mime_stream_file_new(file);
				if(stream)
					parser=g_mime_parser_new_with_stream(stream);
				if(parser)
					msg=g_mime_parser_construct_message(parser);

				if(msg)
				{
					const char *hdr;
					int len;
					time_t timet;
					
					hdr=g_mime_message_get_message_id(msg);
					len=strlen(hdr);
					art.msg_id=pstring_shallow(hdr,len);
					
					g_mime_message_get_date(msg,&art.date,&len);
					
					hdr=g_mime_message_get_header(msg,"Lines");
					if(hdr)
						len=atoi(hdr);
					else
						len=0;
					art.lines=len;
					
					hdr=g_mime_message_get_subject(msg);
					art.subject=pstring_shallow(hdr,strlen(hdr));
					
					
					hdr=g_mime_message_get_sender(msg);
					art.author=pstring_shallow(hdr,strlen(hdr));
					
					hdr=g_mime_message_get_header(msg,"References");
					len=hdr?strlen(hdr):0;
					art.refs=pstring_shallow(hdr,len);
					
					hdr=g_mime_message_get_header(msg,"Xref");
					len=hdr?strlen(hdr):0;
					art.xref=pstring_shallow(hdr,len);
					
					add_article(gid,gsid,&art);
					
					g_object_unref(msg);
				}
				if(parser)
					g_object_unref(parser);
				if(stream)
					g_object_unref(stream);
				fclose(file);
			}
			g_free(fn);
		}
		g_dir_close(dir);
		
		sq_res=sqlite3_prepare(db,"UPDATE article SET read=1, new=0 WHERE id IN "
									"(SELECT article_id FROM group_article WHERE "
									"group_id=?);",-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		sqlite3_bind_int(stmt,1,gid);
		sqlite3_step(stmt);
		sqlite3_finalize(stmt);
		
		sq_res=sqlite3_prepare(db,"UPDATE groups SET qty=(SELECT count(*) "
									"FROM group_article WHERE group_id=?1 ), "
									"qty_read=(SELECT count(*) FROM group_article, article "
									"WHERE group_id=?1 AND article.id=article_id AND read=1 ) "
									"WHERE id=?1 ;",-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		sqlite3_bind_int(stmt,1,gid);
		sqlite3_step(stmt);
		sqlite3_finalize(stmt);
		
		g_free(fldr->name);
		g_free(fldr->path);
		g_free(fldr);
		i=i->next;
	}
	END_TRANSACTION(db,stmt);
	
	g_list_free(dirs);
	
	debug_exit("migrate_folders");
}

static void thread_groups(void)
{
	debug_enter("thread_groups");
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	int sq_res, gid,i;
	GArray *tmp=g_array_sized_new(FALSE,FALSE,sizeof(gint),128);
	
	sq_res=sqlite3_prepare(db,"SELECT DISTINCT group_id FROM group_article;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	while(sqlite3_step(stmt)==SQLITE_ROW)
	{
		gid=sqlite3_column_int(stmt,0);
		g_array_append_val(tmp,gid);
	}
	sqlite3_finalize(stmt);
	
	for(i=0;i<tmp->len;i++)
	{
		gid=g_array_index(tmp,gint,i);
		thread_group(gid);
	}
	
	g_array_free(tmp,TRUE);
	
	debug_exit("thread_groups");
}

