/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2004  K. Haley <haleykd@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <glib.h>
#include <glib/gi18n-lib.h>
#include <pan/base/db.h>
#include <pan/base/db-add.h>
#include <pan/base/message-source.h>
#include <pan/base/debug.h>
#include <pan/base/pstring.h>
#include <pan/base/newsrc.h>
#include <pan/base/pan-config.h>
#include <pan/base/group.h>

static void m_servers(sqlite3 *db);
static void m_load_groups(sqlite3 *db);
static gint import_v12(sqlite3 *db, gint32 gid, gint32 sid, const char *fn, Newsrc *nrc);

void import_db(void)
{
	sqlite3 *db;
	char *errmsg=NULL;
	debug_enter("import_db");
	
	db=get_db();
	
	m_servers(db);
	m_load_groups(db);
	
	debug_exit("import_db");
}

#define SERVER_NAME_KEY "/Pan/Servers/Names"

static void m_servers(sqlite3 *db)
{
	sqlite3_stmt *vm=NULL;
	gchar * p;
	int sq_res,sq_len;
	char *sq_err;
	
	debug_enter("m_servers");
	
	/* build all the servers in the list */
   	p = pan_config_get_string (SERVER_NAME_KEY, NULL); /*commas*/
	if (p != NULL)
	{
		const char * walk = p;
		PString pstr=PSTRING_INIT;
		
		char section_name[PATH_MAX], ins_tmp[40];
		PString name = PSTRING_INIT;
		PString host=PSTRING_INIT,user=PSTRING_INIT,pass=PSTRING_INIT,newsrc_fn=PSTRING_INIT;
		int port,max_conn,idle;
		unsigned long update;
		gboolean newsrcb,auth;

		const char *sql="INSERT INTO server( name, hostname, port, username , "
						"password , rank, max_conn, idle_timeout, need_auth, "
						"newsrc, last_gl_update, newsrc_fn )"
						"VALUES(?,?,?,?,?,?,?,?,?,?,?,?);";
						
		BEGIN_TRANSACTION(db,vm);
		sq_res=sqlite3_prepare(db,sql,-1, &vm, NULL);
		SQL_PREP_VALID(sq_res,db);	
		/* create a server object for each server in the list */
		while (get_next_token_pstring (walk, ',', &walk, &pstr))
		{
			if (pstr.len) {
				name = pstring_shallow (pstr.str, pstr.len);
				name = pstring_strstrip_shallow (&name);
				g_snprintf (section_name, sizeof(section_name), "/Pan/%*.*s", name.len, name.len, name.str);
				if (pan_config_has_section (section_name))
				{

					/* initialize the server from pan_config */
					pan_config_push_prefix (section_name);
					host = pstring_shallow (pan_config_get_string ("Address", NULL), -1);
					port = pan_config_get_int ("Port", 119);
					max_conn = pan_config_get_int ("Max_Connections", 2);
					idle = pan_config_get_int ("Idle_Timeout_Secs", 180);
					update = pan_config_get_ulong ("Last_Grouplist_Update", 0ul);
					newsrcb = pan_config_get_bool ("Use_Newsrc", FALSE);
					newsrc_fn = pstring_shallow (pan_config_get_string ("Newsrc_Filename", NULL), -1);
					auth = pan_config_get_bool ("Auth_Required", FALSE);
					user = pstring_shallow (pan_config_get_string ("Username", NULL), -1);
					pass = pstring_shallow (pan_config_get_string ("Password", NULL), -1);
					pan_config_pop_prefix ();
					
					sqlite3_bind_text(vm,1,name.str,name.len,SQLITE_STATIC);
					sqlite3_bind_text(vm,2,host.str,host.len,SQLITE_STATIC);
					sqlite3_bind_int(vm,3,port);
					sqlite3_bind_text(vm,4,user.str,user.len,SQLITE_STATIC);
					sqlite3_bind_text(vm,5,pass.str,pass.len,SQLITE_STATIC);
					sqlite3_bind_int(vm,6,0); /*rank*/
					sqlite3_bind_int(vm,7,max_conn);
					sqlite3_bind_int(vm,8,idle);
					sqlite3_bind_int(vm,9,auth);
					sqlite3_bind_int(vm,10,newsrcb);
					sqlite3_bind_int(vm,11,update);
					sqlite3_bind_text(vm,12,newsrc_fn.str,newsrc_fn.len,SQLITE_STATIC);
					
					sq_res=sqlite3_step(vm);
					if(sq_res==SQLITE_ERROR)
					{
						printf("servers - error adding server - %s",sqlite3_errmsg(db));
						sqlite3_finalize(vm);
						sqlite3_close(db);
						exit(1);
					}
					sqlite3_reset(vm);
				}
			}
		}
		sqlite3_finalize(vm);
		END_TRANSACTION(db,vm);

		/* cleanup */
		pstring_clear(&host);
		pstring_clear(&user);
		pstring_clear(&pass);
		pstring_clear(&newsrc_fn);
		g_free(p);
	}
	debug_exit("m_servers");
}

enum
{
	GROUP_ALL				= (1<<0),
	GROUP_SUBSCRIBED			= (1<<1), 
	GROUP_NEW				= (1<<2),
	GROUP_FOLDERS				= (1<<3)
};

static gint
m_read_group_data(sqlite3 *db, int sid, //server id
				const char    * dat_str,
                 guint           dat_len,
                 const char    * idx,
				 const char * sname)
{
	gint version;
	const gchar * march_idx = idx;
	gint group_qty = 0;
	
	debug_enter ("m_read_group_data");

	/* sanity clause */
	g_return_val_if_fail (dat_str!=NULL, 0);
	g_return_val_if_fail (idx!=NULL, 0);

	version = get_next_token_int (march_idx, '\n', &march_idx);
	if (version>=1 && version<=13)
	{
		long i;
		long qty = get_next_token_long (march_idx, '\n', &march_idx);
		int res;
		char *errmsg;

		const char *mk_gtbl="CREATE TEMP TABLE tg(name TEXT , description TEXT, "
						"filter_name TEXT, filter_show INT, filter_bits INT, "
						"sort INT, sort_old INT, identity TEXT, download_dir TEXT, "
						"charset TEXT , qty INT, qty_read INT, new INT, folder INT, "
						"server_id INT, subscribed INT, "
						"article_min INT, article_max INT, loaded_since_fetch INT, "
						"article_max_old INT, permision TEXT, gid INT, newsrc TEXT, "
						"id INTEGER PRIMARY KEY);";
		res=sqlite3_exec(db,mk_gtbl,NULL,NULL,&errmsg);
		SQL_EXEC_VALID(res,errmsg);
		
		const char *sql_1="INSERT INTO tg( name, description, download_dir,"
							"charset, filter_name, identity, filter_show,"
							"filter_bits, sort_old, sort, server_id, article_min,"
							"loaded_since_fetch, article_max, article_max_old,"
							"qty, qty_read, permision, newsrc, new, folder, subscribed)"
							"VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);";
		sqlite3_stmt *stmt_1,*stmt_2,*stmt_3;
		
		res=sqlite3_prepare(db,sql_1,-1,&stmt_1,NULL);
		SQL_PREP_VALID(res,db);
		for (i=0; i<qty; ++i)
		{
			const char * newsrc;
			const char * purged;
			glong l;
			gint8 old_sort,new_sort;
			char perm;
			gint32 gid;
			char *freeme;

			/* group name */
			l = get_next_token_long (march_idx, '\n', &march_idx);
			//g = group_new (server, dat_str+l);
			sqlite3_bind_text(stmt_1,1,dat_str+l,-1,SQLITE_STATIC);
			/* readable name removed in version 12 */
			if (version<12)
				skip_next_token (march_idx, '\n', &march_idx);

			/* description */
			freeme=NULL;
			l = get_next_token_long (march_idx, '\n', &march_idx);
			if (0<=l && l<dat_len)
			{
				pan_utf8ize(dat_str+l,strlen(dat_str+l),&freeme);
				sqlite3_bind_text(stmt_1,2,freeme,-1,SQLITE_STATIC);
			}
				//g->description = g_strdup (dat_str+l);

			/* download dir */
			l = get_next_token_long (march_idx, '\n', &march_idx);
			if (0<=l && l<dat_len)
				sqlite3_bind_text(stmt_1,3,dat_str+l,-1,SQLITE_STATIC);
				//g->download_dir = g_strdup (dat_str+l);

			/* default charset added in version 11 */
			if (version>=11) {
				l = get_next_token_long (march_idx, '\n', &march_idx);
				if (0<=l && l<dat_len)
					sqlite3_bind_text(stmt_1,4,dat_str+l,-1,SQLITE_STATIC);
					//g->default_charset = g_strdup (dat_str+l);
			}

			/* newsrc string */
			newsrc = NULL;
			l = get_next_token_long (march_idx, '\n', &march_idx);
			if (0<=l && l<dat_len)
				newsrc = dat_str+l;

			/* purged crosspost string added in version 3 */
			purged = NULL;
			if (version>=3) {
				l = get_next_token_long (march_idx, '\n', &march_idx);
				if (0<=l && l<dat_len)
					purged = dat_str+l;
			}

			/* filter name added in version 4 */
			if (version>=4 ) {
				l = get_next_token_long (march_idx, '\n', &march_idx);
				if (0<=l && l<dat_len)
					sqlite3_bind_text(stmt_1,5,dat_str+l,-1,SQLITE_STATIC);
					//g->filter_name = g_strdup (dat_str+l);
			}

			/* identity name added in version 6 */
			if (version>=6) {
				l = get_next_token_long (march_idx, '\n', &march_idx);
				if (0<=l && l<dat_len)
					sqlite3_bind_text(stmt_1,6,dat_str+l,-1,SQLITE_STATIC);
					//g->identity_name = g_strdup (dat_str+l);
			}

			/*g->flags = (gint8)*/l=   get_next_token_int   (march_idx, '\n', &march_idx);
			sqlite3_bind_int(stmt_1,20,l&GROUP_NEW);
			sqlite3_bind_int(stmt_1,21,l&GROUP_FOLDERS);
			sqlite3_bind_int(stmt_1,22,l&GROUP_SUBSCRIBED);

			if (version>=8)
				l = get_next_token_ulong (march_idx, '\n', &march_idx);
			else
				l = -1;
			if (l!=FILTER_SHOW_MATCHES && l!=FILTER_SHOW_SUBTHREADS && l!=FILTER_SHOW_THREAD)
				l = FILTER_SHOW_MATCHES;
			sqlite3_bind_int(stmt_1,7,l);

			/* the filter bitfields changed between versions 12 and 13 ...
			    reset the fields for file format versions < 13 */
			if (version >= 13) {
				l = (guint32) get_next_token_ulong (march_idx, '\n', &march_idx);
			} else {
				skip_next_token (march_idx, '\n', &march_idx);
				l = ~STATE_FILTER_SCORE_IGNORED;
			}
			/* filter_bits got widened from guint16 to guint32 in version 7 */
			if (version<=6)  
				l |= 0xFFFF8000;
			sqlite3_bind_int(stmt_1,8,l);

			/* old sort style */
			old_sort=0;
			if (version >= 10)
				old_sort = (gint8) get_next_token_int (march_idx, '\n', &march_idx);

			new_sort = (gint8) get_next_token_int (march_idx, '\n', &march_idx);

			/* compare sort styles */
			if (version >= 10)
				if (abs(old_sort) == abs(new_sort))
					old_sort = 0;
			sqlite3_bind_int(stmt_1,9,old_sort);
			sqlite3_bind_int(stmt_1,10,new_sort);

			/* server id */
			sqlite3_bind_int(stmt_1,11,sid);
			
			l = (gulong) get_next_token_ulong (march_idx, '\n', &march_idx);
			sqlite3_bind_int(stmt_1,12,l);

			/* load since fetch */
			if (version>=9)
				l = get_next_token_int (march_idx, '\n', &march_idx) != 0;
			else
				l = FALSE;
			sqlite3_bind_int(stmt_1,13,l);

			/* article high */
			l     = (gulong)  get_next_token_ulong (march_idx, '\n', &march_idx);
			sqlite3_bind_int(stmt_1,14,l);

			/* article_high_old added in version 5 */
			if (version>=5)
				l = (gulong)  get_next_token_ulong (march_idx, '\n', &march_idx);
			//else
				//g->article_high_old =  g->article_high;
			sqlite3_bind_int(stmt_1,15,l);

			/* qty */
			l = (gint32)  get_next_token_int   (march_idx, '\n', &march_idx);
			sqlite3_bind_int(stmt_1,16,l);
			/* read qty */
			l = (gint32)  get_next_token_int   (march_idx, '\n', &march_idx);
			sqlite3_bind_int(stmt_1,17,l);

			/* permissions added in version 2 */
			if (version==1)
			{
				perm = 'y'; /* a reasonable default... */
			}
			else
			{
				gchar ch = '\0';
				if (march_idx!=NULL)
					ch = *march_idx;
				if (ch!='y' && ch!='n' && ch!='m')
					ch = '?';
				perm = ch;
				skip_next_token (march_idx, '\n', &march_idx);
			}
			sqlite3_bind_text(stmt_1,18,&perm,1,SQLITE_STATIC);
			sqlite3_bind_text(stmt_1,19,newsrc,-1,SQLITE_STATIC);
			sqlite3_step(stmt_1);
			sqlite3_reset(stmt_1);

			group_qty++;

			if(freeme)
				g_free(freeme);

		}
		sqlite3_finalize(stmt_1);
		
		/*add new groups to global group list */
		const char *sql_2="INSERT OR IGNORE INTO groups(name,description,qty, "
						"qty_read, filter_name, filter_show, filter_bits, "
						"sort, sort_old,identity,download_dir,charset,new, "
						"subscribed,folder) "
						"SELECT name,description,0,0,filter_name, "
						"filter_show,filter_bits,sort,sort_old,identity, "
						"download_dir,charset,new,subscribed,folder FROM tg;";
		res=sqlite3_exec(db,sql_2,NULL,NULL,&errmsg);
		SQL_EXEC_VALID(res,errmsg);
		
		// QUESTION: would an UPDATE using EXISTS eliminate the need for t2?
		/* set group id in temp table from global group list*/
		res=sqlite3_exec(db,"CREATE TEMP TABLE t2(gid INT, tgid INT);"
							"INSERT INTO t2 SELECT groups.id, tg.id FROM tg, groups "
							"WHERE groups.name=tg.name;",NULL,NULL,&errmsg);
		SQL_EXEC_VALID(res,errmsg);
		
		const char *sql_3="SELECT gid, tgid FROM t2 ;";
		res=sqlite3_prepare(db,sql_3,-1, &stmt_1, NULL);
		SQL_PREP_VALID(res,db);
		res=sqlite3_prepare(db,"UPDATE tg SET gid=? WHERE id=? ;",-1, &stmt_2, NULL);
		SQL_PREP_VALID(res,db);
		while(sqlite3_step(stmt_1)==SQLITE_ROW)
		{
			gint32 gid,tgid;
			gid=sqlite3_column_int(stmt_1,0);
			tgid=sqlite3_column_int(stmt_1,1);
			sqlite3_bind_int(stmt_2,1,gid);
			sqlite3_bind_int(stmt_2,2,tgid);
			res=sqlite3_step(stmt_2);
			sqlite3_reset(stmt_2);
		}
		sqlite3_finalize(stmt_1);
		sqlite3_finalize(stmt_2);
		res=sqlite3_exec(db,"DROP TABLE t2;",NULL,NULL,&errmsg);
		SQL_EXEC_VALID(res,errmsg);
		
		/* add group & server info into group_server table from temp table */
		const char *sql_4="INSERT  INTO group_server(group_id,server_id,article_min,"
					"article_max,article_max_old,permision)"
					"SELECT gid,server_id,article_min,article_max,"
					"article_max_old, permision FROM tg ;";
		res=sqlite3_exec(db,sql_4,NULL,NULL,&errmsg);
		SQL_EXEC_VALID(res,errmsg);
		
		char afn[PATH_MAX];
		const char *gn, *tmp;
		typedef struct 
		{
			gint32 gid;
			Newsrc *nrc;
			char *fn;
		} tas_s;
		tas_s *s;
		GPtrArray *pa_tas=g_ptr_array_new();
		
		/* used for looping and getting article info */
		const char *sql_5="SELECT gid,name,newsrc FROM tg;";
		res=sqlite3_prepare(db,sql_5,-1, &stmt_1, NULL);
		SQL_PREP_VALID(res,db);
		
		while(sqlite3_step(stmt_1)==SQLITE_ROW)
		{
			gn=sqlite3_column_text(stmt_1,1);
			tmp=sqlite3_column_text(stmt_1,2);

			g_snprintf(afn, PATH_MAX, "%s%c%s%c%s",
				get_data_dir(), G_DIR_SEPARATOR, sname,
				G_DIR_SEPARATOR, gn);
			/* have articles? */
			if(pan_file_exists(afn))
			{
				s=g_new0(tas_s,1);
				s->gid=sqlite3_column_int(stmt_1,0);
				s->fn=g_strdup(afn);
				
				if(tmp)
				{
					s->nrc=newsrc_new(NULL,0,0);
					newsrc_init_from_range_str(s->nrc,tmp);
				}
				g_ptr_array_add(pa_tas,s);
			}
		}
		sqlite3_finalize(stmt_1);
		
		res=sqlite3_exec(db,"DROP TABLE tg;",NULL,NULL,&errmsg);
		SQL_EXEC_VALID(res,errmsg);
		
		/* load article info */
		for(i=0;i<pa_tas->len;i++)
		{
			s=(tas_s*)g_ptr_array_index(pa_tas,i);
			import_v12(db,s->gid,sid,s->fn,s->nrc);
			g_free(s->fn);
			if(s->nrc)
				g_object_unref(s->nrc);
			g_free(s);
		}
		g_ptr_array_free(pa_tas,TRUE);
	}


	debug_exit("m_read_group_data");
	return group_qty;
}

typedef struct _sid
{
	int id;
	char *name;
} sid;

static void m_load_groups(sqlite3 *db)
{
	sqlite3_stmt *stm;
	int sq_res,cnt;
	GArray *servers;
	sid t_sid;
	
	debug_enter("m_load_groups");
	
	sq_res=sqlite3_prepare(db,"SELECT count(*) FROM server;",-1,&stm,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_step(stm);
	cnt=sqlite3_column_int(stm,0);
	sqlite3_finalize(stm);
	
	servers=g_array_sized_new(FALSE,TRUE,sizeof(sid),cnt);
	sq_res=sqlite3_prepare(db,"SELECT name,id from server;",-1,&stm,NULL);
	SQL_PREP_VALID(sq_res,db);
	while(sqlite3_step(stm)!=SQLITE_DONE)
	{
		t_sid.id=sqlite3_column_int(stm,1);
		t_sid.name=g_strdup(sqlite3_column_text(stm,0));
		g_array_append_val(servers,t_sid);
	}
	sqlite3_finalize(stm);
	
	char *dat,*idx,path[PATH_MAX];
	int dat_len,idx_len;
	for(cnt=0;cnt<servers->len;cnt++)
	{
		t_sid=g_array_index(servers,sid,cnt);
		
		g_snprintf(path, sizeof(path), "%s%c%s_sub.dat", get_data_dir(), G_DIR_SEPARATOR, t_sid.name);
		pan_file_normalize_inplace(path);
		g_file_get_contents(path, &dat, &dat_len, NULL);
		g_snprintf(path, sizeof(path), "%s%c%s_sub.idx", get_data_dir(), G_DIR_SEPARATOR, t_sid.name);
		pan_file_normalize_inplace(path);
		g_file_get_contents(path, &idx, &idx_len, NULL);
		
		m_read_group_data(db,t_sid.id,dat,dat_len,idx,t_sid.name);
		
		g_snprintf(path, sizeof(path), "%s%c%s_unsub.dat", get_data_dir(), G_DIR_SEPARATOR, t_sid.name);
		pan_file_normalize_inplace(path);
		g_file_get_contents(path, &dat, &dat_len, NULL);
		g_snprintf(path, sizeof(path), "%s%c%s_unsub.idx", get_data_dir(), G_DIR_SEPARATOR, t_sid.name);
		pan_file_normalize_inplace(path);
		g_file_get_contents(path, &idx, &idx_len, NULL);
		
		m_read_group_data(db,t_sid.id,dat,dat_len,idx,t_sid.name);

		g_free(t_sid.name);
	}
	g_array_free(servers,TRUE);
	
	debug_exit("m_load_groups");
}

static gint import_v12(sqlite3 *db, gint32 gid, gint32 sid, const char *fn, Newsrc *nrc)
{
	GIOChannel * in_gio;
	gboolean success = FALSE;
	debug_enter("import_v12");
	/* open the group's file */
	in_gio = g_io_channel_new_file (fn, "r", NULL);
	
	if (in_gio != NULL)
	{
		int version;
		GString * line=NULL;
		GString *mid=NULL,*author_r=NULL,*author_a=NULL;
		GString *subject=NULL,*references=NULL,*xref=NULL;
		gint sub_id, aa_id, ar_id;

		g_io_channel_set_encoding (in_gio, NULL, NULL);
 		line = g_string_sized_new (1024);
		mid = g_string_sized_new (1024);
		author_r = g_string_sized_new (1024);
		author_a = g_string_sized_new (1024);
		subject = g_string_sized_new (1024);
		xref = g_string_sized_new (1024);
		references = g_string_sized_new (1024);

		g_io_channel_read_line_string (in_gio, line, NULL, NULL);
		version = atoi (line->str);
		if (version==11)
		{
			int i;
			long qty;
			char *errmsg;
			article_add_s aas;
			int sq_res;
			sqlite3_stmt *stmt;
			gboolean is_new,is_read;
			gint64 aid;
			
			BEGIN_TRANSACTION(db,stmt);
			
			/* quantity of articles */
			g_io_channel_read_line_string (in_gio, line, NULL, NULL);
			qty = strtol (line->str, NULL, 10);
			
			sq_res=sqlite3_prepare(db,"UPDATE article SET read=? , new=? WHERE id=?;",-1,&stmt,NULL);
			SQL_PREP_VALID(sq_res,db);

			for (i=0; i!=qty; ++i)
			{
#ifdef G_OS_WIN32
#define EOLN_LEN 2 /* \r\n */
#else
#define EOLN_LEN 1 /* \n */
#endif 

#define read_next_line(xstr) \
				g_io_channel_read_line_string(in_gio,xstr,NULL,NULL);\
				g_string_truncate(xstr,xstr->len-EOLN_LEN);

				read_next_line(mid);
				aas.msg_id=pstring_shallow(mid->str,mid->len);
				
				read_next_line(author_a);
				read_next_line(author_r);
				g_string_prepend_c(author_r,'"');
				g_string_append_printf(author_r,"\" <%s>",author_a->str);
				aas.author=pstring_shallow(author_r->str,author_r->len);
				
				read_next_line(subject);
				aas.subject=pstring_shallow(subject->str,subject->len);
				
				read_next_line(references)
				aas.refs=pstring_shallow(references->str,references->len);
				read_next_line(xref);
				aas.xref=pstring_shallow(xref->str,xref->len);
				
				// these are not needed
				if(version!=12)
					read_next_line(line);
				//a->part = (gint16) atoi (line->str);
				read_next_line(line);
				//parts = (gint16) atoi (line->str);
				
				read_next_line(line);
				aas.lines = (guint16) atoi(line->str);
				
				read_next_line(line);
				aas.byte_qty = strtoul(line->str, NULL, 10);
				
				read_next_line(line);
				aas.date = (time_t) strtoul(line->str, NULL, 10);
				
				read_next_line(line);
				aas.uid = strtoul(line->str, NULL, 10);
				
				read_next_line(line);
				is_new = atoi(line->str) != 0;
				if(nrc)
				{
					is_read=newsrc_is_article_read(nrc,aas.uid);
				}else
					is_read=FALSE;
				if(version<12)
					aid=add_article(gid,sid,&aas);
				else
				{
					// FIXME: is my assumption about the subject right?
					int attachment_qty;
					gint16 part,i;
					MessageSources ms;
					char *pre,*norm,*post;
					normalize_subject(&aas.subject,&pre,&norm,&post);
					read_next_line(line); attachment_qty = atoi(line->str);
					for(i=0;i<attachment_qty;++i)
					{
						read_next_line(line); part=(gint16)atoi(line->str);
						read_next_line(line); aas.lines = (guint16) atoi(line->str);
						read_next_line(line); aas.byte_qty = strtoul(line->str, NULL, 10);
						read_next_line(line); aas.date = (time_t) strtoul(line->str, NULL, 10);
						read_next_line(mid); aas.msg_id=pstring_shallow(mid->str,mid->len);
						read_next_line(xref); aas.xref=pstring_shallow(xref->str,xref->len);
						message_sources_add_from_xref(&ms,sid,xref->str,xref->len);
						aas.uid=message_sources_get_number(&ms,gid,sid);
						message_sources_clear(&ms);
						g_string_printf(subject,"%s%s%c%d%s",pre,norm,*post,part,post+1);
						aas.subject=pstring_shallow(subject->str,subject->len);
						aid=add_article(gid,sid,&aas);
					}
					
				}
				
				sqlite3_bind_int(stmt,1,is_read);
				sqlite3_bind_int(stmt,2,is_new);
				sqlite3_bind_int64(stmt,3,aid);
				sqlite3_step(stmt);
				sqlite3_reset(stmt);
			}

			sqlite3_finalize(stmt);
			
			/* update group article counts */
			sq_res=sqlite3_prepare(db,"UPDATE groups SET qty_read="
										"(SELECT sum(read) FROM group_article,article "
										"WHERE group_id=?1 AND article.id=article_id) "
										"WHERE id=?1 ;",-1,&stmt,NULL);
			SQL_PREP_VALID(sq_res,db);
			sqlite3_bind_int(stmt,1,gid);
			sqlite3_step(stmt);
			sqlite3_finalize(stmt);
			
			END_TRANSACTION(db,stmt);
			success=TRUE;
		}

		/* cleanup */
		g_string_free(line, TRUE);
		g_string_free(mid, TRUE);
		g_string_free(author_a, TRUE);
		g_string_free(author_r, TRUE);
		g_string_free(subject, TRUE);
		g_string_free(xref, TRUE);
		g_string_free(references, TRUE);
		g_io_channel_unref(in_gio);
	}

	debug_exit("import_v12");
	/* cleanup */
	return success;
}


