/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <pan/base/debug.h>
#include <pan/base/message-source.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/pstring.h>
#include <pan/base/db.h>
#include <pan/base/serverlist.h>
#include <pan/base/log.h>

const MessageSource MESSAGE_SOURCE_INIT = { 0, 0, 0ul };
const MessageSource MESSAGE_SOURCE_TOXIC = { ~0, ~0, ~0ul };

const MessageSources MESSAGE_SOURCES_INIT = { NULL, NULL, 0 };
const MessageSources MESSAGE_SOURCES_TOXIC = { (void*)0xDEADBEEF, (void*)0xDEADBEEF, ~0 };

void
message_sources_clear(MessageSources * sources)
{
	g_return_if_fail(sources != NULL);

	g_array_free(sources->array,TRUE);
	*sources = MESSAGE_SOURCES_INIT;
}

void
message_sources_add(MessageSources   * sources,
                     gint32             server_id,
                     GQuark             group,
                     gulong             number)
{
	MessageSource source, *it, *end;
	int cnt;

	/* sanity clause */
	g_return_if_fail(server_id!=0); /* bad server */
	g_return_if_fail(group!=0); /* bad group */
	g_return_if_fail(number!=0ul); /* bad number */
	g_return_if_fail(sources!=NULL);
	
	if(sources->array!=NULL)
		for (it=sources->sources, cnt=0; cnt<sources->array->len; ++it,++cnt)
			if(!(it->server_id==server_id && it->group==group))
			{
				//log_add_va(LOG_ERROR,"duplicate source: src=%p len=%d cnt=%d server_id=%d group=%d uid=%d",
				//	sources,sources->array->len,cnt,server_id,group, number);
				return;
			}

	/* add it */
	if(sources->array==NULL)
		sources->array=g_array_sized_new(FALSE,FALSE,sizeof(MessageSource),2);
	
	source.server_id = server_id;
	source.group = group;
	source.number = number;
	g_array_append_val(sources->array,source);
	sources->sources=(MessageSource*)sources->array->data;
	sources->source_qty=sources->array->len;
}

void
message_sources_remove_server(MessageSources   * sources,
                               gint32             server_id)
{
	int remain_qty;
	MessageSource * it, * end;

	g_return_if_fail(sources!=NULL);
	g_return_if_fail(server_id!=0);

	/* how many sources will be left? */
	remain_qty = 0;
	for (it=sources->sources, end=it+sources->source_qty; it!=end; ++it)
		if (it->server_id!=server_id)
			++remain_qty;

//odebug2 ("remove_server: going from %d sources to %d", sources->source_qty, remain_qty);

	/* remove those sources */
	if (!remain_qty)
		message_sources_clear(sources);
	else
	{
		GArray *a=sources->array;
		MessageSource *it=sources->sources+a->len-1;
		int i;
		for (i=a->len-1;i>-1;--i,--it)
			if (it->server_id!=server_id)
				g_array_remove_index(a,i);
		sources->sources = (MessageSource*)a->data;
		sources->source_qty = a->len;
	}
}

void
message_sources_remove_full(MessageSources *sources, gint32 server_id, GQuark group)
{
	int i;
	MessageSource *s;
	for(i=0, s=sources->sources ;i<sources->source_qty;++i, ++s)
		if(s->server_id==server_id && s->group==group)
		{
			g_array_remove_index(sources->array,i);
			--sources->source_qty;
			break;
		}
	
}

void
message_sources_foreach(const MessageSources       * sources,
                         gint32                       server_filter,
                         MessageSourcesForeachFunc    foreach_func,
                         gpointer                     foreach_user_data)
{
	int i;

	g_return_if_fail(sources != NULL);

	for (i=0; i<sources->source_qty; ++i) {
		const MessageSource * s = &sources->sources[i];
		if (server_filter==0 || server_filter==s->server_id)
			(foreach_func)(s->server_id, s->group, s->number, foreach_user_data);
	}
}
void message_sources_group_foreach(const MessageSources     * sources,
									MessageSourcesGroupForeachFunc  func,
									gpointer                   user_data)
{
	int i;
	GHashTable *hash=g_hash_table_new(g_direct_hash,g_direct_equal);

	g_return_if_fail(sources != NULL);

	for (i=0; i<sources->source_qty; ++i) {
		const MessageSource * s = &sources->sources[i];
		if (!g_hash_table_lookup(hash,GINT_TO_POINTER(s->group)) )
		{
			g_hash_table_insert(hash, GINT_TO_POINTER(s->group), GINT_TO_POINTER(s->group) );
			(func)(s->group, user_data);
		}
	}
	g_hash_table_destroy(hash);
}

void
message_sources_add_from_xref(MessageSources   * sources,
                               gint32             server_id,
                               const char       * xref_str,
                               int                xref_len)
{
	PString xref = PSTRING_INIT;

	g_return_if_fail(sources != NULL);
	g_return_if_fail(server_id != 0);
//log_add_va(LOG_INFO,"from xref sid=%d xref=%s",server_id,xref_str);
	if (is_nonempty_string(xref_str))
	{
		/* strip ends and erase leading "Xref: " if present */
		xref = pstring_shallow(xref_str, xref_len);
		xref = pstring_strstrip_shallow(&xref);
		if (xref.len>6 && !strncmp(xref.str,"Xref: ", 6))
		{
			xref = pstring_substr_shallow(&xref, xref.str+6, NULL);
			xref = pstring_strstrip_shallow(&xref);
		}
	}

	if (xref.len)
	{
		PString run;
		char * tmp = g_strndup(xref.str, xref.len);
		const char * march = tmp;
		
		/* walk through the xrefs, of format "group1:number group2:number" */
		while (get_next_token_pstring(march, ' ', &march, &run))
		{
			char * delimit = pan_strchr_len(run.str, run.len, ':');
			if (delimit != NULL )
			{
				GQuark group=g_quark_from_string(run.str);
				const gulong number = strtoul (delimit+1, NULL, 10);
				*delimit = '\0';
				message_sources_add(sources, server_id, group, number);
			}
		}

		g_free(tmp);
	}
	//log_add(LOG_INFO,"from xref done");
}

void
message_sources_copy(MessageSources        * to,
                      const MessageSources  * from)
{
	int i;
	g_return_if_fail(to != NULL);
	g_return_if_fail(from != NULL);

	if (to != from)
	{
		to->source_qty = from->source_qty;
		to->array=g_array_sized_new(FALSE,FALSE,sizeof(MessageSource),from->source_qty);
		for(i=0;i<from->source_qty;++i)
			g_array_append_val(to->array,from->sources[i]);
		to->sources = (MessageSource*)to->array->data;
	}
}

void
message_sources_to_xref(const MessageSources  * sources,
                         GString               * appendme,
                         gint32                  server_filter,
                         gboolean                prepend_server_name)
{
	int i;
	int added_qty;
	int old_len;

	g_return_if_fail(sources != NULL);
	g_return_if_fail(appendme != NULL);
	g_return_if_fail(server_filter != 0);

	old_len = appendme->len;
	for (added_qty=i=0; i<sources->source_qty; ++i)
	{
		const MessageSource * s = &sources->sources[i];
		if (server_filter == s->server_id )
		{
			const char * group_name = g_quark_to_string(s->group);
			if (added_qty)
				g_string_append_c(appendme, ' ');
			g_string_append_printf(appendme, "%s:%lu", group_name, s->number);
			++added_qty;
		}
	}

	if (prepend_server_name && added_qty)
	{
		Server *server=serverlist_get_by_id(server_filter);
		const char * server_name = server->name.str;
		g_string_insert_c(appendme, old_len, ' ');
		g_string_insert(appendme, old_len, server_name);
	}
}

gboolean
message_sources_get_source(const MessageSources     * sources,
                            gint32                     server_key,
                            MessageSource            * setme)
{
	gboolean found = FALSE;
	MessageSource * it, * end;

	g_return_val_if_fail(sources!=NULL, FALSE);
	g_return_val_if_fail(server_key!=0, FALSE);
	g_return_val_if_fail(setme!=0, FALSE);

	for (it=sources->sources, end=it+sources->source_qty; !found && it!=end; ++it)
	{
		if (it->server_id == server_key)
		{
			found = TRUE;
			*setme = *it;
		}
	}

	return found;
}

gulong
message_sources_get_number(const MessageSources  * sources,
                              gint32                  server_id,
                              GQuark                  group)
{
	gulong retval = 0ul;
	MessageSource * it, * end;

	g_return_val_if_fail(sources!=NULL, 0ul);
	g_return_val_if_fail(server_id!=0, 0ul);
	g_return_val_if_fail(group!=0, 0ul);

	for (it=sources->sources, end=it+sources->source_qty; !retval && it!=end; ++it)
		if (it->server_id==server_id && it->group==group)
			retval = it->number;

	return retval;
}

