/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <errno.h>
#include <string.h>
#include <glib/gi18n-lib.h>

#include <pan/base/group.h>
#include <pan/base/article.h>
#include <pan/base/log.h>
#include <pan/base/newsrc.h>
#include <pan/base/newsrc-port.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/util-file.h>
#include <pan/base/db.h>
#include <pan/base/vserver.h>

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

/*****
******
*****/

static void import(sqlite3 *db, gint32 gsid, Newsrc *newsrc)
{
	sqlite3_stmt *stmt,*stmt2;
	gboolean read;
	guint32 uid;
	gint64 aid;
	Article *a;
	int sq_res;
	
	// NOTE: posible optimization by limiting query to newsrc range
	sq_res=sqlite3_prepare(db,"SELECT uid,article.id FROM g_s_a,article "
							"WHERE group_server_id=? AND article.id=article_id;"
							,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt,1,gsid);
	BEGIN_TRANSACTION(db,stmt2);
	sq_res=sqlite3_prepare(db,"UPDATE article SET read=? WHERE id=?;",-1,&stmt2,NULL);
	SQL_PREP_VALID(sq_res,db);
	while(sqlite3_step(stmt)==SQLITE_ROW)
	{
		uid=sqlite3_column_int(stmt,0);
		aid=sqlite3_column_int64(stmt,1);
		read=newsrc_is_article_read(newsrc,uid);
		a=article_get_by_id(aid);
		if(a)
			article_set_read(a,read);
		else
		{
			sqlite3_bind_int64(stmt2,1,aid);
			sqlite3_bind_int(stmt2,0,read);
			sqlite3_step(stmt2);
			sqlite3_reset(stmt2);
		}
	}
	sqlite3_finalize(stmt2);
	sqlite3_finalize(stmt);
	END_TRANSACTION(db,stmt);
}

void
newsrc_import (Server         * server,
               const PString  * filename_in,
               gboolean         subscribed_only)
{
	char * filename;
	char * body = NULL;
	gsize body_len = 0;
	GError * err;
	Newsrc *newsrc=newsrc_new(NULL,0,0);

	/* sanity clause */
	g_return_if_fail (server_is_valid (server));
	g_return_if_fail (pstring_is_set (filename_in));

	/* normalize the filename */
	filename = g_strndup (filename_in->str, filename_in->len);
	replace_gstr (&filename, pan_file_normalize (filename, NULL));
	filename_in = NULL;

	/* read the file into 'body'*/
	err = NULL;
	if (!g_file_get_contents (filename, &body, &body_len, &err))
	{
		log_add_va (LOG_ERROR, _("Error reading file \"%s\": %s"), filename, err->message);
		g_error_free (err);
	}
	else /* march through each line of newsrc... */
	{
		const char * pch = body;
		int groups_imported_qty = 0;
		GString * group_name = g_string_new(NULL);
		GString * range_string = g_string_new(NULL);
		GString * line = g_string_new(NULL);
		GPtrArray *sub_y=g_ptr_array_new();
		int groups_added = 0;
		sqlite3 *db=get_db();
		sqlite3_stmt *stmt;

		while (get_next_token_g_str(pch, '\n', &pch, line))
		{
			gboolean subscribed = FALSE;

			if (newsrc_parse_raw_line(line->str, group_name, &subscribed, range_string))
			{
				Group * group;
				int id,gsid,res;
				const PString group_name_pstr = pstring_shallow(group_name->str, group_name->len);

				/* maybe skip non-subscribed groups */
				if (subscribed_only && !subscribed)
					continue;

				/* get the group which this line is for. */
				res=sqlite3_prepare(db,"SELECT group_id,group_server.id FROM groups, group_server "
								"WHERE group.name=? AND group_id=groups.id AND server_id=? ;",
								-1,&stmt,NULL);
				SQL_PREP_VALID(res,db);
				sqlite3_bind_text(stmt,1,group_name_pstr.str,group_name_pstr.len,SQLITE_STATIC);
				sqlite3_bind_int(stmt,2,server->id);
				if G_LIKELY(sqlite3_step(stmt)==SQLITE_ROW)
				{
					id=sqlite3_column_int(stmt,0);
					gsid=sqlite3_column_int(stmt,1);
					group=group_new(id);
					/* since newsrc is based on uid we can only update read status
					   if we already have the uid */
					newsrc_init_from_range_str(newsrc,range_string->str);
					import(db,gsid,newsrc);
					group_update_counts(group);
				}else
				{
					/* not associated with server */
					++groups_added;
					sqlite3_finalize(stmt);
					res=sqlite3_prepare(db,"SELECT id FROM groups WHERE name=? ;",
									-1,&stmt,NULL);
					SQL_PREP_VALID(res,db);
					sqlite3_bind_text(stmt,1,group_name_pstr.str,group_name_pstr.len,SQLITE_STATIC);
					if G_LIKELY(sqlite3_step(stmt)==SQLITE_ROW)
					{
						id=sqlite3_column_int(stmt,0);
						group=group_new(id);
					}else
					{
						/* group not in db */
						sqlite3_finalize(stmt);
						res=sqlite3_prepare(db,"INSERT INTO groups(name,new) VALUES(?,1);",
										-1,&stmt,NULL);
						SQL_PREP_VALID(res,db);
						sqlite3_bind_text(stmt,1,group_name_pstr.str,group_name_pstr.len,SQLITE_STATIC);
						sqlite3_step(stmt);
						sqlite3_finalize(stmt);
						/* add group */
						res=sqlite3_prepare(db,"SELECT id FROM groups WHERE name=? ;",
										-1,&stmt,NULL);
						SQL_PREP_VALID(res,db);
						sqlite3_bind_text(stmt,1,group_name_pstr.str,group_name_pstr.len,SQLITE_STATIC);
						sqlite3_step(stmt);
						id=sqlite3_column_int(stmt,0);
						group=group_new(id);
					}
					/* associate group with server */
					sqlite3_finalize(stmt);
					res=sqlite3_prepare(db,"INSERT INTO group_server(group_id,server_id) VALUES(?,?);",
									-1,&stmt,NULL);
					SQL_PREP_VALID(res,db);
					sqlite3_bind_int(stmt,1,group->id);
					sqlite3_bind_int(stmt,2,server->id);
					sqlite3_step(stmt);
				}
				sqlite3_finalize(stmt);
				
				if(subscribed && !group_is_subscribed(group))
					g_ptr_array_add(sub_y,group);
				else
					g_object_unref(group);

				/* track how many groups imported */
				++groups_imported_qty;
			}
		}

		/* add groups to subscribed */
		if (sub_y->len)
			vserver_add_groups(1,sub_y);

		/* log our progress */
		log_add_va (LOG_INFO, _("Imported %d groups (%d new) from \"%s\""),
				      groups_imported_qty,
				      groups_added,
				      filename);

		g_string_free (line, TRUE);
		g_string_free (group_name, TRUE);
		g_string_free (range_string, TRUE);
		g_ptr_array_free (sub_y, TRUE);
		int x;
		for(x=0;x<sub_y->len;x++)
			g_object_unref(GROUP(sub_y->pdata[x]));
	}

	/* cleanup */
	g_object_unref(newsrc);
	g_free(filename);
	g_free(body);
}

/*****
******
*****/

void
newsrc_export (Server          * server,
               const PString   * filename_in,
               gboolean          subscribed_only)
{
	FILE * fp;
	char * filename_tmp;
	char * filename;

	/* sanity clause */
	g_return_if_fail (server_is_valid (server));
	g_return_if_fail (pstring_is_set (filename_in));

	filename = g_strndup (filename_in->str, filename_in->len);
	replace_gstr (&filename, pan_file_normalize (filename, NULL));
	filename_in = NULL;

	/* open a temp file to write to */
	filename_tmp = pan_file_make_temp (&fp);
	if (fp == NULL)
	{
		log_add_va (LOG_ERROR, _("Could not create temporary file: %s"), g_strerror(errno));
	}
	else
	{
		int i,sq_res;
		int groups_exported_qty = 0;
		int subbed_exported_qty = 0;
		gboolean ok;
		sqlite3 *db=get_db();
		sqlite3_stmt *stmt, *stmt2;
		GArray *groups=g_array_sized_new(FALSE,FALSE,sizeof(gint32),128);
		
		if(subscribed_only)
			sq_res=sqlite3_prepare(db,"SELECT group_id,name,subscribed,article_min,article_max "
									"FROM group_server, groups "
									"WHERE server_id=? AND groups.id=group_id "
									"AND groups.subscribed!=0 ORDER BY name;", -1, &stmt,NULL);
		else
			sq_res=sqlite3_prepare(db,"SELECT group_id,name,subscribed,article_min,article_max "
									"FROM group_server, groups "
									"WHERE server_id=? AND groups.id=group_id "
									"ORDER BY name;", -1, &stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		sqlite3_bind_int(stmt,1,server->id);
		
		sq_res=sqlite3_prepare(db,"SELECT uid FROM group_server,g_s_a,article "
								"WHERE server_id=? AND group_id=? "
								"AND group_server_id=group_server.id AND article.id=article_id "
								"AND article.read=1;"
								,-1,&stmt2,NULL);
		SQL_PREP_VALID(sq_res,db);
		
		/* walk through the groups */
		while(sqlite3_step(stmt)==SQLITE_ROW)
		{
			const char * name=sqlite3_column_text(stmt,1);
			gboolean subscribed=sqlite3_column_int(stmt,2)==0?FALSE:TRUE;
			gulong low=sqlite3_column_int(stmt,3), high=sqlite3_column_int(stmt,4);
			Newsrc *newsrc=newsrc_new(NULL,low,high);
			// newsrc_mark_range(newsrc,low,high,FALSE);
			
			i=sqlite3_column_int(stmt,0);
			
			sqlite3_bind_int(stmt2,1,server->id);
			sqlite3_bind_int(stmt2,2,i);
			while(sqlite3_step(stmt2)==SQLITE_ROW)
			{
				gulong uid=sqlite3_column_int(stmt2,0);
				//gboolean read=sqlite3_column_int(stmt2,1);
				newsrc_mark_article(newsrc,uid,TRUE);
			}
			sqlite3_reset(stmt2);
			
			char * str = newsrc_export_line(newsrc,name,subscribed);

			if (is_nonempty_string (str))
			{
				fprintf (fp, "%s\n", str);

				/* keep track of how many we've exported */
				++groups_exported_qty;
				if (subscribed)
					++subbed_exported_qty;
			}

			g_free (str);
			g_object_unref(newsrc);
		}
		sqlite3_finalize(stmt);
		sqlite3_finalize(stmt2);

		ok = !ferror(fp);
		fclose (fp);

		if (ok && pan_file_swap_datafile (filename_tmp, filename))
			log_add_va (LOG_INFO, _("Exported %d groups (%d subscribed) to \"%s\""),
					      groups_exported_qty,
					      subbed_exported_qty,
					      filename);

		/* cleanup */
		g_array_free (groups, TRUE);
	}

	/* cleanup */
	g_free (filename);
	g_free (filename_tmp);
}
