/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <glib/gi18n-lib.h>

#include <pan/base/base-prefs.h>
#include <pan/base/debug.h>
#include <pan/base/log.h>
#include <pan/base/newsrc-port.h>
#include <pan/base/pan-config.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/server.h>
#include <pan/base/serverlist.h>
#include <pan/base/util-file.h>
#include <pan/base/db.h>
#include <pan/base/group.h>
#include <pan/base/vserver.h>

#define SERVER_NAME_KEY "/Pan/Servers/Names"
static GSList * serverlist = NULL;
static Server * _active = NULL;

/**
***  Private Utilities
**/

static gint
server_name_compare (gconstpointer a, gconstpointer b)
{
	const Server* sa = (const Server*) a;
	const Server* sb = (const Server*) b;
	return g_strcasecmp (server_get_name (sa), server_get_name (sb));
}

/**
 * Load a list of servers from the user's ~/.gnome/Pan file
 * @return a GSList of Server's for each server found
 */
static GSList*
server_list_load(void)
{
	GSList * l = NULL;
	const gchar * p;
	sqlite3 *db;
	sqlite3_stmt *stmt;
	int sq_res;
	const char *sql="SELECT name, hostname, need_auth, username, password, newsrc,"
					"newsrc_fn, max_conn, port, last_gl_update, rank, id "
					"FROM server;";
	debug_enter("server_list_load");

	/* build all the servers in the list */
	db=get_db();
	sq_res=sqlite3_prepare(db,sql,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	while(sqlite3_step(stmt)==SQLITE_ROW)
	{
		Server *server=server_new();
		p=sqlite3_column_text(stmt,0);
		pstring_set(&server->name,p,strlen(p));
		p=sqlite3_column_text(stmt,1);
		pstring_set(&server->address,p,strlen(p));
		server->need_auth=sqlite3_column_int(stmt,2);
		p=sqlite3_column_text(stmt,3);
		if(p)
			pstring_set(&server->username,p,strlen(p));
		p=sqlite3_column_text(stmt,4);
		if(p)
			pstring_set(&server->password,p,strlen(p));
		server->use_newsrc=sqlite3_column_int(stmt,5);
		p=sqlite3_column_text(stmt,6);
		if(p)
			pstring_set(&server->newsrc_filename,p,strlen(p));
		server->max_connections=sqlite3_column_int(stmt,7);
		server->port=sqlite3_column_int(stmt,8);
		server->last_newgroup_list_time=sqlite3_column_int(stmt,9);
		server->rank=sqlite3_column_int(stmt,10);
		server->id=sqlite3_column_int(stmt,11);
		
		if(server_is_valid(server) && pstring_is_set(&server->name) && pstring_is_set(&server->address))
		{
			l = g_slist_insert_sorted(l, server, server_name_compare);
		}else
		{
			g_object_unref(G_OBJECT(server));
		}
	}
	sqlite3_finalize(stmt);

	debug_exit("server_list_load");
	return l;
}

static void serverlist_save_new_server(Server *srv)
{
	sqlite3 *db;
	sqlite3_stmt *stmt;
	int sq_res;
	const char *sql="INSERT INTO server(name, hostname, port, need_auth, username,"
					"password, newsrc, newsrc_fn, max_conn, last_gl_update, rank) "
					"VALUES(?,?,?,?,?,?,?,?,?,?,?);";
	
	debug_enter("serverlist_save_new_server");
	
	db=get_db();
	sq_res=sqlite3_prepare(db,sql,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_text(stmt,1,srv->name.str,srv->name.len,SQLITE_STATIC);
	sqlite3_bind_text(stmt,2,srv->address.str,srv->address.len,SQLITE_STATIC);
	sqlite3_bind_int(stmt,3,srv->port);
	sqlite3_bind_int(stmt,4,srv->need_auth);
	sqlite3_bind_text(stmt,5,srv->username.str,srv->username.len,SQLITE_STATIC);
	sqlite3_bind_text(stmt,6,srv->password.str,srv->password.len,SQLITE_STATIC);
	sqlite3_bind_int(stmt,7,srv->use_newsrc);
	sqlite3_bind_text(stmt,8,srv->newsrc_filename.str,srv->newsrc_filename.len,SQLITE_STATIC);
	sqlite3_bind_int(stmt,9,srv->max_connections);
	sqlite3_bind_int(stmt,10,srv->last_newgroup_list_time);
	sqlite3_bind_int(stmt,11,srv->rank);
	sq_res=sqlite3_step(stmt);
	if(sq_res!=SQLITE_OK && sq_res!=SQLITE_DONE)
	{
		g_warning("Error adding server: %s",sqlite3_errmsg(db));
	}
	sqlite3_finalize(stmt);
	
	sq_res=sqlite3_prepare(db,"SELECT id from server WHERE name=?;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_text(stmt,1,srv->name.str,srv->name.len,SQLITE_STATIC);
	sqlite3_step(stmt);
	srv->id=sqlite3_column_int(stmt,0);
	sqlite3_finalize(stmt);
	
	debug_exit("serverlist_save_new_server");
}

static gboolean inited = FALSE;

static void
server_init_servers(void)
{
	debug_enter("server_init_servers");

	if (!inited)
	{
		inited = TRUE;
		serverlist = server_list_load();
	}

	debug_exit("server_init_servers");
}

/**
 * serverlist_get_servers:
 * @fillme: GPtrArray
 *
 * Populates @fillme with the list of servers. Caller owns the array NOT the servers.
 */
void
serverlist_get_servers(GPtrArray * fillme)
{
	GSList * l;

	g_return_if_fail(fillme!=NULL);
	debug_enter("serverlist_get_servers");
	
	server_init_servers();
	if(serverlist)
	{
		//g_ptr_array_set_size(fillme, g_slist_length(serverlist));
		for (l=serverlist; l!=NULL; l=l->next)
			g_ptr_array_add(fillme, l->data);
	}
	
	debug_exit("serverlist_get_servers");
}

/**
 * serverlist_shutdown:
 *
 * Frees the internal list of serveres.
 */
void
serverlist_shutdown(void)
{
	debug_enter("serverlist_shutdown");
	GSList * l;

	/* save the updated server info */
	serverlist_save();

	/* close down the servers */
	l = serverlist;
	serverlist = NULL;
	inited=FALSE;
	g_slist_foreach (l, (GFunc)g_object_unref, NULL);
	g_slist_free(l);
	debug_exit("serverlist_shutdown");
}

/**
 * serverlist_get_named_server:
 * @servername: PString*
 *
 * Retreive the server given its name.
 *
 * Returns: A Server* or NULL
 */
Server*
serverlist_get_named_server(const PString* servername)
{
	GSList * l;
	Server * retval = NULL;
	debug_enter("server_get_by_name");

	g_return_val_if_fail(pstring_is_set(servername), NULL);

	server_init_servers();

	for (l=serverlist; retval==NULL && l!=NULL; l=l->next) {
		Server * s = SERVER(l->data);
		if (pstring_equal(servername, &s->name))
			retval = s;
	}

	debug_exit("server_get_by_name");
	return retval;
}

/**
 * serverlist_get_by_id:
 * @id: 
 *
 * Retreive the server given its id.
 *
 * Returns: A Server* or NULL
 */
Server*
serverlist_get_by_id(gint32 id)
{
	GSList * l;
	Server * retval = NULL;
	debug_enter("serverlist_get_by_id");

	server_init_servers();

	for (l=serverlist; retval==NULL && l!=NULL; l=l->next)
	{
		Server * s = SERVER(l->data);
		if (s->id==id)
		{
			retval = s;
			break;
		}
	}

	debug_exit("serverlist_get_by_id");
	return retval;
}

static Server*
get_default_server(void)
{
	GSList * l;
	gchar * name;
	Server * retval = NULL;

	g_return_val_if_fail(serverlist!=NULL, NULL);

	/* look in the gnome config for a default */
	name = pan_config_get_string("/Pan/State/Server", NULL);
	if (is_nonempty_string (name))
	{
		const PString pname = pstring_shallow(name, -1);
		retval = serverlist_get_named_server(&pname);
	}
	g_free(name);

	/* otherwise just use the first server */
	for (l=serverlist; retval==NULL && l!=NULL; l=l->next)
	{
		Server * s = SERVER(l->data);
		retval = s;
	}

	/* return what we've got */
	return retval;
}



/****
*****
*****  SERVER LOADING ON STARTUP
*****
****/

/**
 * serverlist_init:
 *
 */
void
serverlist_init(void)
{
	server_init_servers();
}

/**
 * serverlist_add_server:
 * @server: Server*
 * @err: GError**
 *
 * Adds a server to the list.
 */
void
serverlist_add_server(Server * server, GError ** err)
{
	g_return_if_fail(err==NULL || *err==NULL);
	debug_enter("serverlist_add_server");
	
	server_init_servers();

	if (!server_is_valid(server)) 
	{
		if (err != NULL)
			*err = g_error_new(g_quark_from_string("serverlist"), 1, _("Server is corrupt"));
		return;
	}

	if (serverlist_get_named_server(&server->name) != NULL) 
	{
		if (err != NULL)
			*err = g_error_new(g_quark_from_string("serverlist"), 2, _("A Server with that name already exists."));
		return;
	}

	serverlist = g_slist_insert_sorted(serverlist,
	                                    server,
	                                    server_name_compare);

		serverlist_save_new_server(server);
		
	debug_exit("serverlist_add_server");
}

/**
 * serverlist_remove_server:
 * @server: Server*
 *
 * Removes a server from the list.
 */
void
serverlist_remove_server(Server * server)
{
	g_return_if_fail(server_is_valid(server));
	debug_enter("serverlist_remove_server");

	serverlist = g_slist_remove(serverlist, server);
	/* might be best to just let this leak. */
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	int sq_res;
	
	BEGIN_TRANSACTION(db,stmt);
	
	/* delete server */
	sq_res=sqlite3_prepare(db,"DELETE FROM server WHERE id=?;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt,1,server->id);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	
	/* get list of groups with articles from this server */
	GArray *groups=g_array_new(FALSE,FALSE,sizeof(gint32));
	sq_res=sqlite3_prepare(db,"SELECT group_id FROM group_server,g_s_a "
								"WHERE server_id=? AND group_server_id=group_server.id;"
								,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	while(sqlite3_step(stmt)==SQLITE_ROW)
	{
		gint32 gid=sqlite3_column_int(stmt,0);
		g_array_append_val(groups,gid);
	}
	sqlite3_bind_int(stmt,1,server->id);
	sqlite3_finalize(stmt);
	
	/* delete articles from g_s_a */
	sq_res=sqlite3_prepare(db,"DELETE FROM g_s_a WHERE group_server_id IN "
								"(SELECT id FROM group_server WHERE server_id=?);"
								,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt,1,server->id);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	/* delete articles from group_article */
	sq_res=sqlite3_prepare(db,"DELETE FROM group_article WHERE article_id NOT IN "
								"(SELECT DISTINCT article_id FROM g_s_a);",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	/* update group article counts & rethread group*/
	int i;
	for(i=0;i<groups->len;i++)
	{
		Group *g=group_new(g_array_index(groups,gint32,i));
		group_update_counts(g);
		thread_group(g->id);
		g_object_unref(g);
	}
	g_array_free(groups,TRUE);
	
	/* delete groups from group_server */
	sq_res=sqlite3_prepare(db,"DELETE FROM group_server WHERE server_id=?;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt,1,server->id);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	
	END_TRANSACTION(db,stmt);
	
	/* delete groups from vservers */
	sq_res=sqlite3_prepare(db,"SELECT group_id,vserver FROM vserver_group "
								"WHERE group_id NOT IN "
								"(SELECT DISTINCT group_id FROM group_server) "
								"GROUP BY vserver;"
								,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	gint32 vid=0;
	GPtrArray *groupsp=g_ptr_array_new();
	while(sqlite3_step(stmt)==SQLITE_ROW)
	{
		gint32 gid,nvid;
		gid=sqlite3_column_int(stmt,0);
		nvid=sqlite3_column_int(stmt,1);
		if(vid!=nvid && groupsp->len!=0)
		{
			vserver_remove_groups(vid,groupsp);
			for(i=0;i<groupsp->len;i++)
				g_object_unref(g_ptr_array_index(groupsp,i));
			g_ptr_array_set_size(groupsp,0);
		}
		vid=nvid;
		g_ptr_array_add(groupsp,group_new(gid));
	}
	if(groupsp->len!=0)
	{
		vserver_remove_groups(vid,groupsp);
		for(i=0;i<groupsp->len;i++)
			g_object_unref(g_ptr_array_index(groupsp,i));
	}
	g_ptr_array_free(groupsp,TRUE);
	sqlite3_finalize(stmt);
	
	g_object_unref(G_OBJECT(server));
	debug_exit("serverlist_remove_server");
}

/**
 * serverlist_save:
 *
 * Updates the db.
 */
void
serverlist_save(void)
{
	/* out with the old */
	if(0)
	{
		char * p = pan_config_get_string(SERVER_NAME_KEY, NULL);
		if (p != NULL) {
			GString * tmp = g_string_new(NULL);
			PString pstr = PSTRING_INIT;
			const char * walk = p;
			while (get_next_token_pstring(walk, ',', &walk, &pstr))
			{
				if (pstr.len)
				{
					g_string_assign(tmp, "/Pan/");
					g_string_append_len(tmp, pstr.str, pstr.len);
					pan_config_clean_section(tmp->str);
				}
			}
			g_string_free(tmp, TRUE);
			g_free(p);
		}
	}

	/* in with the new */
	{
		GSList * l = NULL;
		GString * servers = g_string_new(NULL);
		sqlite3 *db;
		sqlite3_stmt *stmt;
		int sq_res;
		const char *sql="UPDATE server SET name=? ,hostname=? ,port=? ,need_auth=? "
						",username=? ,password=? ,newsrc=? ,newsrc_fn=? ,max_conn=? "
						",last_gl_update=? ,rank=? WHERE id=?";

		db=get_db();
		
		BEGIN_TRANSACTION(db,stmt);
		
		sq_res=sqlite3_prepare(db,sql,-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		for (l=serverlist; l!=NULL; l=l->next)
		{
			Server * s = SERVER(l->data);

			if (server_is_valid (s) && pstring_is_set(&s->name))
			{
				if(s->dirty)
				{
					sqlite3_bind_text(stmt,1,s->name.str,s->name.len,SQLITE_STATIC);
					sqlite3_bind_text(stmt,2,s->address.str,s->address.len,SQLITE_STATIC);
					sqlite3_bind_int(stmt,3,s->port);
					sqlite3_bind_int(stmt,4,s->need_auth);
					sqlite3_bind_text(stmt,5,s->username.str,s->username.len,SQLITE_STATIC);
					sqlite3_bind_text(stmt,6,s->password.str,s->password.len,SQLITE_STATIC);
					sqlite3_bind_int(stmt,7,s->use_newsrc);
					sqlite3_bind_text(stmt,8,s->newsrc_filename.str,s->newsrc_filename.len,SQLITE_STATIC);
					sqlite3_bind_int(stmt,9,s->max_connections);
					sqlite3_bind_int(stmt,10,s->last_newgroup_list_time);
					sqlite3_bind_int(stmt,11,s->rank);
					sqlite3_bind_int(stmt,12,s->id);
					sqlite3_step(stmt);
					sqlite3_reset(stmt);
				}
				
				/* char section_name[PATH_MAX];
				g_snprintf (section_name, sizeof(section_name), "/Pan/%*.*s", s->name.len, s->name.len, s->name.str);

				pan_config_push_prefix (section_name);
				pan_config_set_string ("Address", s->address.str);
				pan_config_set_int ("Port", s->port);
				pan_config_set_int ("Max_Connections", s->max_connections);
				pan_config_set_ulong ("Last_Grouplist_Update", s->last_newgroup_list_time);
				pan_config_set_bool ("Use_Newsrc", s->use_newsrc);
				pan_config_set_string ("Newsrc_Filename", s->newsrc_filename.str);
				pan_config_set_bool ("Auth_Required", s->need_auth);
				pan_config_set_string ("Username", s->username.str);
				pan_config_set_string ("Password", s->password.str);
				pan_config_pop_prefix ();

				if (servers->len)
					g_string_append_c (servers, ',');
				g_string_append_len (servers, s->name.str, s->name.len); */
			}
		}
		sqlite3_finalize(stmt);
		
		END_TRANSACTION(db,stmt);


		//pan_config_set_string (SERVER_NAME_KEY, servers->str);
		g_string_free(servers, TRUE);
	}

	//pan_config_sync ();
}

Server *serverlist_get_server(GQuark q)
{
	GSList *s=serverlist;
	Server *ret=NULL;
	
	while(s && !ret)
	{
		Server *x=SERVER(s->data);
		if(x->quark==q)
			ret=x;
		s=s->next;
	}
	return ret;
}

