/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2004  K. Haley <haleykd@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <config.h>
#include <glib.h>
#include <glib/gi18n-lib.h>
#include <pan/base/vserver.h>
#include <pan/base/db.h>
#include <pan/base/debug.h>
#include <pan/base/group-private.h>
#include <pan/base/pan-base-signals.h>

static GPtrArray *list=NULL;
/**
 * vserver_init:
 *
 * Initialize the list of vservers from the db.
 **/
void vserver_init()
{
	debug_enter("vserver_init");
	if(list)
	{
		debug_exit("vserver_init");
		return;
	}
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	vserver *vs;
	int qty, sq_res;
	
	sq_res=sqlite3_prepare(db,"SELECT count(*) FROM vserver;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_step(stmt);
	qty=sqlite3_column_int(stmt,0);
	sqlite3_finalize(stmt);
	list=g_ptr_array_sized_new(qty+4);
	vs=g_new(vserver,1);
	vs->id=0;
	vs->name=g_strdup(_("All Groups"));
	g_ptr_array_add(list,vs);
	vs=g_new(vserver,1);
	vs->id=1;
	vs->name=g_strdup(_("Subscribed"));
	g_ptr_array_add(list,vs);
	vs=g_new(vserver,1);
	vs->id=2;
	vs->name=g_strdup(_("New Groups"));
	g_ptr_array_add(list,vs);
	vs=g_new(vserver,1);
	vs->id=3;
	vs->name=g_strdup(_("Folders"));
	g_ptr_array_add(list,vs);
	
	if G_UNLIKELY(qty==0)
	{
		debug_exit("vserver_init");
		return;
	}
	
	sq_res=sqlite3_prepare(db,"SELECT name,id FROM vserver;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	while(sqlite3_step(stmt)==SQLITE_ROW)
	{
		vs=g_new(vserver,1);
		vs->id=4+sqlite3_column_int(stmt,1);
		vs->name=g_strdup(sqlite3_column_text(stmt,0));
		g_ptr_array_add(list,vs);
	}
	sqlite3_finalize(stmt);
	
	debug_exit("vserver_init");
};

/**
 * vserver_shutdown:
 *
 * Frees the list of vservers.
 **/
void vserver_shutdown()
{
	int i;
	vserver *vs;
	
	debug_enter("vserver_shudown");
	
	for(i=0;i<list->len;i++)
	{
		vs=g_ptr_array_index(list,i);
		g_free((char*)vs->name);
		g_free(vs);
	}
	g_ptr_array_free(list,TRUE);
	debug_exit("vserver_shutdown");
}

/**
 * vserver_add:
 * @name: name of new vserver
 *
 * Creates a new vserver in the db called @name.
 *
 * Return: true if vserver was added, false if it already exists.
 **/
gboolean vserver_add(const char *name)
{
	g_return_val_if_fail(name!=NULL,FALSE);
	g_return_val_if_fail(name[0]!=0,FALSE);
	
	debug_enter("vserver_add");
	
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	vserver *vs;
	int sq_res;
	gboolean ret=FALSE;
	
	sq_res=sqlite3_prepare(db,"INSERT INTO vserver(name) VALUES(?);",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_text(stmt,1,name,-1,SQLITE_STATIC);
	if(sqlite3_step(stmt)==SQLITE_DONE)
	{
		vs=g_new(vserver,1);
		vs->id=4+sqlite3_last_insert_rowid(db);
		vs->name=g_strdup(name);
		g_ptr_array_add(list,vs);
		pan_base_signals_vserver_added(pan_base_signals_get(),name,vs->id);
		ret=TRUE;
	}
	sqlite3_finalize(stmt);
	
	debug_exit("vserver_add");
	return ret;
}

/**
 * vserver_remove:
 * @id: id of vserver to remove
 *
 * Removes a vserver from the db and decrements the subscribed count
 * of all groups in the vserver.
 * The four default vservers ( All, Subscribed, New, Folders) can't be removed.
 **/
void vserver_remove(guint32 id)
{
	g_return_if_fail(id>3);
	
	debug_enter("vserver_remove");
	
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt;
	guint32 vsid=id-4;
	int sq_res;
	GArray *groups=g_array_new(FALSE,FALSE,sizeof(gint32));
	
	BEGIN_TRANSACTION(db,stmt);
	
	const char *sql_1="UPDATE groups SET subscribed=subscribed-1 "
						"WHERE groups.id IN (SELECT group_id FROM vserver_group "
						"WHERE vserver=? );";
	sq_res=sqlite3_prepare(db,sql_1,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt,1,vsid);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	
	const char *sql_2="SELECT groups.id FROM groups,vserver_group "
						"WHERE vserver=? AND groups.id=group_id AND subscribed=0 ;";
	sq_res=sqlite3_prepare(db,sql_2,-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt,1,vsid);
	while(sqlite3_step(stmt)==SQLITE_ROW)
	{
		int i=sqlite3_column_int(stmt,0);
		g_array_append_val(groups,i);
	}
	sqlite3_finalize(stmt);
	
	sq_res=sqlite3_prepare(db,"DELETE FROM vserver_group WHERE vserver=? ;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt,1,vsid);
	sqlite3_step(stmt);
	sqlite3_finalize(stmt);
	
	END_TRANSACTION(db,stmt);
	
	if(groups->len>0)
		pan_base_signals_groups_changed(pan_base_signals_get(),(const Group**)groups->data,groups->len);
	g_array_free(groups,TRUE);
	
	int pos;
	for(pos=4;pos<list->len;pos++)
	{
		if(((vserver*)g_ptr_array_index(list,pos))->id==id)
		{
			g_ptr_array_remove_index(list,pos);
			pan_base_signals_vserver_removed(pan_base_signals_get(),id);
			break;
		}
	}
	
	debug_exit("vserver_remove");
}

/**
 * vserver_get_list:
 *
 * Retreive the list of vservers.
 *
 * Returns: const GPtrArray of vservers
 */
const GPtrArray* vserver_get_list()
{
	debug_enter("vserver_get_list");
	vserver_init();
	debug_exit("vserver_get_list");
	return list;
}

static void do_get(sqlite3_stmt *stmt, GPtrArray *groups)
{
	Group *group;
	gulong x;
	int cnt=0;
//	GTimer *timer=g_timer_new();
	while(sqlite3_step(stmt)==SQLITE_ROW)
	{
		group=group_new(sqlite3_column_int(stmt,0));
		g_ptr_array_add(groups,group);
/* 		if(++cnt==50)
		{
			cnt=0;
			if(g_timer_elapsed(timer,&x)>.05)
			{
				g_main_context_iteration(NULL,FALSE);
				g_timer_start(timer);
			}
		} */
	}
//	g_timer_destroy(timer);
}

static GPtrArray* get_folders(sqlite3 *db)
{
	GPtrArray *ret=NULL;
	sqlite3_stmt *stmt;
	gint qty;
	int sq_res;
	
	sq_res=sqlite3_prepare(db,"SELECT count(*) FROM groups WHERE folder=1;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_step(stmt);
	qty=sqlite3_column_int(stmt,0);
	sqlite3_finalize(stmt);
	ret=g_ptr_array_sized_new(qty);
	if(qty!=0)
	{
		Group *group;
		sq_res=sqlite3_prepare(db,"SELECT id FROM groups WHERE folder=1 ORDER BY name;",-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		do_get(stmt,ret);
		sqlite3_finalize(stmt);
	}
	return ret;
}

static GPtrArray* get_sub(sqlite3 *db)
{
	GPtrArray *ret=NULL;
	sqlite3_stmt *stmt;
	int sq_res;
	gint qty;
	
	sq_res=sqlite3_prepare(db,"SELECT count(*) FROM groups WHERE subscribed!=0;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_step(stmt);
	qty=sqlite3_column_int(stmt,0);
	sqlite3_finalize(stmt);
	ret=g_ptr_array_sized_new(qty);
	if(qty!=0)
	{
		Group *group;
		sq_res=sqlite3_prepare(db,"SELECT id FROM groups WHERE subscribed!=0 ORDER BY name;",-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		do_get(stmt,ret);
		sqlite3_finalize(stmt);
	}
	return ret;
}

static GPtrArray* get_all(sqlite3 *db)
{
	GPtrArray *ret=NULL;
	sqlite3_stmt *stmt;
	gint qty;
	int sq_res;
	
	sq_res=sqlite3_prepare(db,"SELECT count(*) FROM groups;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_step(stmt);
	qty=sqlite3_column_int(stmt,0);
	sqlite3_finalize(stmt);
	ret=g_ptr_array_sized_new(qty);
	if(qty!=0)
	{
		Group *group;
		sq_res=sqlite3_prepare(db,"SELECT DISTINCT group_id FROM group_server,groups "
									"WHERE server_id NOTNULL AND groups.id=group_id ORDER BY name;",-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		do_get(stmt,ret);
		sqlite3_finalize(stmt);
	}
	return ret;
}

static GPtrArray* get_new(sqlite3 *db)
{
	GPtrArray *ret=NULL;
	sqlite3_stmt *stmt;
	gint qty;
	int sq_res;
	
	sq_res=sqlite3_prepare(db,"SELECT count(*) FROM groups WHERE new=1;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_step(stmt);
	qty=sqlite3_column_int(stmt,0);
	sqlite3_finalize(stmt);
	ret=g_ptr_array_sized_new(qty);
	if(qty!=0)
	{
		Group *group;
		sq_res=sqlite3_prepare(db,"SELECT DISTINCT group_id FROM group_server,groups "
									"WHERE new=1 AND groups.id=group_id ORDER BY name;",-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		do_get(stmt,ret);
		sqlite3_finalize(stmt);
	}
	return ret;
}

static GPtrArray* get_groups(sqlite3 *db, guint32 id)
{
	GPtrArray *ret=NULL;
	sqlite3_stmt *stmt;
	gint qty;
	int sq_res;
	
	sq_res=sqlite3_prepare(db,"SELECT count(*) FROM vserver_group WHERE vserver=?;",-1,&stmt,NULL);
	SQL_PREP_VALID(sq_res,db);
	sqlite3_bind_int(stmt,1,id);
	sqlite3_step(stmt);
	qty=sqlite3_column_int(stmt,0);
	sqlite3_finalize(stmt);
	ret=g_ptr_array_sized_new(qty);
	if(qty!=0)
	{
		Group *group;
		sq_res=sqlite3_prepare(db,"SELECT group_id FROM vserver_group,groups "
									"WHERE vserver=? AND groups.id=group_id ORDER BY name;",-1,&stmt,NULL);
		SQL_PREP_VALID(sq_res,db);
		sqlite3_bind_int(stmt,1,id);
		do_get(stmt,ret);
		sqlite3_finalize(stmt);
	}
	return ret;
}

/**
 * vserver_get_groups:
 * @id: vserver to get groups from
 *
 * Retrieve the list of groups in the vserver.
 *
 * Return: GPtrArray of Group objects. Caller owned.
 */
GPtrArray* vserver_get_groups(guint32 id)
{
	debug_enter("vserver_get_groups");
	
	sqlite3 *db=get_db();
	GPtrArray *ret;
	
	switch(id)
	{
		case 0: ret=get_all(db); break;
		case 1: ret=get_sub(db); break;
		case 2: ret=get_new(db); break;
		case 3: ret=get_folders(db); break;
		default: ret=get_groups(db,id-4); break;
	}
	
	debug_exit("vserver_get_groups");
	return ret;
}

/**
 * vserver_add_groups:
 * @id: id of vserver
 * @groups: GPtrArray of Group to be added
 *
 * Adds the list of groups the the vserver incrementing subscribed.
 * DOES NOT work on ALL, NEW or FOLDERS.
 */
void vserver_add_groups(guint32 id, const GPtrArray *groups)
{
	g_return_if_fail(id==1 || id>3);
	g_return_if_fail(groups!=NULL);
	g_return_if_fail(groups->len>0);
	
	debug_enter("vserver_add_groups");
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt_1,*stmt_2;
	Group *g;
	int i, sq_res;
	
	BEGIN_TRANSACTION(db,stmt_1);
	
	if(id==1) /* all subscribed */
	{
		sq_res=sqlite3_prepare(db,"UPDATE groups SET subscribed=1 WHERE id=?;",-1,&stmt_1,NULL);
		SQL_PREP_VALID(sq_res,db);
		for(i=0;i<groups->len;i++)
		{
			g=(Group*)g_ptr_array_index(groups,i);
			if(!group_is_folder(g) && !group_is_subscribed(g))
			{
				sqlite3_bind_int(stmt_1,1,g->id);
				sqlite3_step(stmt_1);
				sqlite3_reset(stmt_1);
				g->_priv->subscribed=1;
			}
		}
		sqlite3_finalize(stmt_1);
	}else
	{
		guint32 vsid=id-4;
		sq_res=sqlite3_prepare(db,"INSERT INTO vserver_group(vserver,group_id) VALUES(?,?);",-1,&stmt_2,NULL);
		SQL_PREP_VALID(sq_res,db);
		for(i=0;i<groups->len;i++)
		{
			g=(Group*)g_ptr_array_index(groups,i);
			sqlite3_bind_int(stmt_2,1,vsid);
			sqlite3_bind_int(stmt_2,2,g->id);
			sqlite3_step(stmt_2);
			sqlite3_reset(stmt_2);
			g->_priv->subscribed=1;
		}
		sqlite3_finalize(stmt_2);
	}

	END_TRANSACTION(db,stmt_1);

	pan_base_signals_groups_changed(pan_base_signals_get(),(const Group**)groups->pdata,groups->len);
	debug_exit("vserver_add_groups");
}

/**
 * vserver_remove_groups:
 * @id: vserver id
 * @groups: GPtrArray of Group
 *
 * Remove the specified groups from the vserver and decrement subscribed. The
 * SUBSCRIBED vserver is a special case, it unsubscribes the group from all extra
 * vservers and sets subscribed to 0.
 * DOES NOT WORK on ALL, NEW, or FOLDERS.
 * Uses TRANSACTION.
 */
void vserver_remove_groups(guint32 id, const GPtrArray *groups)
{
	g_return_if_fail(id==1 || id>3);
	g_return_if_fail(groups!=NULL);
	g_return_if_fail(groups->len>0);

	debug_enter("vserver_remove_groups");
	sqlite3 *db=get_db();
	sqlite3_stmt *stmt_1,*stmt_2;
	Group *g;
	int i, sq_res;
	
	BEGIN_TRANSACTION(db,stmt_1);

	if(id==1) /* all subscribed */
	{
		sq_res=sqlite3_prepare(db,"DELETE FROM vserver_group WHERE group_id=?;",-1,&stmt_1,NULL);
		SQL_PREP_VALID(sq_res,db);
		sq_res=sqlite3_prepare(db,"UPDATE groups SET subscribed=0 WHERE id=?;",-1,&stmt_2,NULL);
		SQL_PREP_VALID(sq_res,db);
		for(i=0;i<groups->len;i++)
		{
			g=g_ptr_array_index(groups,i);

			sqlite3_bind_int(stmt_1,1,g->id);
			sqlite3_step(stmt_1);
			sqlite3_reset(stmt_1);

			sqlite3_bind_int(stmt_2,1,g->id);
			sqlite3_step(stmt_2);
			sqlite3_reset(stmt_2);
		}
		sqlite3_finalize(stmt_1);
		sqlite3_finalize(stmt_2);
	}else
	{
		guint32 vsid=id-4;
		sq_res=sqlite3_prepare(db,"DELETE FROM vserver_group WHERE vserver=? AND group_id=?;",-1,&stmt_1,NULL);
		SQL_PREP_VALID(sq_res,db);
		sq_res=sqlite3_prepare(db,"UPDATE groups SET subscribed=subscribed-1 WHERE id=?;",-1,&stmt_2,NULL);
		SQL_PREP_VALID(sq_res,db);
		for(i=0;i<groups->len;i++)
		{
			g=g_ptr_array_index(groups,i);
			sqlite3_bind_int(stmt_1,1,vsid);
			sqlite3_bind_int(stmt_1,2,g->id);
			sqlite3_step(stmt_1);
			sqlite3_reset(stmt_1);
			if(!group_is_folder(g))
			{
				sqlite3_bind_int(stmt_2,1,g->id);
				sqlite3_step(stmt_1);
				sqlite3_reset(stmt_1);
			}
		}
		sqlite3_finalize(stmt_1);
		sqlite3_finalize(stmt_2);
	}

	END_TRANSACTION(db,stmt_1);

	pan_base_signals_groups_changed(pan_base_signals_get(),(const Group**)groups->pdata,groups->len);
	debug_exit("vserver_remove_groups");
}

