/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <glib.h>
#include <pan/base/debug.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/serverlist.h>
#include <pan/flagset.h>
#include <pan/queue.h>
#include <pan/task-bodies.h>
#include <pan/base/pan-base-signals.h>

static GHashTable * _id_to_mid = NULL;

static void expired(PanBaseSignals *unused, GPtrArray *articles)
{
	int i;
	for(i=0;i<articles->len;++i)
	{
		Article *a=ARTICLE(g_ptr_array_index(articles,i));
		g_hash_table_remove(_id_to_mid,a);
	}
}

void
flagset_init(void)
{
	_id_to_mid = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, g_object_unref);
	g_signal_connect(pan_base_signals_get(),"articles-deleted",(GCallback)expired,NULL);
}

gboolean
flagset_has_article(const Article * article)
{
	g_return_val_if_fail(IS_ARTICLE(article), FALSE);

	return g_hash_table_lookup(_id_to_mid, article) != NULL;
}

void
flagset_add_articles(const GPtrArray     * articles)
{
	int i;
	Article *a;

	/* sanity clause */
	g_return_if_fail(articles!=NULL && articles->len>=1);

	/* add these message-ids to the set */
	for (i=0; i<articles->len; ++i)
	{
		a=ARTICLE(g_ptr_array_index(articles,i));
		g_object_ref(a);
		article_set_flagged(a,TRUE);
		g_hash_table_replace(_id_to_mid, a, a);
	}
}


void
flagset_remove_articles(const GPtrArray     * articles)
{
	int i;
	Article *a;

	/* sanity clause */
	g_return_if_fail(articles!=NULL && articles->len>=1);

	/* add these message-ids to the set */
	for (i=0; i<articles->len; ++i)
	{
		a=ARTICLE(g_ptr_array_index(articles,i));
		g_object_ref(a);
		article_set_flagged(a,FALSE);
		g_hash_table_remove(_id_to_mid, a);
	}
}


void
flagset_flush(void)
{
	/* get an array of the identifiers and ref them so they
	 * can be used after the hash is destroyed */
	GPtrArray * mids = g_ptr_array_new();
	pan_hash_to_ptr_array(_id_to_mid, mids);
	pan_g_ptr_array_foreach(mids, (GFunc)g_object_ref, NULL);

	/* clear the flagset's hash table */
	g_hash_table_destroy(_id_to_mid);
	_id_to_mid = g_hash_table_new_full(pstring_hash, pstring_equal, NULL, g_object_unref);

	/* if any messages were flagged, queue them */
	if (mids->len)
		queue_add(TASK (task_bodies_new((Article**)mids->pdata, mids->len)));

	/* cleanup */
	pan_g_ptr_array_foreach(mids, (GFunc)g_object_unref, NULL);
	g_ptr_array_free(mids, TRUE);
}
