
/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <gtk/gtkstyle.h>
#include <glib/gi18n-lib.h>

#include <pan/base/article.h>
#include <pan/base/debug.h>
#include <pan/base/log.h>
#include <pan/base/pan-glib-extensions.h>

#include <pan/xpm/pan-pixbufs.h>

#include <pan/filters/filter-score.h>

#include <pan/header-pane-renderer.h>
#include <pan/articlelist.h>
#include <pan/article-toolbar.h>
#include <pan/prefs.h>
#include <pan/queue.h>
#include <pan/util.h>
#include <pan/globals.h>

typedef struct
{
	const guint8 * pixbuf_txt;
	GdkPixbuf * pixbuf;
	GdkPixmap * pixmap;
	GdkBitmap * mask;
}
Icon;

enum
{
	ICON_READ,
	ICON_UNREAD,
	ICON_NEW,
	ICON_COMPLETE,
	ICON_INCOMPLETE,
	ICON_CACHED,
	ICON_BINARY_DECODED,
	ICON_QUEUED,
	ICON_ERROR,
	ICON_FLAGGED_FOR_DL,
	ICON_NEWSGROUP,
	ICON_QTY
};

static Icon _icons[ICON_QTY] = {
	{ icon_article_read, NULL, NULL, NULL },
	{ icon_article_unread, NULL, NULL, NULL },
	{ icon_article_new, NULL, NULL, NULL },
	{ icon_binary_complete, NULL, NULL, NULL },
	{ icon_binary_incomplete, NULL, NULL, NULL },
	{ icon_disk, NULL, NULL, NULL },
	{ icon_binary, NULL, NULL, NULL },
	{ icon_bluecheck, NULL, NULL, NULL },
	{ icon_x, NULL, NULL, NULL },
	{ icon_blue_flag, NULL, NULL, NULL },
	{ icon_newsgroup, NULL, NULL, NULL }
};

static GtkStyle * normal_style[2]                = { NULL, NULL };
static GtkStyle * read_style[2]                  = { NULL, NULL };
static GtkStyle * score_style[SCORE_COLOR_QTY][2] = {
	{ NULL, NULL },
	{ NULL, NULL },
	{ NULL, NULL },
	{ NULL, NULL },
	{ NULL, NULL }
};

GdkColor read_color;
GdkColor unread_color;
GdkColor score_color[SCORE_COLOR_QTY][2];

static PangoFontDescription * normal_pfd = NULL;
static PangoFontDescription * new_replies_pfd = NULL;

/***
****
***/

static void render_action(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static void render_state(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static void render_subject(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static void render_lines(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static void render_date(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static void render_author(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
static void render_score(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data);

//struct _HeaderPaneRenderer
//{
	GtkWidget * main_window;
	GtkTreeView * tree;
	gboolean rebuild_style;
	int score_column_index;
	int subject_column_index;
//};

// callbacks
static void style_set_cb(GtkWidget * w, GtkStyle * style, gpointer user_data)
{
	static gboolean nested = FALSE;

	/* _ The `nested' check is to prevent recursion from
	 *   articlelist_reset_style_nolock().
	 * _ The GTK_WIDGET_REALIZED() check is to prevent us from
	 *   refreshing the header pane every time the user toggles
	 *   tabbed pane layout on/off -- reparenting the header pane
	 *   widget triggers a style_set signal. (#112812)
	 */
	if (GTK_WIDGET_REALIZED(w) && !nested)
	{
		nested = TRUE;
		header_pane_renderer_reset_style_nolock();
		nested = FALSE;
	}
}


static void header_pane_renderer_free(gpointer renderer_gpointer)
{
	g_free (renderer_gpointer);
}

// resource cleanup
static void tree_destroy(GtkObject *obj, gpointer data)
{
	int i;
	for(i=0;i<ICON_QTY;++i)
	{
		g_object_unref(_icons[i].pixbuf);
		_icons[i].pixbuf=NULL;
	}
	for(i=0;i<2;++i)
	{
		gtk_style_unref(normal_style[i]);
		gtk_style_unref(read_style[i]);
		normal_style[i]=NULL;
		read_style[i]=NULL;
	}
	for(i=0;i<SCORE_COLOR_QTY;++i)
	{
		gtk_style_unref(score_style[i][0]);
		gtk_style_unref(score_style[i][1]);
		score_style[i][0]=NULL;
		score_style[i][1]=NULL;
	}
	pango_font_description_free(normal_pfd);
	pango_font_description_free(new_replies_pfd);
	normal_pfd=NULL;
	new_replies_pfd=NULL;
	main_window=NULL;
	tree=NULL;
}

GtkWidget* header_pane_renderer_new(void)
{
	GtkWidget * w;
	GtkTreeView * treev;
	GtkWidget * s_window;
	GtkWidget * vbox;
	//HeaderPaneRenderer * retval = g_new0(HeaderPaneRenderer, 1);

	if(tree)
		return gtk_widget_get_ancestor(GTK_WIDGET(tree),GTK_TYPE_VBOX);
	
	vbox = gtk_vbox_new(FALSE, GUI_PAD_SMALL);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), GUI_PAD_SMALL);

	/* filter */
	gtk_box_pack_start(GTK_BOX(vbox), article_toolbar_new(), FALSE, FALSE, 0);

	/* create the widget */
	Pan.article_tree=GTK_TREE_VIEW(gtk_tree_view_new());
	treev=Pan.article_tree;
	
	main_window = vbox;
	tree = treev;
	rebuild_style = TRUE;
	
	/* init the icons */
	if(_icons[0].pixbuf==NULL)
	{
		guint i;
		//GdkColormap * cmap = gdk_colormap_get_system ();

		for (i=0; i<ICON_QTY; ++i) {
			_icons[i].pixbuf = gdk_pixbuf_new_from_inline(-1, _icons[i].pixbuf_txt, FALSE, NULL);
			/* gdk_pixbuf_render_pixmap_and_mask_for_colormap (_icons[i].pixbuf, cmap,
									&_icons[i].pixmap,
									&_icons[i].mask,
									128); */
		}
	}

	/* init the fonts */
	/* fonts */
	if(normal_pfd==NULL)
	{
		if (header_pane_custom_font_enabled) {
				normal_pfd = pango_font_description_from_string(header_pane_custom_font);
				new_replies_pfd = pango_font_description_copy(normal_pfd);
		} else {
				GtkStyle * style = gtk_widget_get_default_style();
				normal_pfd = pango_font_description_copy(style->font_desc);
				new_replies_pfd = pango_font_description_copy(style->font_desc);
		}
		pango_font_description_set_weight(new_replies_pfd, PANGO_WEIGHT_BOLD);
		gtk_widget_modify_font(GTK_WIDGET(tree), new_replies_pfd);
	}
	/* ui settings */
	//gtk_ctree_set_line_style (ctree, GTK_CTREE_LINES_DOTTED);
	gtk_tree_selection_set_mode(gtk_tree_view_get_selection(tree), GTK_SELECTION_MULTIPLE);

	/* columns */
	{
		GtkCellRenderer *cell;
		GtkTreeViewColumn *col;
		
		cell=gtk_cell_renderer_pixbuf_new();
		col=gtk_tree_view_column_new_with_attributes(" ",cell,NULL);
		gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED);
		gtk_tree_view_column_set_fixed_width(col,16);
		gtk_tree_view_column_set_resizable(col,FALSE);
		gtk_tree_view_column_set_cell_data_func(col,cell,render_action,NULL,NULL);
		gtk_tree_view_append_column(tree,col);
		
		cell=gtk_cell_renderer_pixbuf_new();
		col=gtk_tree_view_column_new_with_attributes(" ",cell,NULL);
		gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED);
		gtk_tree_view_column_set_resizable(col,FALSE);
		gtk_tree_view_column_set_fixed_width(col,16);
		gtk_tree_view_column_set_cell_data_func(col,cell,render_state,NULL,NULL);
		gtk_tree_view_append_column(tree,col);
		
		cell=gtk_cell_renderer_text_new();
		col=gtk_tree_view_column_new_with_attributes(_("Subject"),cell,NULL);
		gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED);
		gtk_tree_view_column_set_fixed_width(col,220);
		gtk_tree_view_column_set_resizable(col,TRUE);
		gtk_tree_view_column_set_expand(col,TRUE);
		gtk_tree_view_column_set_cell_data_func(col,cell,render_subject,NULL,NULL);
		gtk_tree_view_column_set_sort_column_id(col,COLUMN_SUBJECT);
		gtk_tree_view_append_column(tree,col);
		gtk_tree_view_set_expander_column(tree,col);

		cell=gtk_cell_renderer_text_new();
		col=gtk_tree_view_column_new_with_attributes(_("Score"),cell,NULL);
		gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED);
		gtk_tree_view_column_set_fixed_width(col,40);
		gtk_tree_view_column_set_resizable(col,TRUE);
		gtk_tree_view_column_set_alignment(col,1);
		gtk_tree_view_column_set_cell_data_func(col,cell,render_score,NULL,NULL);
		gtk_tree_view_column_set_sort_column_id(col,COLUMN_SCORE);
		gtk_tree_view_append_column(tree,col);
		
		cell=gtk_cell_renderer_text_new();
		col=gtk_tree_view_column_new_with_attributes(_("Lines"),cell,NULL);
		gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED);
		gtk_tree_view_column_set_fixed_width(col,40);
		gtk_tree_view_column_set_resizable(col,TRUE);
		gtk_tree_view_column_set_alignment(col,1);
		gtk_tree_view_column_set_cell_data_func(col,cell,render_lines,NULL,NULL);
		gtk_tree_view_column_set_sort_column_id(col,COLUMN_LINES);
		gtk_tree_view_append_column(tree,col);
		
		cell=gtk_cell_renderer_text_new();
		col=gtk_tree_view_column_new_with_attributes(_("Author"),cell,NULL);
		gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED);
		gtk_tree_view_column_set_fixed_width(col,120);
		gtk_tree_view_column_set_resizable(col,TRUE);
		gtk_tree_view_column_set_cell_data_func(col,cell,render_author,NULL,NULL);
		gtk_tree_view_column_set_sort_column_id(col,COLUMN_AUTHOR);
		gtk_tree_view_append_column(tree,col);
		
		cell=gtk_cell_renderer_text_new();
		col=gtk_tree_view_column_new_with_attributes(_("Date"),cell,NULL);
		gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED);
		gtk_tree_view_column_set_fixed_width(col,120);
		gtk_tree_view_column_set_resizable(col,TRUE);
		gtk_tree_view_column_set_cell_data_func(col,cell,render_date,NULL,NULL);
		gtk_tree_view_column_set_sort_column_id(col,COLUMN_DATE);
		gtk_tree_view_append_column(tree,col);
		
		//gui_restore_column_widths(GTK_WIDGET(tree), "header_pane_2");
	}
	g_object_set(treev,"fixed-height-mode",TRUE,NULL);

	//g_object_set_data_full(G_OBJECT(tree), "pan_renderer", retval, header_pane_renderer_free);

	/* wrap it in a scrolled window */
	s_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(s_window),
		GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(s_window), GTK_WIDGET(tree));

	gtk_box_pack_start(GTK_BOX(vbox), s_window, TRUE, TRUE, 0);
	
	g_signal_connect(tree,"destroy",(GCallback)tree_destroy,NULL);
	g_signal_connect(tree,"style-set",(GCallback)style_set_cb,NULL);
	
	return vbox;
}

void header_pane_renderer_update_columns(void)
{
	int i;

	g_return_if_fail (Pan.article_tree!=NULL);

	/* FIXME: this could/should also update the column widths ? */

/* 	gtk_clist_freeze (_article_clist);
	for (i=0; i<articlelist_column_qty; ++i)
		gtk_clist_set_column_title (_article_clist, i, column_to_title (i));
	gtk_clist_thaw (_article_clist);
 */
}

void header_pane_renderer_reset_style_nolock(void)
{
	header_pane_renderer_reset_style();
	
	//gdk_window_invalidate_rect(
}
/***
****
****
****   SORTING
****
****
***/

int articlelist_columns[] = {
	COLUMN_ACTION_STATE,
	COLUMN_ARTICLE_STATE,
	COLUMN_SUBJECT,
	COLUMN_SCORE,
	COLUMN_LINES,
	COLUMN_AUTHOR,
	COLUMN_DATE
};
int articlelist_column_qty = G_N_ELEMENTS (articlelist_columns);

static int get_column_number_from_column_type (int type)
{
	int i;
	for (i=0; i<articlelist_column_qty; ++i)
		if (articlelist_columns[i] == type)
			return i;

	/* column is not visible */
	return -1;
}


/***
****
****
****  POPULATING THE TREE:  UTILITY FUNCTIONS
****
****
***/



static const Icon* get_column_state_icon(const Article * a)
{
	const Icon * retval = NULL;
	const gboolean is_read = article_is_read(a);
	const gboolean is_new = article_is_new(a) && !is_read;
	const int multipart_state = article_get_multipart_state(a);

	g_return_val_if_fail (a!=NULL, NULL);

	if (article_get_error_flag(a))
		retval = _icons + ICON_ERROR;
	else if (multipart_state == MULTIPART_STATE_ALL)
		retval = _icons + ICON_COMPLETE;
	else if (multipart_state == MULTIPART_STATE_SOME)
		retval = _icons + ICON_INCOMPLETE;
	else if (is_new)
		retval = _icons + ICON_NEW;
	else if (!is_new && !is_read)
		retval = _icons + ICON_UNREAD;
	else
		retval = _icons + ICON_READ;

	return retval;
}

static const Icon* get_column_action_icon(const Article * a)
{
	const Icon * retval = NULL;

	g_return_val_if_fail(IS_ARTICLE(a), NULL);

	if (article_is_flagged(a))
		retval = _icons + ICON_FLAGGED_FOR_DL;
	else if (article_get_decode_state(a) == DECODE_STATE_DECODED)
		retval = _icons + ICON_BINARY_DECODED;
	else if (queue_get_article_status(a))
		retval = _icons + ICON_QUEUED;
	else if (article_is_cached(a))
		retval = _icons + ICON_CACHED;

	return retval;
}

#define NORMAL 0
#define NEW_REPLIES 1

static void ensure_styles_inited(void)
{
	int i;
	GtkStyle * s;

	if (!rebuild_style)
		return;
	rebuild_style = FALSE;

	score_column_index = get_column_number_from_column_type(COLUMN_SCORE);
	subject_column_index = get_column_number_from_column_type(COLUMN_SUBJECT);

	/* build the styles */
	s = gtk_widget_get_style(main_window);
	
	for (i=0; i<2; ++i)
	{
		int j;
		PangoFontDescription * pfd = NULL;

		/* find the font description */
		if (i==NORMAL)
			pfd = normal_pfd;
		else if (i==NEW_REPLIES)
			pfd = new_replies_pfd;
		else
			g_assert(0);

		/* build the normal style */
		if (normal_style[i] != NULL)
			g_object_unref(G_OBJECT(normal_style[i]));
		normal_style[i] = gtk_style_copy(s);
		normal_style[i]->fg[0] = unread_color;
		if (pfd != NULL) 
		{
			PangoFontDescription * tmp = normal_style[i]->font_desc;
			normal_style[i]->font_desc = pango_font_description_copy(pfd);
			pango_font_description_free(tmp);
		}

		/* build the read style */
		if (read_style[i] != NULL)
			g_object_unref(G_OBJECT(read_style[i]));
		read_style[i] = gtk_style_copy(s);
		read_style[i]->fg[0] = read_color;
		if (pfd != NULL) 
		{
			PangoFontDescription * tmp = read_style[i]->font_desc;
			read_style[i]->font_desc = pango_font_description_copy(pfd);
			pango_font_description_free(tmp);
		}

		for (j=0; j<SCORE_COLOR_QTY; ++j) 
		{
			if (score_style[j][i] != NULL)
				g_object_unref(G_OBJECT(score_style[j][i]));
			score_style[j][i] = gtk_style_copy(s);
			score_style[j][i]->base[0] = score_color[j][0];
			score_style[j][i]->fg[0] = score_color[j][1];
			if (pfd != NULL)
			{
				PangoFontDescription * tmp = score_style[j][i]->font_desc;
				score_style[j][i]->font_desc = pango_font_description_copy(pfd);
				pango_font_description_free(tmp);
			}
		}
	}
}

static GtkStyle* get_node_style(GtkTreeModel *model, GtkTreeIter  * iter,
							const Article       * article)
{
	int score;

	GtkStyle * style = NULL;
	const gboolean is_unread = !article_is_read(article);
	const gboolean is_new = article_is_new(article) && is_unread;
	const gboolean new_children = article->new_children>0;
	const gboolean unread_children = article->unread_children>0;
	GtkTreeView *tree=GTK_TREE_VIEW(Pan.article_tree);
	gboolean collapsed_subthread=FALSE;
	
	if(gtk_tree_model_iter_has_child(model,iter))
	{
		gtk_tree_model_get(model,iter,ARTICLE_COL_EXPANDED,&collapsed_subthread,-1);
		collapsed_subthread=!collapsed_subthread;
	}
	const gboolean do_mark_new = is_new || (collapsed_subthread && new_children);
	const gboolean do_mark_unread = is_unread || (collapsed_subthread && unread_children);
	
	ensure_styles_inited();
	
	if (do_mark_new)
		style = normal_style[NEW_REPLIES];
	else if (do_mark_unread)
		style = normal_style[NORMAL];
	else
		style = read_style[NORMAL];
	
	return style;
}

static GtkStyle* get_node_style_color(GtkTreeModel *model, GtkTreeIter  * iter,
							const Article       * article)
{
	int score;

	GtkStyle * style = NULL;
	const gboolean is_unread = !article_is_read(article);
	const gboolean is_new = article_is_new(article) && is_unread;
	const gboolean new_children = article->new_children>0;
	const gboolean unread_children = article->unread_children>0;
	GtkTreeView *tree=GTK_TREE_VIEW(Pan.article_tree);
	gboolean collapsed_subthread=FALSE;
	if(gtk_tree_model_iter_has_child(model,iter))
	{
		gtk_tree_model_get(model,iter,ARTICLE_COL_EXPANDED,&collapsed_subthread,-1);
		collapsed_subthread=!collapsed_subthread;
	}
	const gboolean do_mark_new = is_new || (collapsed_subthread && new_children);
	const gboolean do_mark_unread = is_unread || (collapsed_subthread && unread_children);

	ensure_styles_inited();

	/* find the score to use */
	if (article == NULL)
		score = 0;
	else if (collapsed_subthread)
		score = article->subthread_score;
	else 
		score = article_get_score(article,articlelist_get_group()->id);

	/* if article has a score, then change the style */
	if (score != 0)
	{
		int color_index;

		/* get the color index from the score mode */
		switch (filter_score_get_score_mode (score))
		{
			case SCORE_WATCHED:  color_index = SCORE_COLOR_WATCHED; break;
			case SCORE_HIGH:     color_index = SCORE_COLOR_HIGH; break;
			case SCORE_MEDIUM:   color_index = SCORE_COLOR_MEDIUM; break;
			case SCORE_LOW:      color_index = SCORE_COLOR_LOW; break;
			case SCORE_IGNORED:  color_index = SCORE_COLOR_IGNORED; break;
			default:             color_index = SCORE_COLOR_WATCHED; pan_warn_if_reached (); break;
		}

		style = score_style[color_index][do_mark_new ? NEW_REPLIES : NORMAL];
	}
	else
		style=get_node_style(model,iter,article);
	
	return style;
}

static int find_missing_part_qty(const Article * a)
{
	int missing_qty;
	int parts;
	
	g_return_val_if_fail (IS_ARTICLE(a), -1);
	g_return_val_if_fail (a->multipart_state == MULTIPART_STATE_SOME, -1);
	
	parts=article_get_num_parts(a);
	missing_qty=a->parts-parts;
	return missing_qty;
}

static void get_formatted_subject(char            * buf,
                       gulong            buflen,
                       const Article   * a,
                       gboolean          expanded)
{
	gboolean show_missing_qty;
	gboolean show_followup_qty;

	g_return_if_fail (buf!=NULL);
	g_return_if_fail (buflen!=0ul);
	g_return_if_fail (a!=NULL);

	show_followup_qty = !expanded && (a->new_children>0);
	show_missing_qty = article_get_multipart_state(a) == MULTIPART_STATE_SOME;

	if (!show_followup_qty && !show_missing_qty)
		article_get_subject(a,buf,buflen);
	else
	{
		int len=article_get_subject(a,buf,buflen);
		const int missing_qty = show_missing_qty ? find_missing_part_qty(a) : 0;
		if (show_followup_qty && show_missing_qty)
			g_snprintf(buf+len, buflen-len, _(" (%u) [%d missing pieces]"), a->new_children, missing_qty);
		else if (show_followup_qty)
			g_snprintf(buf+len, buflen-len, " (%u)", a->new_children);
		else /* show_missing_qty */
			g_snprintf(buf+len, buflen-len, _(" [%d missing pieces]"), missing_qty);
	}

}

/***
****
***/


void header_pane_renderer_reset_style(void)
{
        PangoFontDescription * old_normal_pfd = normal_pfd;
        PangoFontDescription * old_new_replies_pfd = new_replies_pfd;
                                                                                                                           
        if (header_pane_custom_font_enabled)
		{
                normal_pfd = pango_font_description_from_string(header_pane_custom_font);
                new_replies_pfd = pango_font_description_copy(normal_pfd);
        }
		else
		{
                GtkStyle * style = gtk_widget_get_default_style();
                normal_pfd = pango_font_description_copy(style->font_desc);
                new_replies_pfd = pango_font_description_copy(style->font_desc);
        }
        pango_font_description_set_weight(new_replies_pfd, PANGO_WEIGHT_BOLD);
        //gtk_widget_modify_font(GTK_WIDGET(tree), new_replies_pfd);
                                                                                                                           
        pango_font_description_free(old_normal_pfd);
        pango_font_description_free(old_new_replies_pfd);

        rebuild_style = TRUE;
}

static void render_action(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	Article *a;
	gtk_tree_model_get(model,iter,ARTICLE_COL_OBJECT,&a,-1);
	const Icon *icon=get_column_action_icon(a);
	g_object_set(cell,"pixbuf",icon?icon->pixbuf:NULL,NULL);
	return;
}
static void render_state(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	Article *a;
	gtk_tree_model_get(model,iter,ARTICLE_COL_OBJECT,&a,-1);
	const Icon *icon=get_column_state_icon(a);
	g_object_set(cell,"pixbuf",icon?icon->pixbuf:NULL,NULL);
	return;
}
static void render_subject(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	Article *a;
	GtkStyle *style;
	gboolean expanded;
	GdkColor fg,base;
	char buf[1024];
	gtk_tree_model_get(model,iter,ARTICLE_COL_OBJECT,&a,ARTICLE_COL_EXPANDED,&expanded,-1);
	
	if (header_pane_color_subject_column)
		style=get_node_style_color(model,iter,a);
	else
		style=get_node_style(model,iter,a);
	memset(buf,0,1024);
	get_formatted_subject(buf,1024,a,expanded);
	fg=style->fg[0];
	base=style->base[0];
	g_object_set(cell,"text",buf,/* "foreground-gdk",fg,
				"background-gdk",base, */"font-desc",style->font_desc,NULL);
	return;
}
static void render_lines(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	Article *a;
	GtkStyle *style;
	char buf[10];
	gtk_tree_model_get(model,iter,ARTICLE_COL_OBJECT,&a,-1);
	
	style=get_node_style(model,iter,a);
	snprintf(buf,10,"%d",a->line_qty);
	g_object_set(cell,"text",buf,"font-desc",style->font_desc,NULL);
	return;
}
static void render_date(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	Article *a;
	GtkStyle *style;
	char *freeme;
	gtk_tree_model_get(model,iter,ARTICLE_COL_OBJECT,&a,-1);
	
	style=get_node_style(model,iter,a);
	// FIXME: make this take  buffer?
	freeme = get_date_display_string(a->date, header_pane_date_format);
	g_object_set(cell,"text",freeme,"font-desc",style->font_desc,NULL);
	g_free(freeme);
	return;
}
static void render_author(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	Article *a;
	GtkStyle *style;
	char buf[1024];
	gtk_tree_model_get(model,iter,ARTICLE_COL_OBJECT,&a,-1);
	
	style=get_node_style(model,iter,a);
	article_get_short_author_str(a,buf,1024);
	g_object_set(cell,"text",buf,"font-desc",style->font_desc,NULL);
	return;
}
static void render_score(GtkTreeViewColumn *column, GtkCellRenderer *cell,
						   GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	Article *a;
	GtkStyle *style;
	gboolean expanded;
	GdkColor fg,base;
	char buf[8];
	gtk_tree_model_get(model,iter,ARTICLE_COL_OBJECT,&a,ARTICLE_COL_EXPANDED,&expanded,-1);
	const int score=(!expanded && gtk_tree_model_iter_has_child(model,iter))?a->subthread_score:article_get_score(a,articlelist_get_group()->id);
	// FIXME: something wrong here
	if (header_pane_color_score_column)
		style=get_node_style_color(model,iter,a);
	else
		style=get_node_style(model,iter,a);
	snprintf(buf,6,"%d",score);
	fg=style->fg[0];
	base=style->base[0];
	g_object_set(cell,"text",buf,/* "foreground-gdk",fg,
				"background-gdk",base, */"font-desc",style->font_desc,NULL);
	return;
}

